/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.basislib.berechtigung;

/**
 * Klasse OberflaechenFunktion.
 * 
 * Ersetzt die in TBuV-104 geforderte XML-Datei zur Definition von Funktionen. Der Entwickler kann mit dieser Klasse
 * eine Funktion anlegen und zustzlich einen {@link IOberflaechenBerechtigungZuhoerer} angeben, der ihn ber nderungen
 * der Berechtigung informiert.
 * 
 * @version $Revision: 1.2 $ / $Date: 2009/12/08 08:50:11 $ / ($Author: HofmannJ $)
 * 
 * @author HofmannJ
 * 
 */
public class OberflaechenFunktion
{
  /**
   * Name der Funktion.
   */
  private String _funktionsName = null;

  /**
   * Alias-Name der Funktion.
   */
  private String _funktionsAlias = null;

  /**
   * Eindeutige ElementId fr ein Element, dessen Oberflchenberechtigung geprft werden soll.
   */
  private String _elementId = null;

  /**
   * Eindeutige ID des Dialogs oder des Views, in der die OberflaechenBerechtigungsFunktion bentigt wird.
   */
  private String _dialogId = null;

  /**
   * Listener der Funktion.
   */
  private IOberflaechenBerechtigungZuhoerer _zuhoerer = null;

  /**
   * Konstruktor fr eine OberflaechenFunktion.
   * 
   * @param funktionsName
   *          Name der Funktion.
   * @param listener
   *          {@link IOberflaechenBerechtigungZuhoerer} der Funktion.
   * @param dialogId
   *          Eindeutige Identifizierung des Dialogs, fr den die Berechtigung geprft werden soll.
   * @param elementId
   *          Eindeutige Id des Elements, fr dass die Berechtigung geprft werden soll.
   * 
   */
  public OberflaechenFunktion(String funktionsName, IOberflaechenBerechtigungZuhoerer listener, String dialogId,
      String elementId)
  {
    this._funktionsName = funktionsName;
    this._zuhoerer = listener;
    this._dialogId = dialogId;
    this._elementId = elementId;
  }

  /**
   * Konstruktor fr ein OberflaechenFunktion.
   * 
   * @param funktionsName
   *          Name der Funktion.
   * @param funktionsAlias
   *          Alias-Name der Funktion.
   * @param listener
   *          {@link IOberflaechenBerechtigungZuhoerer} der Funktion.
   * @param dialogId
   *          Identifizierung des Dialogs, fr den die Berechtigung geprft werden soll.
   * @param elementId
   *          Eindeutige Id des Elements, fr dass die Berechtigung geprft werden soll.
   * 
   */
  public OberflaechenFunktion(String funktionsName, String funktionsAlias, IOberflaechenBerechtigungZuhoerer listener,
      String dialogId, String elementId)
  {
    this._funktionsName = funktionsName;
    this._funktionsAlias = funktionsAlias;
    this._zuhoerer = listener;
    this._dialogId = dialogId;
    this._elementId = elementId;
  }

  /**
   * Liefert den Namen der Funktion.
   * 
   * @return Funktionsname
   */
  public String getFunktionsName()
  {
    return _funktionsName;
  }

  /**
   * Liefert den Alias der Funktion.
   * 
   * @return Funktionsalias
   */
  public String getFunktionsAlias()
  {
    return _funktionsAlias;
  }

  /**
   * Liefert die ElementID, auf die sich diese Oberflchenfunktion bezieht.
   * 
   * @return ElementID.
   */
  public String getElementId()
  {
    return _elementId;
  }

  /**
   * Liefert die DialogID, auf die sich diese Oberflchenfunktion bezieht.
   * 
   * @return DialogId
   */
  public String getDialogId()
  {
    return _dialogId;
  }

  /**
   * Liefert den {@link IOberflaechenBerechtigungZuhoerer} der Funktion.
   * 
   * @return {@link IOberflaechenBerechtigungZuhoerer} der Funktion.
   */
  public IOberflaechenBerechtigungZuhoerer getZuhoerer()
  {
    return _zuhoerer;
  }

}
