/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.basislib.einstellungen;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferenceStore;
import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import de.bsvrz.buv.rw.basislib.Activator;
import de.bsvrz.buv.rw.basislib.einstellungen.hilfsklassen.DefinitionenEinstellungen;
import de.bsvrz.buv.rw.basislib.einstellungen.hilfsklassen.PraeferenzSeiteVoreinstellungenOptionen;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Diese Klasse speichert Einstellungen als Parameter innerhalb einer
 * XML-Struktur. <br> {@link EinstellungenSpeicher}
 * <table>
 * <th>Methode:</th>
 * <th>Aktion:</th>
 * <tr>
 * <td>{@link EinstellungenSpeicher#hinzufuegenParameter(String, String)}</td>
 * <td>Hinzufgen eines Parameters zu den Einstellungen.</td>
 * </tr>
 * <tr>
 * <td>{@link EinstellungenSpeicher#editiereParameter(String, String)}</td>
 * <td>Editieren eines vorhandenen Parameters.</td>
 * </tr>
 * <tr>
 * <td>{@link EinstellungenSpeicher#ueberschreibeParameter(String, String)}</td>
 * <td>berschreiben eines vorhandenen Parameters, legt diesen neu an sofern
 * noch nicht vorhanden.</td>
 * </tr>
 * <tr>
 * <td>{@link EinstellungenSpeicher#loescheParameter(String)}</td>
 * <td>Lscht einen Parameter.</td>
 * </tr>
 * <tr>
 * <td>{@link EinstellungenSpeicher#getParameter(String)}</td>
 * <td>Liefert einen bestimmten Parameter.</td>
 * </tr>
 * <tr>
 * <td>{@link EinstellungenSpeicher#getAlleParameter()}</td>
 * <td>Liefert eine {@link HashMap} mit allen vorhandenen Parametern.</td>
 * </tr>
 * </table>
 * <br>
 * <b>Achtung:</b> Ein Parameter darf maximal <i>32000</i> Zeichen lang sein.
 * Wurde ein Parameter editiert, ist der {@link EinstellungenSpeicher} durch die
 * entsprechenden Methoden der Klasse {@link EinstellungenWaehler} zu sichern. <br>
 * <br>
 * Durch Synchronisation sind die ffentlichen Methoden dieser Klasse
 * Threadsicher!
 * 
 * @author Dambach Werke GmbH
 * 
 */
public class EinstellungenSpeicher
{
	/**
	 * Der Debugger
	 */
	private Debug _debug = Debug.getLogger();

	/**
	 * JDOM Document, in dem die Verarbeitung erfolgt
	 */
	private Document _doc = null;

	/**
	 * Wurzelelement des JDOM-Baums
	 */
	private Element _elementWurzel = null;

	/**
	 * Schmutziges Flag, gibt an, ob etwas gendert oder hinzugefgt wurde, was
	 * erst noch gespeichert werden muss.
	 * 
	 */
	private boolean _schmutzig = false;

	/**
	 * Maximale Anzahl von Zeichen pro TeilParameter
	 */
	private static final int _zeichenProTeilParameter = 32000;

	/**
	 * Maximale Anzahl von Zeichen pro Eintrag = 2097152 ~ 2 MB
	 */
	private static final int _maxZeichen = 2 * 1024 * 1024;

	/**
	 * Gibt an, ober der der Wert eines Eintrags um einen Startstring oder einen
	 * Endstring ergnzt werden soll.
	 */
	private boolean _eintragErgaenzen = true;

	/**
	 * Startstring, um welchen ein Eintrag ergnzt werden kann.
	 */
	private String _startString = "[START]";

	/**
	 * Endstring, um welchen ein Eintrag ergnzt werden kann.
	 */
	private String _endeString = "[ENDE]";

	public EinstellungenSpeicher()
	{
		init();
	}

	/**
	 * Fhrt die Initialisierung durch.
	 * 
	 */
	private void init()
	{
		erzeugeXmlBasisDoc();
	}

	/**
	 * Liest eine XML-Struktur aus einer XML-Struktur ein.
	 * 
	 * @param xmlString
	 *          String mit XML-Struktur.
	 */
	public synchronized void einlesenXmlAusString(String xmlString)
	{
		// synchronized (this)
		// {
		SAXBuilder builder = new SAXBuilder();
		builder.setValidation(false);
		try
		{
			StringReader reader = new StringReader(xmlString);

			Document tempDoc = builder.build(reader);
			if (tempDoc.getRootElement().getName().equals(
					EinstellungenBezeichnerXml.elementEinstellungen))
			{
				// alles ok
				_doc = tempDoc;
				_elementWurzel = _doc.getRootElement();
				_schmutzig = false;

			} else
			{
				System.out
						.println(getClass().getName()
								+ "::einlesenXmlAusString => Fehler: Ungltiges Format der bergebenen XML-Struktur.");
			}

			// _doc = builder.build(reader);
			// _elementWurzel = _doc.getRootElement();
		} catch (JDOMException e)
		{
			_debug.fine(getClass().getName()
					+ "::einlesenXmlAusString => JDOMException: \n" + e.toString());
		} catch (IOException e)
		{
			_debug.fine(getClass().getName()
					+ "::einlesenXmlAusString => IOException: \n" + e.toString());
		}
		// }
	}

	/**
	 * Liest eine XML-Datei ein
	 * 
	 * @param dateiPfad
	 *          Kompletter Pfad zur Datei
	 * @throws IllegalArgumentException
	 *           Wird geworfen, wenn die Datei nicht existiert
	 */
	public synchronized String einlesenXmlAusDatei(String dateiPfad)
			throws IllegalArgumentException
	{
		// synchronized (this)
		// {
		// SAXBuilder builder = new SAXBuilder();
		StringBuffer dateiInhalt = new StringBuffer();

		File file = new File(dateiPfad);
		if (file.exists())
		{
			FileReader fr;
			try
			{
				String line;
				fr = new FileReader(file);
				BufferedReader br = new BufferedReader(fr);
				while ((line = br.readLine()) != null)
				{
					dateiInhalt.append(line);
					dateiInhalt.append("\n");
				}
				einlesenXmlAusString(dateiInhalt.toString());

			} catch (FileNotFoundException e)
			{
				_debug.fine(getClass().getName()
						+ "::einlesenXmlausDatei => Die Datei existiert nicht: \n"
						+ e.toString());
			} catch (IOException ioe)
			{
				_debug.fine(getClass().getName()
						+ "::einlesenXmlausDatei => IO-Fehler: \n" + ioe.toString());
			}
		} else
		{
			throw new IllegalArgumentException("Die Datei " + dateiPfad
					+ " existiert nicht!");
		}
		// }

		return dateiInhalt.toString();
	}

	/**
	 * Erzeugt das Basisdokument, lehnt sich an toolbar.dtd an
	 * 
	 */
	private void erzeugeXmlBasisDoc()
	{
		_doc = new Document();
		_elementWurzel = new Element(
				EinstellungenBezeichnerXml.elementEinstellungen);
		_doc.setRootElement(_elementWurzel);
	}

	/**
	 * Liefert eine formatierte Ausgabe der erzeugten XML-Struktur
	 * 
	 * @param mitZeitStempel
	 *          Wenn <code>true</code>, dann wird im Wurzelelement ein Attribut
	 *          <code>zeitstempel</code> untergebracht, sonst nicht.
	 * @return Formatierte XML-Struktur
	 */
	public String ausgabeXmlFormatiert(boolean mitZeitStempel)
	{
		synchronized (this)
		{
			String returnString = null;

			SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");
			String datum = sdf.format(new Date());

			if (mitZeitStempel)
			{
				Attribute attTs = new Attribute("zeitstempel", datum);
				_elementWurzel.setAttribute(attTs);
			}

			if (_doc == null)
			{
				_doc = new Document(_elementWurzel);
			}

			Format prettyFormat = Format.getPrettyFormat();
			XMLOutputter out = new XMLOutputter(prettyFormat);
			Format f = out.getFormat();
			f.setEncoding(EinstellungenBezeichnerXml.xmlEncoding);
			out.setFormat(f);

			returnString = out.outputString(_doc);

			_debug.fine(getClass().getName()
					+ "::ausgabeXmlFormatiert => Ausgabestring:\n" + returnString);

			_schmutzig = false;
			return returnString;
		}
	}

	/**
	 * Schreibt einen String in eine Datei
	 * 
	 * @param dateiPfad
	 *          Kompletter Pfad zur Datei, in die geschrieben werden soll
	 * @param schreibString
	 *          String, der in die Datei geschrieben werden soll
	 * @param ueberschreiben
	 *          true == berschreiben der Datei, falls schon vorhanden
	 * @return Liefert <code>true</code>, wenn die Datei erfolgreich geschrieben
	 *         werden konnte, sonst <code>false</code>.
	 * @throws IllegalArgumentException
	 *           Wir geworfen, wenn die Datei existiert und nicht berschrieben
	 *           werden kann
	 */
	public boolean schreibeString(String dateiPfad, String schreibString,
			boolean ueberschreiben) throws IllegalArgumentException
	{
		boolean erfolgreich = false;

		synchronized (this)
		{
			File file = new File(dateiPfad);

			// File existiert, auf berschreiben prfen
			if (file.exists() && ueberschreiben)
			{
				try
				{
					PrintWriter pw = new PrintWriter(file);
					pw.print(schreibString);
					pw.close();
					erfolgreich = true;
				} catch (FileNotFoundException e)
				{
					e.printStackTrace();
				}
			}
			// File existiert nicht und kann einfach angelegt werden
			else if (!file.exists())
			{
				try
				{
					PrintWriter pw = new PrintWriter(file);
					pw.print(schreibString);
					pw.close();
					erfolgreich = true;
				} catch (FileNotFoundException e)
				{
					e.printStackTrace();
				}
			}
			// File existiert und ueberschreiben == flase
			else
			{
				throw new IllegalArgumentException("Die Datei " + dateiPfad
						+ " existiert und ueberschreiben == flase");
			}
		}

		return erfolgreich;
	}

	/**
	 * Fgt einen Parameter zu den Voereinstellungen hinzu, sofern dieser dort
	 * noch nicht vorhanden ist.
	 * 
	 * @param id
	 *          Id des Parameters
	 * @param wert
	 *          Wert des Parameters, maximal 2 * 1024 ^ 2Zeichen.
	 * @return Liefert <code>true</code>, wenn der Parameter gesetzt werden konnte
	 *         und <code>false</code>, falls dieser schon vorhanden war oder
	 *         Aufgrund der Lnge des Wertes (max 2 * 1024 ^ 2 Zeichen) nicht
	 *         gesetzt werden konnte.
	 */
	public boolean hinzufuegenParameter(String id, String wert)
	{
		synchronized (this)
		{
			if (wert.length() <= _maxZeichen)
			{
				String elementParameter = EinstellungenBezeichnerXml.elementParameter;
				@SuppressWarnings("unchecked")
				List kinder = _elementWurzel.getChildren(elementParameter);
				if (kinder.size() == 0)
				{
					_elementWurzel.addContent(erzeugeParameterElement(id, wert));
					_schmutzig = true;
					return true;
				} else
				{
					if (!pruefeObParameterVorhanden(id))
					{
						Element parameter = erzeugeParameterElement(id, wert);
						_elementWurzel.addContent(parameter);
						_schmutzig = true;
						return true;
					} else
					{
						// _debug.info(getClass().getName()
						// + "::hinzufuegenParameter \n\t=>Der Parameter mit der ID " + id
						// + " ist in den Einstellungen bereits vorhanden! Breche ab.");
						return false;
					}
				}
			} else
			{
				return false;
			}
		}
	}

	/**
	 * Editiert einen vorhandenen Parameter.
	 * 
	 * @param id
	 *          Id des Parameters.
	 * @param wert
	 *          Wert des Parameters, maximal 2 * 1024 ^ 2 Zeichen.
	 * @return Liefert <code>true</code>, wenn der Parameter editiert werden
	 *         konnte, sonnst <code>false</code>.
	 */
	public boolean editiereParameter(String id, String wert)
	{
		boolean erfolgreich = false;

		if (wert.length() <= _maxZeichen)
		{
			synchronized (this)
			{
				if (pruefeObParameterVorhanden(id))
				{
					// Element parameter = getElement(id);
					// parameter.setText(wert);
					// _schmutzig = true;
					// erfolgreich = true;

					loescheParameter(id);
					_schmutzig = true;
					erfolgreich = hinzufuegenParameter(id, wert);
				} else
				{
					erfolgreich = hinzufuegenParameter(id, wert);
				}
			}
		}

		return erfolgreich;
	}

	/**
	 * berschreibt einen Parameter. Existiert dieser noch nicht, wird er
	 * angelegt.
	 * 
	 * @param id
	 *          Id des Parameters
	 * @param wert
	 *          Wert des Parameters maximal 2 * 1024 ^ 2 Zeichen.
	 */
	public boolean ueberschreibeParameter(String id, String wert)
	{
		boolean erfolgreich = false;

		if (wert.length() <= _maxZeichen)
		{
			synchronized (this)
			{
				_schmutzig = true;
				if (pruefeObParameterVorhanden(id))
				{
					erfolgreich = editiereParameter(id, wert);
				} else
				{
					erfolgreich = hinzufuegenParameter(id, wert);
				}
			}
		}

		return erfolgreich;
	}

	/**
	 * Lscht einen Parameter.
	 * 
	 * @param id
	 *          Id des Parameters, der gelscht werden soll.
	 * @return Liefert <code>true</code>, wenn der Parameter gelscht werden
	 *         konnte, sonst <code>false</code>.
	 */
	public boolean loescheParameter(String id)
	{
		synchronized (this)
		{
			if (pruefeObParameterVorhanden(id))
			{
				Element loeschElement = getElement(id);
				loescheElement(loeschElement);
				_schmutzig = true;
				return true;
			} else
			{
				_debug
						.error(getClass().getName()
								+ "::loescheElement \n\t=>Der Parameter mit der ID "
								+ id
								+ " ist in den Einstellungen nicht vorhanden und wird damit auch nicht gelscht! Breche ab.");
				return false;
			}
		}
	}

	/**
	 * Liefert den Wert eines Parameters.
	 * 
	 * @param id
	 *          Id des Parameters.
	 * @return Wert des Parameters falls dieser existiert, sonst <code>null</code>
	 *         .
	 */
	public synchronized String getParameter(String id)
	{
		// synchronized (this)
		// {
		String returnString = null;
		if (pruefeObParameterVorhanden(id))
		{
			Element parameter = getElement(id);

			/*
			 * Prfen ob ein- oder mehrteilig
			 */
			String attMehrteilig = parameter
					.getAttributeValue(EinstellungenBezeichnerXml.attMehrteilig);
			if (attMehrteilig == null)
			{
				/*
				 * Nicht mehrteilig
				 */
				returnString = entpackeWert(parameter.getText());
			} else if (attMehrteilig.equalsIgnoreCase("false"))
			{
				/*
				 * Nicht mehrteilig
				 */
				returnString = entpackeWert(parameter.getText());
			} else if (attMehrteilig.equalsIgnoreCase("true"))
			{
				/*
				 * Mehrteilig!!!
				 */
				// String strAnzahlTeile = parameter
				// .getAttributeValue(EinstellungenBezeichnerXml.attAnzahlTeile);
				StringBuffer stringBuffer = new StringBuffer();
				@SuppressWarnings("unchecked")
				Iterator<Element> itTeilParam = parameter.getChildren(
						EinstellungenBezeichnerXml.elementTeilParameter).iterator();
				while (itTeilParam.hasNext())
				{
					Element teilParameter = itTeilParam.next();
					// System.out.println("Lese Teil: "
					// + teilParameter
					// .getAttributeValue(EinstellungenBezeichnerXml.attTeilId));
					stringBuffer.append(entpackeWert(teilParameter.getText()));
				}

				returnString = stringBuffer.toString();

			}

		} else
		{
			// _debug.info(getClass().getName()
			// + "::getParameter \n\t=>Der Parameter mit der ID " + id
			// + " ist in den Einstellungen nicht vorhanden! Breche ab.");
		}

		return returnString;
		// }
	}

	/**
	 * Liefert eine HashMap, bestehend aus der ID des Parameters als Key (vom Typ
	 * String) und dem Wert des Paramerer als Value (vom Typ String). <br>
	 * <b>Achtung:</b> nderungen wirken sich <b><i>nicht</i></b> aus die
	 * Hintergrunddatenstruktur aus. Diese HashMap bietet nur lesenden Zugriff!
	 * 
	 * @return HashMap mit der ID des Parameters als Key und dem Wert des
	 *         Parameters als Value
	 */
	public synchronized HashMap<String, String> getAlleParameter()
	{
		// synchronized (this)
		// {
		HashMap<String, String> hm = new HashMap<String, String>();
		@SuppressWarnings("unchecked")
		List<Element> parameter = _elementWurzel
				.getChildren(EinstellungenBezeichnerXml.elementParameter);
		Iterator<Element> it = parameter.iterator();
		while (it.hasNext())
		{
			Element elementParameter = it.next();
			String id = elementParameter
					.getAttributeValue(EinstellungenBezeichnerXml.attId);
			hm.put(id, getParameter(id));
		}

		return hm;
		// }
	}

	/**
	 * Liefert das Schmutzig-Flag zurck. Ist dieses gesetzt, mssen die Daten
	 * dieser Klasse erst noch gesichert werden.
	 * 
	 * @return Schmutzig Flag.
	 */
	public boolean istSchmutzig()
	{
		synchronized (this)
		{
			return _schmutzig;
		}
	}

	/**
	 * Erzeugt ein neuen Parameter als {@link Element}.
	 * 
	 * @param id
	 *          Id des Parameters.
	 * @param wert
	 *          Wert des Parameters.
	 * @return Parameter als {@link Element}.
	 */
	private Element erzeugeParameterElement(String id, String wert)
	{
		Element parameter = new Element(EinstellungenBezeichnerXml.elementParameter);
		Attribute attId = new Attribute(EinstellungenBezeichnerXml.attId, id);
		parameter.setAttribute(attId);

		int anzahlTeile = berechneAnzahlTeile(wert);

		if (anzahlTeile == 1)
		{
			wert = verpackeWert(wert);
			parameter.setText(wert);
		} else
		{
			Attribute attMehrteilig = new Attribute(
					EinstellungenBezeichnerXml.attMehrteilig, "true");
			parameter.setAttribute(attMehrteilig);
			Attribute attAnzTeile = new Attribute(
					EinstellungenBezeichnerXml.attAnzahlTeile, "" + anzahlTeile);
			parameter.setAttribute(attAnzTeile);

			/*
			 * String aufsplitten und in TeilParameter Elemente aufteilen
			 */
			int idx = 0;
			for (int i = 0; i < anzahlTeile; i++)
			{
				Element elementTeilParameter = new Element(
						EinstellungenBezeichnerXml.elementTeilParameter);
				Attribute teilId = new Attribute(EinstellungenBezeichnerXml.attTeilId,
						"" + i);
				elementTeilParameter.setAttribute(teilId);

				if (idx + _zeichenProTeilParameter < wert.length())
				{
					String wertSplit = wert
							.substring(idx, idx + _zeichenProTeilParameter);

					wertSplit = verpackeWert(wertSplit);

					elementTeilParameter.setText(wertSplit);
				} else
				{
					String wertSplit = wert.substring(idx, wert.length());

					wertSplit = verpackeWert(wertSplit);

					elementTeilParameter.setText(wertSplit);
				}

				parameter.addContent(elementTeilParameter);

				idx = idx + _zeichenProTeilParameter;
			}
		}

		return parameter;
	}

	/**
	 * Prft ob der Parameter mit bestimmter ID schon vorhanden ist.
	 * 
	 * @param id
	 *          Id des Parameters
	 * @return Liefert <code>true</code>, wenn der Parameter schon vorhanden ist,
	 *         sonst <code>false</code>.
	 */
	private boolean pruefeObParameterVorhanden(String id)
	{
		boolean vorhanden = false;
		@SuppressWarnings("unchecked")
		List<Element> kinder = _elementWurzel
				.getChildren(EinstellungenBezeichnerXml.elementParameter);
		Iterator<Element> it = kinder.iterator();
		while (it.hasNext())
		{
			Element element = it.next();
			if (element.getName().equals(EinstellungenBezeichnerXml.elementParameter))
			{
				if (element.getAttributeValue(EinstellungenBezeichnerXml.attId).equals(
						id))
				{
					vorhanden = true;
					break;
				}
			}
		}

		return vorhanden;
	}

	/**
	 * Liefert das {@link Element} eines Parameters.
	 * 
	 * @param id
	 *          Id des Parameters.
	 * @return {@link Element} des Parameters.
	 */

	private Element getElement(String id)
	{
		Element returnElement = null;
		@SuppressWarnings("unchecked")
		List<Element> kinder = _elementWurzel
				.getChildren(EinstellungenBezeichnerXml.elementParameter);
		Iterator<Element> it = kinder.iterator();
		while (it.hasNext())
		{

			Element element = it.next();
			if (element.getName().equals(EinstellungenBezeichnerXml.elementParameter))
			{
				if (element.getAttributeValue(EinstellungenBezeichnerXml.attId).equals(
						id))
				{
					returnElement = element;
					break;
				}

			}

		}
		return returnElement;

	}

	/**
	 * Lscht ein {@link Element}.
	 * 
	 * @param element
	 *          Zu lschendes {@link Element}. Wird via detach-Methode aus dem
	 *          JDOM-Baum entfernt.
	 */
	private void loescheElement(Element element)
	{
		element.detach();
	}

	/**
	 * Prft, ob es sich bei der ID fr einen Parameter um eine Ausnahme handelt
	 * oder nicht.<br>
	 * Ausnahmen sind die folgenden Parameter:
	 * <ul>
	 * <li>xmlStrukturMenuleiste</li>
	 * <li>xmlStrukturSymbolleite</li>
	 * <li>xmlStrukturStatusleiste</li>
	 * </ul>
	 * 
	 * @param id
	 *          Zu prfende ParameterID
	 * @return Liefert <code>true</code>, wenn es sich um eine Ausnahme handelt,
	 *         sonst <code>false</code>.
	 */
	@SuppressWarnings("unused")
	private boolean istAusnahme(String id)
	{
		if ((id.equals("xmlStrukturMenuleiste"))
				|| (id.equals("xmlStrukturSymbolleite"))
				|| (id.equals("xmlStrukturStatusleiste")))
		{
			return true;
		} else
		{
			return false;
		}
	}

	/**
	 * Berechnet in wie viele Teile ein String zwecks Aufteilung i.d. XML-Struktur
	 * zur Anpassung an den DaV aufgespalten wird. Die Berechnung erfolgt an Hand
	 * der Vorgabe durch die Konstante
	 * {@link EinstellungenSpeicher#_zeichenProTeilParameter}.
	 * 
	 * @param pruefString
	 *          Zu prfender String.
	 * @return Anzahl der Teile, in welche dieser String aufgespalten wird.
	 */
	private int berechneAnzahlTeile(String pruefString)
	{
		int rest = pruefString.length() % _zeichenProTeilParameter;
		int wertOhneRest = pruefString.length() - rest;
		int anz = wertOhneRest / _zeichenProTeilParameter;

		if (rest > 0)
		{
			anz++;
		}

		return anz;
	}

	/**
	 * Prft ber den Parameter <code>paramEintragErgaenzen</code> der Plug-In
	 * {@link PreferenceStore}s, ob der Wert eines Parameters mit Start und Stop
	 * Strings ergnzt werden soll oder nicht. Die Parameter
	 * <code>paramStartString</code> und <code>paramEndeString</code> der Plug-In
	 * Prefernces definieren die zu ergnzenden Start & Stop-Strings. Theoretisch
	 * knnte die Konfiguration ber
	 * {@link PraeferenzSeiteVoreinstellungenOptionen} erfolgen - nach Ansage von
	 * Herrn Kni wurde dies Seite jedoch entfernt.
	 */
	protected void syncMitPraeferenzen()
	{
		IPreferenceStore store = Activator.getDefault().getPreferenceStore();

		if (store != null)
		{
			this._eintragErgaenzen = store
					.getBoolean(DefinitionenEinstellungen.PARAM_EINTRAG_ERGAENZEN);

			this._startString = store
					.getString(DefinitionenEinstellungen.PARAM_START_STRING);
			this._endeString = store
					.getString(DefinitionenEinstellungen.PARAM_ENDE_STRING);
		}

	}

	/**
	 * Prft, ob der Wert eines Parameters mittels der Start/Stop-Ergnzungen
	 * (ber Flag{@link EinstellungenSpeicher#_eintragErgaenzen}) verpackt werden
	 * soll oder nicht und liefert ihn verpackt oder unverpackt zurck.
	 * 
	 * @param wert
	 *          Wert Zu verpackender Wert.
	 * @return Verpackter/Nichtverpackter Wert.
	 */
	protected String verpackeWert(String wert)
	{
		syncMitPraeferenzen();

		if (_eintragErgaenzen)
		{
			StringBuffer sb = new StringBuffer();
			sb.append(_startString);
			sb.append(wert);
			sb.append(_endeString);
			return sb.toString();
		} else
		{
			return wert;
		}
	}

	/**
	 * Prft an Hand des Flags {@link EinstellungenSpeicher#_eintragErgaenzen} ob
	 * ein Wert entpackt werden soll oder nicht.
	 * 
	 * @param wert
	 *          Wert, der entpackt werden soll.
	 * @return Liefert den entpackten/nichtentpackten Wert.
	 */
	protected String entpackeWert(String wert)
	{
		String re = "";

		syncMitPraeferenzen();
		// if (_eintragErgaenzen)
		// {
		// if (wert.length() >= _startString.length())
		// {
		// if ((wert.substring(0, _startString.length())).equals(_startString))
		// {
		// System.out.println("START: "
		// + wert.substring(0, _startString.length()));
		//
		// if ((wert.substring(wert.length() - _endeString.length(), wert
		// .length())).equals(_endeString))
		// {
		// System.out.println("STOP: "
		// + wert.substring(wert.length() - _endeString.length(), wert
		// .length()));
		// System.out.println("DA");
		//
		// re = wert.substring(_startString.length(), wert.length()
		// - _endeString.length());
		// }
		// }
		// }
		// }
		// else
		// {
		// re = wert;
		// }

		/*
		 * Start Index des eigentlichen Wertes
		 */
		int ptrStart = 0;

		/*
		 * Ende Index des eigentlichen Wertes
		 */
		int ptrEnde = wert.length();

		/*
		 * Startstring prfen
		 */
		if (wert.length() >= _startString.length())
		{
			if ((wert.substring(0, _startString.length())).equals(_startString))
			{
				ptrStart = _startString.length();
			}
		}

		/*
		 * Endstring prfen
		 */
		if (wert.length() >= _endeString.length())
		{
			if ((wert.substring(wert.length() - _endeString.length(), wert.length()))
					.equals(_endeString))
			{
				ptrEnde = wert.length() - _endeString.length();
			}
		}

		re = wert.substring(ptrStart, ptrEnde);

		return re;
	}

	public void test()
	{
		String testWert = _startString + " Hallo Welt! " + _endeString;
		System.out.println("->" + testWert + "<-");
		System.out.println("=>" + entpackeWert(testWert) + "<=\n");

		_eintragErgaenzen = true;
		String wert1 = " Hallo Welt! ";
		String wert2 = "";
		for (int i = 1; i < 1000 * 3 * 2; i++)
		{
			wert2 = wert2 + " AAA BBB CCC ";
		}

		hinzufuegenParameter("1", wert1);
		hinzufuegenParameter("2", wert2);
		System.out.println(ausgabeXmlFormatiert(true));
		einlesenXmlAusString(ausgabeXmlFormatiert(true));

		String par1 = getParameter("1");
		String par2 = getParameter("2");

		System.out.println("wert1 == par1 => " + wert1.equals(par1));
		System.out.println("wert2 == par2 => " + wert2.equals(par2));

	}

	public static void main(String argv[])
	{
		EinstellungenSpeicher es = new EinstellungenSpeicher();
		es.test();
	}

}
