/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.basislib.ort.hilfsklassen;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;

import de.bsvrz.buv.rw.basislib.ort.hilfsklassen.StrassenTeilSegment.AsbStationierung;
import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.sys.funclib.debug.Debug;


/**
 * Diese Klasse stellt smtliche Informationen bereit die notwendig sind, um
 * Ortsangaben ber uere Straensegmente durchzufhren. Es werden geliefert:<br>
 * <ul>
 * <li>Straen</li>
 * <li>Alle Knoten der ASB-Ortsangabe, die auf dieser Strae verfgbar sind</li>
 * <li>Informationen ber die zu einem ASB-Knoten zugehrigen, erreichbaren
 * Knoten</li>
 * <li>Prfung, ob eine ASB-Stationierung im Bereich zwischen zwei Knoten liegt</li>
 * </ul>
 * 
 * @author Dambach Werke GmbH
 * 
 */
public class AsbOrtsangabeAlt
{
	/**
	 * Der Debugger
	 */
	private Debug _debug = Debug.getLogger();

	/**
	 * Verbindung zum DaV
	 */
	private ClientDavConnection _verbindung = null;

	/**
	 * Enthlt die Instanz der Klasse {@link KonfigurationNeu}, welche die hier
	 * bentigten Datenstrukturen zur Verfgung stellt.
	 */
	private KonfigurationNeu _konf = null;

	/**
	 * {@link Struktur} in positiver TMC-Richtung
	 */
	private Struktur _strukturPos = null;

	/**
	 * {@link Struktur} in negativer TMC-Richtung
	 */
	private Struktur _strukturNeg = null;

	/**
	 * Referenz auf die {@link AsbKnotenVerwaltung}, positive Richtung.
	 */
	private AsbKnotenVerwaltung _asbKnotenVerwaltungPositiv = null;

	/**
	 * Referenz auf die {@link AsbKnotenVerwaltung}, negative Richtung.
	 */
	private AsbKnotenVerwaltung _asbKnotenVerwaltungNegativ = null;

	/**
	 * Speichert eine Liste aller zu einer Strae verfgbaren AsbKnoten. Wird
	 * durch die Methode {@link AsbOrtsangabeAlt#filtereAsbKnoten()}, die von der
	 * Methode {@link AsbOrtsangabeAlt#initialisiereStrukturNachStrasse(String)}
	 * aufgerufen wird, gefllt.
	 */
	private LinkedList<AsbKnoten> _listeKnotenAsbGesamt = new LinkedList<AsbKnoten>();

	/**
	 * Konstruktor. bernimmt die Verbindung zum DaV und fhrt die
	 * Basis-Initialisierung durch.
	 * 
	 * @param verbindung
	 */
	public AsbOrtsangabeAlt(ClientDavConnection verbindung)
	{
		this._verbindung = verbindung;
		einlesenBasisStruktur();
	}

	/**
	 * Liest die Basisstruktur ein.
	 */
	private void einlesenBasisStruktur()
	{
		if (_verbindung != null)
		{
			_konf = KonfigurationNeu.getInstanz();
			_konf.bestimmeObjekte(_verbindung);
		} else
		{
			_debug
					.error(getClass().getName()
							+ "::einlesenBasisStruktur \n\t=>Es wurde keine Verbindung zum DaV bergeben. Kann die Datenstrukturen nicht aus dem DaV einlesen.");
		}

	}

	/**
	 * Baut die Struktur auf.
	 * 
	 * @param pidSelektierteStrasse
	 *          PID der Strae, zu der die Struktur aufgebaut werden soll.
	 */
	public void initialisiereStrukturNachStrasse(String pidSelektierteStrasse)
	{
		_konf.filtereNachStrasse(pidSelektierteStrasse);

		/*
		 * Liste der uerenStraenSegmente in positiver Richtung
		 * (atg.ueresStraenSegment Attribut TmcRichtung)
		 */
		LinkedList<AeusseresStrassenSegment> aeussereStrassenSegmentePositiv = new LinkedList<AeusseresStrassenSegment>();

		/*
		 * Liste der uerenStraenSegmente in negativer Richtung
		 * (atg.ueresStraenSegment Attribut TmcRichtung)
		 */
		LinkedList<AeusseresStrassenSegment> aeussereStrassenSegmenteNegativ = new LinkedList<AeusseresStrassenSegment>();

		/*
		 * Liste der Knoten der uerenStraenSegmente in positiver Richtung
		 */
		LinkedList<StrassenKnoten> knotenPositiv = new LinkedList<StrassenKnoten>();

		/*
		 * Liste der Knoten der uerenStraenSegmente in negativer Richtung
		 */
		LinkedList<StrassenKnoten> knotenNegativ = new LinkedList<StrassenKnoten>();

		Iterator<Entry<String, AeusseresStrassenSegment>> it = _konf
				.getAeussereStrassenSegmenteGefiltert().entrySet().iterator();
		while (it.hasNext())
		{
			Entry<String, AeusseresStrassenSegment> entry = it.next();
			AeusseresStrassenSegment aeusseresStrassenSegment = entry.getValue();
			if (aeusseresStrassenSegment.getRichtungStrasse().equals("positiv"))
			{
				aeussereStrassenSegmentePositiv.add(aeusseresStrassenSegment);
				// VON-Knoten
				if ((aeusseresStrassenSegment.getVonKnoten() != null)
						&& (!knotenPositiv
								.contains(aeusseresStrassenSegment.getVonKnoten())))
				{
					knotenPositiv.add(aeusseresStrassenSegment.getVonKnoten());
				}
				// NACH-Knoten
				if ((aeusseresStrassenSegment.getNachKnoten() != null)
						&& (!knotenPositiv.contains(aeusseresStrassenSegment
								.getNachKnoten())))
				{
					knotenPositiv.add(aeusseresStrassenSegment.getNachKnoten());
				}

			} else if (aeusseresStrassenSegment.getRichtungStrasse()
					.equals("negativ"))
			{
				aeussereStrassenSegmenteNegativ.add(aeusseresStrassenSegment);
				// VON-Knoten
				if ((aeusseresStrassenSegment.getVonKnoten() != null)
						&& (!knotenNegativ
								.contains(aeusseresStrassenSegment.getVonKnoten())))
				{
					knotenNegativ.add(aeusseresStrassenSegment.getVonKnoten());
				}

				// NACH-KNOTEN
				if ((aeusseresStrassenSegment.getNachKnoten() != null)
						&& (!knotenNegativ.contains(aeusseresStrassenSegment
								.getNachKnoten())))
				{
					knotenNegativ.add(aeusseresStrassenSegment.getNachKnoten());
				}
			}

		}

		_debug.fine("uereStraenSegmente (positiv): "
				+ aeussereStrassenSegmentePositiv.size());
		_debug.fine("uereStraenSegmente (negativ): "
				+ aeussereStrassenSegmenteNegativ.size());
		_debug.fine("StraenKnoten (positiv): " + knotenPositiv.size());
		_debug.fine("StraenKnoten (negativ): " + knotenNegativ.size());

		/*
		 * Struktur (TmcRichtung positiv) erzeugen
		 */
		_strukturPos = new Struktur("positiv");

		// uereStraenSegmente hinzufgen
		Iterator<AeusseresStrassenSegment> itAePos = aeussereStrassenSegmentePositiv
				.iterator();
		while (itAePos.hasNext())
		{
			AeusseresStrassenSegment aeusseresStrassenSegment = itAePos.next();
			_strukturPos.erzeugeStrukturElement(aeusseresStrassenSegment);
		}

		// StraenKnoten hinzufgen
		Iterator<StrassenKnoten> itSKPos = knotenPositiv.iterator();
		while (itSKPos.hasNext())
		{
			StrassenKnoten strassenKnoten = itSKPos.next();
			_strukturPos.erzeugeStrukturElement(strassenKnoten);
		}

		// Bestimmen der Struktur in TmcRichtung positiv
		_strukturPos.bestimmeVorgaengerNachfolger();

		/*
		 * Struktur (TmcRichtung negativ) erzeugen
		 */
		_strukturNeg = new Struktur("negativ");

		// urereStraenSegmente zur negativen Struktur hinzufgen
		Iterator<AeusseresStrassenSegment> itAeNeg = aeussereStrassenSegmenteNegativ
				.iterator();
		while (itAeNeg.hasNext())
		{
			AeusseresStrassenSegment aeusseresStrassenSegment = itAeNeg.next();
			_strukturNeg.erzeugeStrukturElement(aeusseresStrassenSegment);
		}

		// StraenKnoten zur negativen Struktur hizufgen.
		Iterator<StrassenKnoten> itSKNeg = knotenNegativ.iterator();
		while (itSKNeg.hasNext())
		{
			StrassenKnoten strassenKnoten = itSKNeg.next();
			_strukturNeg.erzeugeStrukturElement(strassenKnoten);
		}

		// Bestimmen der Struktur in TmcRichtung negativ
		_strukturNeg.bestimmeVorgaengerNachfolger();

//		 gebeStrukturAus(_strukturPos);
//		 gebeStrukturAus(_strukturNeg);

		Iterator<StrukturElement> itkombi = getVonAuswahlfeld().iterator();
		int j = 0;
		while (itkombi.hasNext())
		{
			StrukturElement element = itkombi.next();
			// _debug.fine("E" + j + ": " + element.toString());

			if (_strukturNeg.containsValue(element))
			{
				// _debug.fine("NEG OK");

			} else
			{
				// _debug.fine("NEG FALSE");
			}

			if (_strukturPos.containsValue(element))
			{
				// _debug.fine("POS OK");
			} else
			{
				// _debug.fine("POS FALSE");
			}

			LinkedList<StrukturElementMitPfad> li = bestimmeErreichbareKnotenAeste(element);
			_debug.fine("SIZE: " + li.size());

			// Iterator<StrukturElementMitPfad> itTe = li.iterator();
			// int zaehler = 0;
			// while (itTe.hasNext())
			// {
			// StrukturElementMitPfad pfad = itTe.next();
			// zaehler++;
			// }

			j++;
		}

		/**
		 * Herausfiltern der ASB-Knoten, passend zur aktuell initialisierten Strae.
		 */
		filtereAsbKnoten();
	}

	/**
	 * Gibt die Elemente einer {@link Struktur} auf der Konsole aus.
	 * 
	 * @param struktur
	 *          Auszugebende {@link Struktur}
	 */
	protected void gebeStrukturAus(Struktur struktur)
	{
		/*
		 * Ermitteln der Elemente ohne Vorgnger
		 */
		StrukturElement ohneVor = struktur.bestimmeStrukturElementOhneVorgaenger();

		// System.out.println("Strukturelement ohne Vorgnger \n\t=> "
		// + ohneVor.toString());

		int i = 0;
		Iterator<StrukturElement> itAus = struktur.bestimmeNachfolgerFuerAusgabe(
				ohneVor).iterator();
		while (itAus.hasNext())
		{
			StrukturElement element = itAus.next();
			System.out.println("StrukturElement " + i + ": " + element.toString());
			i++;
		}

	}

	/**
	 * Liefert die nach Key sortierte Kombination der {@link StrukturElement}e
	 * der positiven und negativen {@link Struktur} als {@link LinkedList} zurck.<br>
	 * <br>
	 * 
	 * Diese Liste bietet sich zum Fllen des <i>ersten Auswahlfeldes</i> an, in
	 * dem ein Knoten bzw. ein Ast (ueres Straensegment) der Strae ausgewhlt
	 * werden kann.
	 * 
	 * @return LinkedList der {@link StrukturElement}e der positiven und
	 *         negativen {@link Struktur}en.
	 */
	public LinkedList<StrukturElement> getVonAuswahlfeld()
	{
		/**
		 * Enthlt alle positiven Strukturelemente, nach KEY sortiert.
		 */
		TreeMap<String, StrukturElement> kombiPos = new TreeMap<String, StrukturElement>();

		/**
		 * Enthlt alle negativen Strukturelemente, nach KEY sortiert.
		 * 
		 */
		TreeMap<String, StrukturElement> kombiNeg = new TreeMap<String, StrukturElement>();

		/*
		 * Positive Struktur hinzufgen
		 */
		if ((_strukturPos != null) && (_strukturPos.size() > 0))
		{
			Iterator<Entry<String, StrukturElement>> itPos = _strukturPos
					.getStrukturElemente().entrySet().iterator();

			while (itPos.hasNext())
			{
				Entry<String, StrukturElement> entry = itPos.next();
				if (!kombiPos.containsKey(entry.getKey()))
				{
					kombiPos.put(entry.getKey(), entry.getValue());
					_debug.fine(getClass().getName()
							+ "::kombiniereStrukturen PUT positives StrukturElement: "
							+ entry.getValue());

				} else
				{
					// Element bereits vorhanden, nicht erneut hinzufgen
				}

			}

		}

		/*
		 * Negative Struktur hinzufgen
		 */
		if ((_strukturNeg != null) && (_strukturNeg.size() > 0))
		{
			Iterator<Entry<String, StrukturElement>> itNeg = _strukturNeg
					.getStrukturElemente().entrySet().iterator();

			while (itNeg.hasNext())
			{
				Entry<String, StrukturElement> entry = itNeg.next();
				if (!kombiNeg.containsKey(entry.getKey()))
				{
					kombiNeg.put(entry.getKey(), entry.getValue());
					_debug.fine(getClass().getName()
							+ "::kombiniereStrukturen PUT positives StrukturElement: "
							+ entry.getValue());

				} else
				{
					// Element bereits vorhanden, nicht erneut hinzufgen
				}
			}
		}

		/*
		 * Liste erstellen und zurckliefern
		 */
		LinkedList<StrukturElement> ru = new LinkedList<StrukturElement>();

		Iterator<Entry<String, StrukturElement>> it = kombiPos.entrySet()
				.iterator();
		while (it.hasNext())
		{
			Entry<String, StrukturElement> entry = it.next();
			if (!ru.contains(entry.getValue()))
			{
				ru.add(entry.getValue());
			} else
			{
				_debug
						.error("Das StrukturElement ("
								+ entry.getValue().toStringLang()
								+ ") ist bereits i.d. sortierten und kombinierten Liste der StrukturElemente vorhanden!");

			}
		}

		Iterator<Entry<String, StrukturElement>> itNeg = kombiNeg.entrySet()
				.iterator();
		while (itNeg.hasNext())
		{
			Entry<String, StrukturElement> entry = itNeg.next();
			if (!ru.contains(entry.getValue()))
			{
				ru.add(entry.getValue());
			} else
			{
				_debug
						.error("Das StrukturElement ("
								+ entry.getValue().toStringLang()
								+ ") ist bereits i.d. sortierten und kombinierten Liste der StrukturElemente vorhanden!");

			}
		}

		return ru;
	}

	/**
	 * Bestimmt die von einem {@link StrukturElement} aus erreichbaren Knoten und
	 * ste. Diese werden in einer Liste vom Typ {@link StrukturElementMitPfad}
	 * zurckgegeben, so dass zu jedem erreichbaren {@link StrukturElement}
	 * (gekapselt als {@link StrukturElementMitPfad}) eine Pfadangabe vom
	 * Ausgangs {@link StrukturElement} verfgbar ist.
	 * 
	 * @param element
	 *          {@link StrukturElement}, von dem die erreichbaren
	 *          {@link StrukturElement}e bestimmt werden sollen.
	 * @return Liste der erreichbaren {@link StrukturElement}e (gekapselt als
	 *         {@link StrukturElementMitPfad}) inklusive Pfad vom Ausgangs
	 *         {@link StrukturElement} an.
	 */
	public LinkedList<StrukturElementMitPfad> bestimmeErreichbareKnotenAeste(
			StrukturElement element)
	{
		TreeMap<String, StrukturElementMitPfad> pos = bestimmeIntern(_strukturPos,
				element);
		TreeMap<String, StrukturElementMitPfad> neg = bestimmeIntern(_strukturNeg,
				element);

		LinkedList<StrukturElementMitPfad> liste = new LinkedList<StrukturElementMitPfad>();

		if ((pos != null) && (pos.size() > 0))
		{
			Iterator<Entry<String, StrukturElementMitPfad>> itPos = pos.entrySet()
					.iterator();
			while (itPos.hasNext())
			{
				Entry<String, StrukturElementMitPfad> entry = itPos.next();
				liste.add(entry.getValue());
			}
		}

		if ((neg != null) && (neg.size() > 0))
		{

			Iterator<Entry<String, StrukturElementMitPfad>> itNeg = neg.entrySet()
					.iterator();
			while (itNeg.hasNext())
			{
				Entry<String, StrukturElementMitPfad> entry = itNeg.next();
				liste.add(entry.getValue());
			}
		}

		return liste;
	}

	/**
	 * Bestimmt alle von einem {@link StrukturElement} erreichbaren
	 * {@link StrukturElement}e innerhalb einer {@link Struktur}. Der Pfad der
	 * erreichbaren {@link StrukturElement}e wird zusammen mit dem Ausgangs
	 * {@link StrukturElement} in einer Instanz der Klasse
	 * {@link StrukturElementMitPfad} gekapselt.
	 * 
	 * 
	 * @param struktur
	 *          {@link Struktur}, innerhalb der alle von einem
	 *          {@link StrukturElement} erreichbaren {@link StrukturElement}e
	 *          bestimmt werden sollen.
	 * @param element
	 *          Ausgangs {@link StrukturElement}, von dem ausgehend die
	 *          erreichbaren {@link StrukturElement}e innerhalb der
	 *          {@link Struktur} bestimmt werden.
	 * @return Alle von einem Ausgangs- {@link StrukturElement} aus erreichbaren
	 *         {@link StrukturElement}e. Wird gekapselt in einer TreeMap als
	 *         {@link StrukturElementMitPfad} und beinhaltet sowohl den Pfad als
	 *         auch das erreichte {@link StrukturElement}.
	 */
	private TreeMap<String, StrukturElementMitPfad> bestimmeIntern(
			Struktur struktur, StrukturElement element)
	{
		TreeMap<String, StrukturElementMitPfad> map = null;

		if (struktur.containsValue(element))
		{
			if (map == null)
			{
				map = new TreeMap<String, StrukturElementMitPfad>();
			}

			/*
			 * Bestimmen der Vor- und Nachfolgerlisten
			 */
			LinkedList<StrukturElement> nachfolgerListe = struktur
					.bestimmeNachfolgerListe(element);
			LinkedList<StrukturElement> vorgaengerListe = struktur
					.bestimmeVorgaengerListe(element);

			// Nachfolger durchlaufen
			for (int i = 1; i < nachfolgerListe.size(); i++)
			{
				List<StrukturElement> subList = nachfolgerListe.subList(0, i + 1);
				LinkedList<StrukturElement> tempListNachfolger = new LinkedList<StrukturElement>();

				tempListNachfolger.addAll(subList);

				StrukturElementMitPfad elementMitPfad = new StrukturElementMitPfad(
						nachfolgerListe.get(i), true, tempListNachfolger);
				if (!map.containsKey(elementMitPfad.getKey()))
				{
					map.put(elementMitPfad.getKey(), elementMitPfad);
				} else
				{
					//Das Element (Nachfolger) ist bereits vorhanden!
				}
			}

			// Vorgnger durchlaufen
			for (int i = 1; i < vorgaengerListe.size(); i++)
			{
				List<StrukturElement> subList = vorgaengerListe.subList(0, i + 1);
				LinkedList<StrukturElement> tempListVorgaenger = new LinkedList<StrukturElement>();
				tempListVorgaenger.addAll(subList);

				StrukturElementMitPfad elementMitPfad = new StrukturElementMitPfad(
						vorgaengerListe.get(i), false, tempListVorgaenger);
				if (!map.containsKey(elementMitPfad.getKey()))
				{
					map.put(elementMitPfad.getKey(), elementMitPfad);
				} else
				{
					//Das Element ist bereits (Vorgnger) vorhanden!
				}
			}
		}

		return map;
	}

	/**
	 * Liefert eine Liste aller verfgbaren ASB-Knoten zurck. Wurden die
	 * Strukturen nach keiner Strae initialisiert (vgl. Method
	 * {@link AsbOrtsangabeAlt#initialisiereStrukturNachStrasse(String)}), wird eine
	 * leere Liste zurckgeliefert.
	 * 
	 * @return Liste aller verfgbaren {@link AsbKnoten}. Kann leer sein, wenn
	 *         nach keiner Strae initialisiert wurde.
	 */
	public LinkedList<AsbKnoten> getVerfuegbareAsbKnoten()
	{
		return _listeKnotenAsbGesamt;
	}

	/**
	 * Bestimmt eine Liste aller von einem {@link AsbKnoten} aus erreichbaren
	 * {@link AsbKnoten}. Die Liste ist leer, wenn von dem {@link AsbKnoten} aus
	 * keine Knoten erreichbar sind oder keine Knotenverwaltung initialisiert
	 * werden konnte.
	 * 
	 * @param ausgangsKnoten
	 *          Ausgangs {@link AsbKnoten}, von dem aus alle erreichbaren
	 *          {@link AsbKnoten} bestimmt werden sollen.
	 * @return Eine Liste der erreichbaren {@link AsbKnoten}. Kann leer sein,
	 *         wenn keine {@link AsbKnoten} erreichbar sind oder keine
	 *         {@link AsbKnotenVerwaltung} initialisiert werden konnte.
	 */
	public LinkedList<AsbKnoten> bestimmeErreichbareAsbKnoten(
			AsbKnoten ausgangsKnoten)
	{
		LinkedList<AsbKnoten> re = new LinkedList<AsbKnoten>();

		AsbKnotenVerwaltung knotenVerwaltung = null;

		if (ausgangsKnoten.getTmcRichtung().equals(EnumTmcRichtung.positiv))
		{
			knotenVerwaltung = _asbKnotenVerwaltungPositiv;
		} else if (ausgangsKnoten.getTmcRichtung().equals(EnumTmcRichtung.negativ))
		{
			knotenVerwaltung = _asbKnotenVerwaltungNegativ;
		}

		if (knotenVerwaltung != null)
		{
			re = knotenVerwaltung
					.bestimmeListeAllerErreichbarenKnoten(ausgangsKnoten);
		} else
		{
			_debug.warning(getClass().getName()
					+ "::bestimmeErreichbareAsbKnoten => Es konnte keine "
					+ "Knotenverwaltung zum Bestimmung der erreichbaren "
					+ "Knoten bestimmt werden.");
		}

		return re;
	}

	/**
	 * Prft, ob eine angegebene Stationierung im Bereich zwischen zwei Knoten
	 * vorhanden ist bzw. gltig ist.
	 * 
	 * @param knotenA
	 *          {@link AsbKnoten} A
	 * @param knotenB
	 *          {@link AsbKnoten} B
	 * @param stationierung
	 *          Stationierung zwischen den beiden Knoten, die geprft werden soll.
	 *          Die Angabe erfolgt in Metern.
	 * @return Liefert <code>true</code>, wenn die Stationierung zwischen den
	 *         beiden Knoten gltig ist, sonst <code>false</code>.
	 */
	public boolean pruefeAsbStationierung(AsbKnoten knotenA, AsbKnoten knotenB,
			long stationierung)
	{
		boolean re = false;

		if (knotenA.getTmcRichtung().equals(knotenB.getTmcRichtung()))
		{
			AsbKnotenVerwaltung knotenVerwaltung = null;
			if (knotenA.getTmcRichtung().equals(EnumTmcRichtung.positiv))
			{
				knotenVerwaltung = _asbKnotenVerwaltungPositiv;
			} else if (knotenA.getTmcRichtung().equals(EnumTmcRichtung.negativ))
			{
				knotenVerwaltung = _asbKnotenVerwaltungNegativ;
			}

			if (knotenVerwaltung != null)
			{
				re = knotenVerwaltung.istInBereich(knotenA, knotenB, stationierung);
			} else
			{
				_debug.warning(getClass().getName()
						+ "::pruefeAsbStationierung => Es konnte keine Knotenverwaltung "
						+ "zum Prfen der Stationierung bestimmt werden.");
			}

		} else
		{
			_debug.warning(getClass().getName()
					+ "::pruefeAsbStationierung => Die TMC-Richtungen der "
					+ "bergebenen Objekte ist nicht gleich. KnotenA: " + knotenA
					+ " KnotenB: " + knotenB);
		}

		return re;
	}

	/**
	 * Filtert die zu einer Strae gehrigen ASB-Knoten an Hand der durch die
	 * Methode {@link AsbOrtsangabeAlt#initialisiereStrukturNachStrasse(String)}
	 * bestimmten Strukturen <code>_sturkturPos</code> und
	 * <code>_strukturNeg</code>.
	 */
	private void filtereAsbKnoten()
	{
		_asbKnotenVerwaltungPositiv = new AsbKnotenVerwaltung(
				EnumTmcRichtung.positiv);
		_asbKnotenVerwaltungNegativ = new AsbKnotenVerwaltung(
				EnumTmcRichtung.negativ);

		StrukturElement ohneVorgaengerPositiveRichtung = _strukturPos
				.bestimmeStrukturElementOhneVorgaenger();

		StrukturElement ohneVorgaengerNegativeRichtung = _strukturNeg
				.bestimmeStrukturElementOhneVorgaenger();

		bestimmeAsbKnotenAnfangEnde(ohneVorgaengerPositiveRichtung,
				_asbKnotenVerwaltungPositiv);
		bestimmeAsbKnotenAnfangEnde(ohneVorgaengerNegativeRichtung,
				_asbKnotenVerwaltungNegativ);

		 System.out.println("===POSITIV START===");
		 System.out.println(_asbKnotenVerwaltungPositiv);
		 System.out.println("===POSITIV ENDE===");
		 System.out.println("===Negativ Start===");
		 System.out.println(_asbKnotenVerwaltungNegativ);
		 System.out.println("===Negativ Ende===");

		LinkedList<AsbKnoten> listeAllerKnotenPos = _asbKnotenVerwaltungPositiv
				.bestimmeListeAllerKnoten();
		LinkedList<AsbKnoten> listeAllerKnotenNeg = _asbKnotenVerwaltungNegativ
				.bestimmeListeAllerKnoten();
		_listeKnotenAsbGesamt = new LinkedList<AsbKnoten>();
		_listeKnotenAsbGesamt.addAll(listeAllerKnotenPos);
		_listeKnotenAsbGesamt.addAll(listeAllerKnotenNeg);
		Collections.sort(_listeKnotenAsbGesamt);

	}

	private void bestimmeAsbKnotenAnfangEnde(StrukturElement ohneVorgaenger,
			AsbKnotenVerwaltung asbKnotenVerwaltung)
	{
		/*
		 * Bestimmen der erreichbaren Knoten und ste
		 */
		LinkedList<StrukturElementMitPfad> erreichbareKnotenAeste = bestimmeErreichbareKnotenAeste(ohneVorgaenger);
		Iterator<StrukturElementMitPfad> it = erreichbareKnotenAeste.iterator();
		while (it.hasNext())
		{
			StrukturElementMitPfad mitPfad = it.next();

			/*
			 * Pfad extrahieren
			 */
			Iterator<StrukturElement> itPfad = mitPfad.getPfad().iterator();
			while (itPfad.hasNext())
			{
				/*
				 * Pfad Element holen
				 */
				StrukturElement elementImPfad = itPfad.next();

				/*
				 * ueres Straensegment?
				 */
				if (elementImPfad.isAeusseresStrassenSegment())
				{
					AeusseresStrassenSegment aes = (AeusseresStrassenSegment) elementImPfad
							.getElement();

					/*
					 * Bestimmen der StraenTeilsegmente des ueren StraenSegmentes
					 */
					Vector<StrassenTeilSegment> strassenTeilSegmente = aes
							.getStrassenTeilSegmente();

					/*
					 * Extrahieren der AsbStationierungen aus den StraenTeilSegmenten
					 */
					Iterator<StrassenTeilSegment> itSts = strassenTeilSegmente.iterator();
					while (itSts.hasNext())
					{
						StrassenTeilSegment strassenTeilSegment = itSts.next();

						LinkedList<AsbStationierung> asbStationierungen = strassenTeilSegment
								.getAsbStationierungen();

						Iterator<AsbStationierung> itAsb = asbStationierungen.iterator();
						while (itAsb.hasNext())
						{
							AsbStationierung asbStationierung = itAsb.next();

							/*
							 * Stationierung zur Verwaltung hinzufgen.
							 */
							asbKnotenVerwaltung.addAsbStationierung(asbStationierung);
						}
					}
				}
			}
		}
	}

	/**
	 * Liefert die Elemente einer Struktur der {@link AeusseresStrassenSegment}e
	 * einer Strae. <br>
	 * Die {@link AeusseresStrassenSegment}e dieser Strae knnen sortiert in
	 * Reihenfolge ausgelesen werden, sofern ein Zusammenhang besteht.
	 * 
	 * @author HofmannJ
	 * 
	 */
	public class Struktur
	{
		/**
		 * Liste der Elemente dieser Struktur
		 */
		private HashMap<String, StrukturElement> _elemente = new HashMap<String, StrukturElement>();

		/**
		 * TMC-Richtung der Elemente dieser Struktur
		 */
		private String _tmcRichtung = null;

		private int _zaehler = 0;

		/**
		 * Konstruktor. bernimmt die TmcRichtung der Elemente dieser
		 * {@link Struktur}.
		 * 
		 * @param tmcRichtungPositiv
		 *          TmcRichtung dieser Struktur
		 */
		public Struktur(String tmcRichtungPositiv)
		{
			_debug.fine(getClass().getName() + "::Struktur \n\t=> Struktur erzeugt.");

			this._tmcRichtung = tmcRichtungPositiv;
		}

		/**
		 * Erzeugt ein neues {@link StrukturElement} und liefert dieses zurck. Ist
		 * bereits eines mit dem gleichen {@link AeusseresStrassenSegment}
		 * vorhanden, so wird dieses zurckgeliefert.
		 * 
		 * @param aeusseresStrassenSegment
		 *          {@link AeusseresStrassenSegment}, aus dem ein
		 *          {@link StrukturElement} erzeugt werden soll.
		 * @return {@link StrukturElement}
		 */
		public StrukturElement erzeugeStrukturElement(
				AeusseresStrassenSegment aeusseresStrassenSegment)
		{
			/*
			 * StrukturElement besorgen
			 */
			StrukturElement strukturElement = null;
			if (containsKey(aeusseresStrassenSegment.getPid()))
			{
				strukturElement = this.getStrukturElement(aeusseresStrassenSegment
						.getPid());
			} else
			{
				strukturElement = new StrukturElement(aeusseresStrassenSegment,
						_tmcRichtung);
				this.addStrukturElement(strukturElement);
			}

			return strukturElement;
		}

		/**
		 * Erzeugt ein neues {@link StrukturElement} und liefert dieses zurck. Ist
		 * bereits eines mit dem gleichen {@link StrassenKnoten} vorhanden, so wird
		 * dieses zurckgeliefert.
		 * 
		 * @param strassenKnoten
		 *          {@link StrassenKnoten}, aus dem ein {@link StrukturElement}
		 *          erzeugt werden soll.
		 * @return {@link StrukturElement}
		 */
		public StrukturElement erzeugeStrukturElement(StrassenKnoten strassenKnoten)
		{
			/*
			 * StrukturElement besorgen
			 */
			StrukturElement strukturElement = null;
			if (containsKey(strassenKnoten.getPid()))
			{
				strukturElement = this.getStrukturElement(strassenKnoten.getPid());
			} else
			{
				strukturElement = new StrukturElement(strassenKnoten, _tmcRichtung);
				this.addStrukturElement(strukturElement);
			}

			return strukturElement;
		}

		/**
		 * Fgt ein {@link StrukturElement} in die {@link HashMap} ein, sofern
		 * dieses noch nicht enthalten ist.
		 * 
		 * @param element
		 *          Hinzuzufgendes {@link StrukturElement}
		 */
		public void addStrukturElement(StrukturElement element)
		{
			if (!_elemente.containsKey(element.getPid()))
			{
				_elemente.put(element.getPid(), element);
			}
		}

		/**
		 * Liefert ein {@link StrukturElement} an Hand ihres Keys aus der
		 * {@link HashMap} zurck.
		 * 
		 * @param pid
		 *          Key des StrukturElements, entsprechend der PID des
		 *          {@link AeusseresStrassenSegment}.
		 * @return {@link StrukturElement}, zugehrig zum Key.
		 */
		public StrukturElement getStrukturElement(String pid)
		{
			if (_elemente.containsKey(pid))
			{
				return _elemente.get(pid);
			} else
			{
				return null;
			}
		}

		/**
		 * Liefert alle {@link StrukturElement}e in einer {@link HashMap} zurck.
		 * 
		 * @return {@link HashMap} (String als Key, {@link StrukturElement} als
		 *         Value)
		 */
		public HashMap<String, StrukturElement> getStrukturElemente()
		{
			return _elemente;
		}

		/**
		 * Liefert die TmcRichtung der Elemente dieser {@link Struktur}.
		 * 
		 * @return TmcRichtung der Elemente dieser {@link Struktur}
		 */
		public String getTmcRichtung()
		{
			return _tmcRichtung;
		}

		/**
		 * Prft, ob ein {@link StrukturElement} mit einem bestimmten Key in der
		 * {@link HashMap} der {@link StrukturElement}e vorhanden ist.
		 * 
		 * @param key
		 *          Schlssel des auf Existenz zu prfenden {@link StrukturElement}s.
		 * @return Liefert <code>true</code>, wenn das Element in der
		 *         {@link HashMap} vorhanden ist, sonst <code>false</code>.
		 */
		public boolean containsKey(String key)
		{
			return _elemente.containsKey(key);
		}

		/**
		 * Prft, ob die {@link HashMap} dieser {@link Struktur} ein bestimmtes
		 * Objekt enthlt.
		 * 
		 * @param object
		 *          Objekt, auf dessen Existenz geprft werden soll.
		 * @return Liefert <code>true</code>, wenn das Objekt in der
		 *         {@link HashMap} vorhanden ist, sonst <code>false</code>.
		 */
		public boolean containsValue(Object object)
		{
			return _elemente.containsValue(object);
		}

		/**
		 * Liefert die Gre der Struktur.
		 * 
		 * @return Gre der Struktur
		 */
		public int size()
		{
			return _elemente.size();
		}

		/**
		 * Bestimmt die Vorgnger und Nachfolger fr jedes {@link StrukturElement}.
		 */
		protected void bestimmeVorgaengerNachfolger()
		{
			Iterator<Entry<String, StrukturElement>> it = _elemente.entrySet()
					.iterator();
			while (it.hasNext())
			{
				Entry<String, StrukturElement> entry = it.next();
				StrukturElement strukturElement = entry.getValue();

				/*
				 * ueresStraenSegment
				 */
				if (strukturElement.isAeusseresStrassenSegment())
				{
					// Aussen
					AeusseresStrassenSegment segment = (AeusseresStrassenSegment) strukturElement
							.getElement();

					StrassenKnoten nachKnoten = segment.getNachKnoten();
					if (nachKnoten != null)
					{
						// Nach
						if (_elemente.containsKey(nachKnoten.getPid()))
						{
							StrukturElement ssNachfolger = _elemente.get(nachKnoten.getPid());
							ssNachfolger.addVorgaenger(strukturElement);

							strukturElement.addNachfolger(ssNachfolger);
						}
					}

					StrassenKnoten vonKnoten = segment.getVonKnoten();
					if (vonKnoten != null)
					{
						// Von
						if (_elemente.containsKey(vonKnoten.getPid()))
						{
							StrukturElement ssVorgaenger = _elemente.get(vonKnoten.getPid());
							ssVorgaenger.addNachfolger(strukturElement);

							strukturElement.addVorgaenger(ssVorgaenger);
						}
					}

				}

				/*
				 * StraenKnoten
				 */
				if (strukturElement.isStrassenKnoten())
				{
					StrassenKnoten knoten = (StrassenKnoten) strukturElement.getElement();

					/*
					 * Abgehend (Nachfolgend)
					 */
					Iterator<AeusseresStrassenSegment> itAbgehend = knoten
							.getAbgehendeAeussereStrassenSegmente().iterator();
					while (itAbgehend.hasNext())
					{
						// Abgehend
						AeusseresStrassenSegment segment = itAbgehend.next();
						if (_elemente.containsKey(segment.getPid()))
						{
							StrukturElement ssNachfolger = _elemente.get(segment.getPid());
							ssNachfolger.addVorgaenger(strukturElement);

							strukturElement.addNachfolger(ssNachfolger);
						}
					}

					/*
					 * Hinzufhrend
					 */
					Iterator<AeusseresStrassenSegment> itHinzu = knoten
							.getHinfuehrendeAeussereStrassenSegmente().iterator();
					while (itHinzu.hasNext())
					{
						// Hinzufhrend
						AeusseresStrassenSegment segment = itHinzu.next();
						if (_elemente.containsKey(segment.getPid()))
						{
							StrukturElement ssVorgaenger = _elemente.get(segment.getPid());
							ssVorgaenger.addNachfolger(strukturElement);

							strukturElement.addVorgaenger(ssVorgaenger);
						}
					}

				}

			}
		}

		/**
		 * Bestimmt das {@link StrukturElement}, welches keinen Vorgnger hat.
		 * 
		 * @return {@link StrukturElement} ohne Vorgnger, sonst <code>null</code>.
		 */
		public StrukturElement bestimmeStrukturElementOhneVorgaenger()
		{
			StrukturElement element = null;

			Iterator<Entry<String, StrukturElement>> it = _elemente.entrySet()
					.iterator();
			while (it.hasNext())
			{
				Entry<String, StrukturElement> entry = it.next();
				int size = entry.getValue().getVorgaenger().size();
				if (size == 0)
				{
					element = entry.getValue();
					break;
				}
			}

			_debug
					.fine(getClass().getName()
							+ "::bestimmeStrukturElementOhneNachfolger \n\t=>Element ohne Vorgaenger: "
							+ element.getPid());

			return element;
		}

		/**
		 * Diese Funktion liefert eine geordnete Liste der
		 * {@link AeusseresStrassenSegment}e.
		 * 
		 * @param element
		 *          {@link StrukturElement}, an Hand dessen Nachfolgern (und damit
		 *          auch Vorgngern) die {@link AeusseresStrassenSegment} in
		 *          Reihenfolge bestimmt werden sollen.
		 * @return Liste mit in Reihenfolgen gebrachten {@link StrukturElement}en.
		 */
		public LinkedList<StrukturElement> bestimmeNachfolgerFuerAusgabe(
				StrukturElement element)
		{
			_debug
					.fine(getClass().getName()
							+ "::bestimmeNachfolger \n\tNachfolger werden bestimmt, siehe nachfolgende Liste.");

			LinkedList<StrukturElement> liste = new LinkedList<StrukturElement>();

			LinkedList<StrukturElement> nach = bestimmeNachfolgerRekursivFuerAusgabe(element);
			if (nach != null)
			{
				liste.addAll(nach);
			}

			return liste;
		}

		/**
		 * Bestimmt <i>rekursiv</i> die Nachfolger eines Elementes <br>
		 * <b>Nicht fr ffentlichen Gebrauch bestimmt.</b>
		 * 
		 * @param element
		 * @return
		 */
		private LinkedList<StrukturElement> bestimmeNachfolgerRekursivFuerAusgabe(
				StrukturElement element)
		{
			LinkedList<StrukturElement> liste = new LinkedList<StrukturElement>();
			liste.add(element);

			_debug.fine("\t" + _zaehler + " Element in Nachfolge: "
					+ element.getPid());
			_zaehler++;

			if (element.getNachfolger().size() > 0)
			{
				LinkedList<StrukturElement> nach = bestimmeNachfolgerRekursivFuerAusgabe(element
						.getNachfolger().get(0));
				if (nach != null)
				{
					liste.addAll(nach);
				}
			}

			return liste;
		}

		public String toString()
		{
			String re = "";
			Iterator<Entry<String, StrukturElement>> it = _elemente.entrySet()
					.iterator();
			while (it.hasNext())
			{
				Entry<String, StrukturElement> entry = it.next();
				re = re + "PID: " + entry.getKey() + " Anz. Nachfolger: "
						+ entry.getValue().getNachfolger().size() + " Anz. Vorgaenger: "
						+ entry.getValue().getVorgaenger().size() + "\n";

			}

			return re;
		}

		/**
		 * Bestimmt die Liste der Nachfolger eines {@link StrukturElement}es. Dabei
		 * wird das {@link StrukturElement} an erster Stelle der Liste gesetzt und
		 * das letzte, nachfolgende Element an die letzte Stelle. <br>
		 * Der Verlauf der Liste ist linear und reprsentiert den Verlauf der
		 * nachfolgenden {@link StrukturElement}e auf der Strae.
		 * 
		 * @param element
		 *          {@link StrukturElement}, von dem die Liste der nachfolgenden
		 *          {@link StrukturElement}e bestimmt werden soll.
		 * @return Liste der nachfolgenden {@link StrukturElement}e mit dem
		 *         Ausgangs {@link StrukturElement} am Anfang und dem letzen,
		 *         nachfolgenden {@link StrukturElement} am Ende.
		 */
		public LinkedList<StrukturElement> bestimmeNachfolgerListe(
				StrukturElement element)
		{
			LinkedList<StrukturElement> liste = new LinkedList<StrukturElement>();
			LinkedList<StrukturElement> nachfolgerListe = bestimmeNachfolgerListeRekursiv(element);
			if ((nachfolgerListe != null) && (nachfolgerListe.size() > 0))
			{
				liste.addAll(nachfolgerListe);
			}

			return liste;
		}

		/**
		 * Bestimmt rekursiv die Nachfolger eines {@link StrukturElement}es.
		 * 
		 * @param element
		 *          {@link StrukturElement}, dessen Nachfolger bestimmt werden
		 *          sollen.
		 * @return Liste von Nachfolgern des {@link StrukturElement}es, das
		 *         bergebene Element steht an erster Stelle der Liste, der letzte
		 *         Nachfolger an letzter Stelle.
		 */
		private LinkedList<StrukturElement> bestimmeNachfolgerListeRekursiv(
				StrukturElement element)
		{
			LinkedList<StrukturElement> liste = new LinkedList<StrukturElement>();
			liste.add(element);

			if (element.getNachfolger().size() > 0)
			{
				LinkedList<StrukturElement> nach = bestimmeNachfolgerListeRekursiv(element
						.getNachfolger().get(0));
				if (nach != null)
				{
					liste.addAll(nach);
				}
			}

			return liste;
		}

		/**
		 * Bestimmt die Liste der Vorgnger eines {@link StrukturElement}es. Dabei
		 * wird das {@link StrukturElement} an erster Stelle der Liste gesetzt und
		 * das letzte, vorgehende Element an die letzte Stelle. <br>
		 * Der Verlauf der Liste ist linear und reprsentiert den Verlauf der
		 * vorhergehenden {@link StrukturElement}e auf der Strae.
		 * 
		 * @param element
		 *          {@link StrukturElement}, vom dem die Liste der vorhergehenden
		 *          {@link StrukturElement}e bestimmt werden soll.
		 * @return Liste der vorhergehenden {@link StrukturElement}e mit dem
		 *         Ausgangs {@link StrukturElement} am Anfang und dem letzten
		 *         vorhergehenden {@link StrukturElement} am Ende.
		 */
		public LinkedList<StrukturElement> bestimmeVorgaengerListe(
				StrukturElement element)
		{
			LinkedList<StrukturElement> liste = new LinkedList<StrukturElement>();
			LinkedList<StrukturElement> vorgaengerListe = bestimmeVorgaengerListeRekursiv(element);
			if ((vorgaengerListe != null) && (vorgaengerListe.size() > 0))
			{
				liste.addAll(vorgaengerListe);
			}

			return liste;
		}

		/**
		 * Bestimmt rekursiv die Vorgnger eines {@link StrukturElement}es.
		 * 
		 * @param element
		 *          {@link StrukturElement}, dessen Vorgnger bestimmt werden
		 *          sollen.
		 * @return Liste von Vorgngern des {@link StrukturElement}es, das
		 *         bergebene Element steht an erster Stelle der Liste, der letzte
		 *         Vorgnger an letzter Stelle.
		 */
		private LinkedList<StrukturElement> bestimmeVorgaengerListeRekursiv(
				StrukturElement element)
		{
			LinkedList<StrukturElement> liste = new LinkedList<StrukturElement>();
			liste.add(element);

			if (element.getVorgaenger().size() > 0)
			{
				LinkedList<StrukturElement> vorgaenger = bestimmeVorgaengerListeRekursiv(element
						.getVorgaenger().get(0));
				if (vorgaenger != null)
				{
					liste.addAll(vorgaenger);
				}
			}

			return liste;
		}

	}

	/**
	 * Reprsentiert ein Element in der Struktur. Dieses verfgt ber je eine
	 * Liste von Vorgngern und Nachfolgern und kann ein
	 * {@link AeusseresStrassenSegment} oder einen {@link StrassenKnoten} als Wert
	 * aufnehmen.
	 * 
	 * @author HofmannJ
	 * 
	 */
	public class StrukturElement
	{
		private Object _element = null;

		/**
		 * TMC-Richtung dieses StrukturElements
		 */
		private String _tmcRichtung = null;

		private LinkedList<StrukturElement> _vorgaenger = new LinkedList<StrukturElement>();

		private LinkedList<StrukturElement> _nachfolger = new LinkedList<StrukturElement>();

		/**
		 * Konstruktor, bernimmt ein {@link AeusseresStrassenSegment} als Element
		 * der Klasse {@link StrukturElement}.
		 * 
		 * @param element
		 *          {@link AeusseresStrassenSegment}
		 * 
		 * @param tmcRichtung
		 *          TmcRichtung dieses {@link StrukturElement}es.
		 */
		public StrukturElement(AeusseresStrassenSegment element, String tmcRichtung)
		{
			this._element = element;
			this._tmcRichtung = tmcRichtung;
		}

		/**
		 * Konstruktor, bernimmt einen {@link StrassenKnoten} als Element der
		 * Klasse {@link StrukturElement}.
		 * 
		 * @param knoten
		 *          {@link StrassenKnoten}
		 * @param tmcRichtung
		 *          TmcRichtung dieses {@link StrukturElement}es.
		 */
		public StrukturElement(StrassenKnoten knoten, String tmcRichtung)
		{
			this._element = knoten;
			this._tmcRichtung = tmcRichtung;
		}

		/**
		 * Liefert das Element dieser Datenkapselungsklasse als {@link Object}
		 * zurck.
		 * 
		 * @return {@link Object}
		 */
		public Object getElement()
		{
			return _element;
		}

		/**
		 * Prft, ob es sich beim Element um einen {@link StrassenKnoten} handelt.
		 * 
		 * @return Liefert <code>true</code>, wenn es sich um einen
		 *         {@link StrassenKnoten} handelt, sonst <code>false</code>.
		 */
		public boolean isStrassenKnoten()
		{
			if (_element instanceof StrassenKnoten)
			{
				return true;
			} else
			{
				return false;
			}
		}

		/**
		 * Prft, ob es sich beim Element um ein {@link AeusseresStrassenSegment}
		 * handelt.
		 * 
		 * @return Liefert <code>true</code>, wenn es sich um ein
		 *         {@link AeusseresStrassenSegment} handelt, sonst
		 *         <code>false</code>.
		 */
		public boolean isAeusseresStrassenSegment()
		{
			if (_element instanceof AeusseresStrassenSegment)
			{
				return true;
			} else
			{
				return false;
			}
		}

		/**
		 * Liefert die Pid des Elementes
		 * 
		 * @return Pid Liefert die PID des Elementes (entweder die PID des
		 *         {@link StrassenKnoten}s oder des
		 *         {@link AeusseresStrassenSegment}s, sonst <code>null</code>).
		 */
		public String getPid()
		{
			if (isAeusseresStrassenSegment())
			{
				AeusseresStrassenSegment as = (AeusseresStrassenSegment) _element;
				return as.getPid();
			} else if (isStrassenKnoten())
			{
				StrassenKnoten sk = (StrassenKnoten) _element;
				return sk.getPid();
			} else
			{
				return null;
			}
		}

		/**
		 * Liefert die TmcRichtung dieses {@link StrukturElement}es.
		 * 
		 * @return TmcRichtung dieses {@link StrukturElement}es.
		 */
		public String getTmcRichtung()
		{
			return _tmcRichtung;
		}

		/**
		 * Fgt einen Vorgnger hinzu, falls dieser noch nicht i.d. Liste der
		 * Vorgnger vorhanden ist.
		 * 
		 * @param vorgaenger
		 *          Vorgnger
		 */
		public void addVorgaenger(StrukturElement vorgaenger)
		{
			if (!_vorgaenger.contains(vorgaenger))
			{
				_vorgaenger.add(vorgaenger);
			}
		}

		/**
		 * Liefert die Liste der Vorgnger.
		 * 
		 * @return Liste der Vorgnger
		 */
		public LinkedList<StrukturElement> getVorgaenger()
		{
			return _vorgaenger;
		}

		/**
		 * Fgt einen Nachfolger vom Typ {@link StrukturElement} zur Liste der
		 * Nachfolger hinzu.
		 * 
		 * @param nachfolger
		 *          {@link StrukturElement} als Nachfolger
		 */
		public void addNachfolger(StrukturElement nachfolger)
		{
			if (!_nachfolger.contains(nachfolger))
			{
				_nachfolger.add(nachfolger);
			}
		}

		/**
		 * Liefert die Liste der Nachfolger.
		 * 
		 * @return Liste der Nachfolger.
		 */
		public LinkedList<StrukturElement> getNachfolger()
		{
			return _nachfolger;
		}

		public String toString()
		{
			String re = "";

			if (isStrassenKnoten())
			{
				StrassenKnoten sk = (StrassenKnoten) _element;
				re = sk.toString();
			} else if (isAeusseresStrassenSegment())
			{
				AeusseresStrassenSegment aeSS = (AeusseresStrassenSegment) _element;
				re = aeSS.toString();
			}

			if (_tmcRichtung != null)
			{
				re = re + " (TmcRichtung: " + _tmcRichtung + ")";
			}

			return re;
		}

		public String toStringLang()
		{
			String re = "" + toString() + "\n";

			String vor = "Vorgnger:  ";
			Iterator<StrukturElement> itVor = _vorgaenger.iterator();
			while (itVor.hasNext())
			{
				StrukturElement element = itVor.next();
				vor = vor + element.toString() + " ";
			}

			String nach = "Nachfolger: ";
			Iterator<StrukturElement> itNach = _nachfolger.iterator();
			while (itNach.hasNext())
			{
				StrukturElement element = itNach.next();
				nach = nach + element.toString() + " ";
			}

			re = re + vor + "\n" + nach + "\n";

			return re;
		}

	}

	/**
	 * Diese Klasse kapselt ein {@link StrukturElement} und beinhaltet einen
	 * kompletten Pfad von {@link StrukturElement}en bis zu dem hier gekapselten
	 * {@link StrukturElement}. Das erste Element in diesem Pfad gibt das
	 * Ausgangs {@link StrukturElement} an.
	 * 
	 * @author HofmannJ
	 * 
	 */
	public class StrukturElementMitPfad
	{
		/**
		 * Speichert das {@link StrukturElement}.
		 */
		private StrukturElement _strukturElement = null;

		/**
		 * Mit Hilfe dieses Flags wird angezeigt, dass der Verlauf im Pfad in
		 * positiver Richtung, d.h. das erste Element ist der Vorgnger vom 2.ten
		 * ist der Vorgnger vom 3. ten usw, erfolgt <br>
		 * 
		 * Folgende Struktur sei ein Beispiel fr die aus {@link Struktur}
		 * gelieferte Grundstruktur. <br>
		 * X---a---X---b---X---c---X---d---X <br>
		 * <ul>
		 * <li> Soll jetzt ein Pfad von b nach dem Letzten Knoten X gebildet werden,
		 * ist die Richtung positiv un die Gesamtlnge von b geht i.d. Berechnung
		 * mit ein </li>
		 * <li> Soll ein Pfad von c zum ersten Knoten X gebildet werden ist davon
		 * auszugehen, dass wir uns am Anfang von c befinden und c somit nicht mit
		 * seiner vollen Lnge i.d. Berechnung mit ein geht. Die Richtung ist im
		 * Vergleich zu der von {@link Struktur} gelieferten Richtung negativ. <br>
		 * In dieser Situation muss vom Anfang des {@link AeusseresStrassenSegment}
		 * ausgegangen werden. </li>
		 * <li> 3.tes X nach a: Richtung negativ, A geht mit voller Lnge ein </li>
		 * <li> 2.tes X nach c: Richtung positiv, C geht mit voller Lnge ein </li>
		 * </ul>
		 * 
		 * <b>Zusammenfassung</b>: Es wird immer vom Angang eines
		 * {@link AeusseresStrassenSegment}es ausgegangen. In <i>positiver</i>
		 * Richtung geht dieses komplett mit ein, in <i>negativer</i> Richtung
		 * nicht (d.h. man geht nur vom Anfang dieses
		 * {@link AeusseresStrassenSegment} aus).
		 * 
		 */
		private boolean _richtungPositiv = true;

		/**
		 * Speichert den Pfad. Das erste Element ist das Ausgangsstrukturelement,
		 * das letzte Element ist das Zielelement, entsprechend dem in
		 * {@link StrukturElementMitPfad#_strukturElement} gespeichertem Element.
		 */
		private LinkedList<StrukturElement> _pfad = null;

		/**
		 * Konstruktor. bernimmt ein {@link StrukturElement}.
		 * 
		 * @param strukturElement
		 *          Zu bernehmendes {@link StrukturElement}
		 * @param richtungPositiv
		 *          Gibt an, ob im Vergleich zur Klasse {@link Struktur}, aus der
		 *          der Pfad gebildet wurde, der Pfad in positiver Richtung
		 *          (Vorgnger im Pfad ist auch Vorgnger in {@link Struktur}) oder
		 *          in negativer Richtung (Vorgnger im Pfad ist Nachfolger in
		 *          {@link Struktur}) verluft.
		 * 
		 */
		public StrukturElementMitPfad(StrukturElement strukturElement,
				boolean richtungPositiv)
		{
			this._strukturElement = strukturElement;
			this._richtungPositiv = richtungPositiv;
		}

		/**
		 * Konstruktor. bernimmt ein {@link StrukturElement}, die Richtung sowie
		 * einen Pfad, auf dem das {@link StrukturElement} zu erreichen ist.
		 * 
		 * @param strukturElement
		 *          Zu bernehmendes {@link StrukturElement}
		 * @param richtungPositiv
		 *          Gibt an, ob im Vergleich zur Klasse {@link Struktur}, aus der
		 *          der Pfad gebildet wurde, der Pfad in positiver Richtung
		 *          (Vorgnger im Pfad ist auch Vorgnger in {@link Struktur}) oder
		 *          in negativer Richtung (Vorgnger im Pfad ist Nachfolger in
		 *          {@link Struktur}) verluft.
		 * @param pfad
		 *          Pfad, auf dem das {@link StrukturElement} zu erreichen ist.
		 * @throws IllegalArgumentException
		 *           Wird geworfen, wenn es sich beim letzten Element im Pfad nicht
		 *           um das {@link StrukturElement} handelt.
		 * 
		 */
		public StrukturElementMitPfad(StrukturElement strukturElement,
				boolean richtungPositiv, LinkedList<StrukturElement> pfad)
		{
			this._strukturElement = strukturElement;
			this._richtungPositiv = richtungPositiv;
			setPfad(pfad);
		}

		/**
		 * Liefert das hinterlegte {@link StrukturElement} zurck.
		 * 
		 * @return Liefert das hinterlegte {@link StrukturElement} zurck.
		 */
		public StrukturElement getStrukturElement()
		{
			return _strukturElement;
		}

		/**
		 * Liefert die TmcRichtung des hinterlegten {@link StrukturElement}es
		 * zurck.
		 * 
		 * @return TmcRichtung des hinterlegten {@link StrukturElement}es zurck
		 */
		public String getStrukturElementTmcRichtung()
		{
			return _strukturElement.getTmcRichtung();
		}

		/**
		 * Liefert die PID des in dieser Klasse hinterlegten {@link StrukturElement}es
		 * zurck.
		 * 
		 * @return PID des hinterlegten {@link StrukturElement}es
		 */
		public String getPid()
		{
			String re = null;

			if (_strukturElement != null)
			{
				re = _strukturElement.getPid();
			}

			return re;
		}

		/**
		 * Liefert die PID des in dieser Klasse hinterlegten {@link StrukturElement}es,
		 * erweitert um die TmcRichtung, als Schlssel zurck.
		 * 
		 * @return
		 */
		public String getKey()
		{
			String re = null;
			if (_strukturElement != null)
			{
				re = getPid();

				if (_richtungPositiv)
				{
					re = re + " (TmcRichtung: positiv)";
				} else
				{
					re = re + " (TmcRichtung: negativ)";
				}
			}

			return re;
		}

		/**
		 * Liefert den Pfad, auf dem das {@link StrukturElement} zu erreichen ist.
		 * Beim letzten Element in der Liste handelt es sich um das in
		 * {@link StrukturElementMitPfad} hinterlegte {@link StrukturElement}, beim
		 * ersten Element um das Ausgangselement.
		 * 
		 * @return Pfad, auf dem das {@link StrukturElement} zu erreichen ist.
		 */
		public LinkedList<StrukturElement> getPfad()
		{
			return _pfad;
		}

		/**
		 * Setzt den Pfad, auf dem das {@link StrukturElement} zu erreichen ist.
		 * Beim letzten Element in der Liste handelt es sich um das in
		 * {@link StrukturElementMitPfad} hinterlegte {@link StrukturElement}, beim
		 * ersten Element um das Ausgangselement.
		 * 
		 * @param pfad
		 *          Pfad, auf dem das {@link StrukturElement} zu erreichen ist.
		 * 
		 * @throws IllegalArgumentException
		 *           Wird geworfen, wenn es sich beim letzten Element im Pfad nicht
		 *           um das in dieser Klasse bereits hinterlegte
		 *           {@link StrukturElement} handelt.
		 */
		public void setPfad(LinkedList<StrukturElement> pfad)
				throws IllegalArgumentException
		{
			if (pfad.getLast() != _strukturElement)
			{
				throw new IllegalArgumentException(
						"Das letzte Element im Pfad entspricht nicht dem hinterlegten StrukturElement!");
			} else
			{
				this._pfad = pfad;
			}

		}

		/**
		 * Prfung auf Lnge
		 * 
		 * @param stationierung
		 * @return
		 */
		public boolean pruefe(long stationierung)
		{
			boolean re = true;

			/*
			 * Richtungsbereinigung positiv/negativ
			 */
			LinkedList<StrukturElement> bereinigt = new LinkedList<StrukturElement>();
			for (int i = 0; i < _pfad.size(); i++)
			{
				StrukturElement strukturElement = _pfad.get(i);

				if ((i == 0) && (!_richtungPositiv)
						&& (strukturElement.isAeusseresStrassenSegment()))
				{
					// Nicht bernehmen
				} else
				{
					if (strukturElement.isAeusseresStrassenSegment())
					{
						bereinigt.add(strukturElement);
					}
				}

			}

			return re;
		}

		public String toString()
		{
			String re = "" + _strukturElement.toString() + " ";
			re = re + " Anzahl Elemente im Pfad: " + _pfad.size();
			return re;
		}

		/**
		 * Lange Version der <code>toString</code>-Methode mit kompletter
		 * Pfadausgabe.
		 * 
		 * @return Lang-Version von <code>toString</code>.
		 */
		public String toStringLang()
		{
			String re = toString();

			if (_pfad.size() > 0)
			{
				re = re + "[";
				Iterator<StrukturElement> it = _pfad.iterator();
				while (it.hasNext())
				{
					StrukturElement element = it.next();
					re = re + element.getPid() + "; ";
				}

				re = re.substring(0, re.length() - 2);
				re = re + "]";
			}

			return re;
		}

	}

	/**
	 * Liefert eine dem Namen nach sortierte Liste der {@link Strasse}n zurck.
	 * 
	 * @return Sortierte Liste der {@link Strasse}n.
	 */
	public LinkedList<Strasse> getStrassen()
	{

		HashMap<String, Strasse> strassen = _konf.getStrassen();

		TreeMap<String, Strasse> strassenSortiert = new TreeMap<String, Strasse>();

		strassenSortiert.putAll(strassen);

		LinkedList<Strasse> listeStrassen = new LinkedList<Strasse>();
		Iterator<Entry<String, Strasse>> it = strassenSortiert.entrySet()
				.iterator();
		while (it.hasNext())
		{
			Entry<String, Strasse> entry = it.next();

			if (!entry.getValue().getName().equals("null"))
			{
				listeStrassen.add(entry.getValue());
			}

		}

		return listeStrassen;
	}

	/**
	 * Liefert ein StrukturElement an Hand seiner PID zurck.
	 * 
	 * @param pid
	 *          PID des {@link StrukturElement}es
	 * @param tmcRichtung
	 *          TmcRichtung des {@link StrukturElement}es (<code>positiv</code>
	 *          oder <code>negativ</code>)
	 * @return {@link StrukturElement}
	 */
	public StrukturElement getStrukturElement(String pid, String tmcRichtung)
	{
		StrukturElement strukturElement = null;

		if (_strukturPos.containsKey(pid))
		{
			StrukturElement strukturElementPos = _strukturPos.getStrukturElement(pid);
			if (strukturElementPos.getTmcRichtung().equals(tmcRichtung))
			{
				strukturElement = strukturElementPos;
			}
		}

		if ((strukturElement == null) && (_strukturNeg.containsKey(pid)))
		{
			StrukturElement strukturElementNeg = _strukturNeg.getStrukturElement(pid);
			if (strukturElementNeg.getTmcRichtung().equals(tmcRichtung))
			{
				strukturElement = strukturElementNeg;
			}
		}

		return strukturElement;
	}

	private class AsbKnotenVerwaltung
	{
		/**
		 * TMC-Richtung
		 */
		private EnumTmcRichtung _tmcRichtung = null;

		/**
		 * Liste der {@link AsbKnotenAnfangEnde}
		 */
		private LinkedList<AsbKnotenAnfangEnde> _asbKnoten = new LinkedList<AsbKnotenAnfangEnde>();

		/**
		 * Konstruktor, bernimmt die {@link EnumTmcRichtung} als Parameter.
		 * 
		 * @param tmcRichtung
		 *          {@link EnumTmcRichtung}
		 */
		public AsbKnotenVerwaltung(EnumTmcRichtung tmcRichtung)
		{
			this._tmcRichtung = tmcRichtung;
		}

		/**
		 * Fgt eine {@link AsbStationierung} zur {@link AsbKnotenVerwaltung} hinzu.
		 * Existiert diese Stationierung nicht, wird diese als neuer
		 * {@link AsbKnotenAnfangEnde} hinzugefgt. Dieser beinhaltet den Anfangs
		 * und den Endknoten, die {@link AsbStationierung} wird in einer Liste von
		 * AsbStationierungsBereichen verwaltet. Wird erneut eine
		 * {@link AsbStationierung} mit gleichem Anfangs- und Endknoten hinzugefgt,
		 * so wird diese nur noch in der Bereichsliste des bereits existierenden
		 * {@link AsbKnotenAnfangEnde} hinzugefgt.
		 * 
		 * @param asbStationierung
		 *          Hinzuzufgende {@link AsbStationierung}.
		 */
		public void addAsbStationierung(AsbStationierung asbStationierung)
		{
			AsbKnotenAnfangEnde asbKnotenBereich = null;

			/*
			 * Prfen, ob eine ASB-Stationierung mit gleichen Anfangs und Endknoten
			 * existiert
			 */
			String anfangsKnoten = asbStationierung.getAnfangsKnoten();
			String endKnoten = asbStationierung.getEndKnoten();

			Iterator<AsbKnotenAnfangEnde> it = _asbKnoten.iterator();
			while (it.hasNext())
			{
				AsbKnotenAnfangEnde pruefKnotenBereich = it.next();
				if ((pruefKnotenBereich.getAnfangsKnoten().getKnoten()
						.equals(anfangsKnoten))
						&& (pruefKnotenBereich.getEndKnoten().getKnoten().equals(endKnoten)))
				{
					asbKnotenBereich = pruefKnotenBereich;
					break;
				}
			}

			/*
			 * Wenn asbKnotenBereich == null, dann neu anlegen und zur Liste der
			 * AsbKnotenBereiche hinzufgen
			 */
			if (asbKnotenBereich == null)
			{
				asbKnotenBereich = new AsbKnotenAnfangEnde(asbStationierung,
						_tmcRichtung);
				_asbKnoten.add(asbKnotenBereich);
			}

			/*
			 * Neuen Bereich hinzufgen
			 */
			asbKnotenBereich.addBereich(asbStationierung);
		}

		/**
		 * Bestimmt eine kombinierte Liste aller Knoten. Diese wird aus Anfangs- und
		 * EndKnoten erstellt. Jeder Knoten kommt, egal ob Anfangs- oder Endknoten,
		 * nur ein mal in der Liste vor.
		 * 
		 * @return Kombinierte Liste aller Knoten.
		 */
		public LinkedList<AsbKnoten> bestimmeListeAllerKnoten()
		{
			LinkedList<AsbKnoten> re = new LinkedList<AsbKnoten>(); // Von String auf
			// AsbKnoten
			// umgestellt

			Iterator<AsbKnotenAnfangEnde> it = _asbKnoten.iterator();
			while (it.hasNext())
			{
				AsbKnotenAnfangEnde asbKnotenAnfangEnde = it.next();

				/*
				 * Ist der AnfangsKnoten schon vorhanden? Wenn nein, hinzufgen.
				 */
				if (!istKnotenAsbInListe(re, asbKnotenAnfangEnde.getAnfangsKnoten()))
				{
					re.add(asbKnotenAnfangEnde.getAnfangsKnoten());
				}

				/*
				 * Ist der EndKnoten schon vorhanden? Wenn nein, hinzufgen.
				 */
				// if (!re.contains(asbKnotenAnfangEnde.getEndKnoten()))
				if (!istKnotenAsbInListe(re, asbKnotenAnfangEnde.getEndKnoten()))
				{
					re.add(asbKnotenAnfangEnde.getEndKnoten());
				}
			}

			// Collections.sort(re, Collator.getInstance());

			Collections.sort(re);

			return re;
		}

		/**
		 * Bestimmt eine Liste aller von einem Knoten aus erreichbaren Knoten.
		 * 
		 * @return Liste aller von einem Knoten aus erreichbaren Knoten
		 */
		public LinkedList<AsbKnoten> bestimmeListeAllerErreichbarenKnoten(
				AsbKnoten ausgangsKnoten)
		{
			LinkedList<AsbKnoten> re = new LinkedList<AsbKnoten>(); // Von String auf
			// AsbKnoten
			// umgestellt

			Iterator<AsbKnotenAnfangEnde> it = _asbKnoten.iterator();
			while (it.hasNext())
			{
				AsbKnotenAnfangEnde knotenAnfangEnde = it.next();

				/*
				 * Anfangsknoten == Ausgangsknoten?
				 */
				if (knotenAnfangEnde.getAnfangsKnoten().equals(ausgangsKnoten))
				{
					// if (!re.contains(knotenAnfangEnde.getEndKnoten()))
					if (!istKnotenAsbInListe(re, knotenAnfangEnde.getEndKnoten()))
					{
						re.add(knotenAnfangEnde.getEndKnoten());
					}
				}

				/*
				 * Endknoten == Ausgabgsknoten?
				 */
				if (knotenAnfangEnde.getEndKnoten().equals(ausgangsKnoten))
				{
					// if (!re.contains(knotenAnfangEnde.getAnfangsKnoten()))
					if (!istKnotenAsbInListe(re, knotenAnfangEnde.getAnfangsKnoten()))
					{
						re.add(knotenAnfangEnde.getAnfangsKnoten());
					}
				}
			}

			// Collections.sort(re, Collator.getInstance());
			Collections.sort(re);

			return re;
		}

		/**
		 * Prft, ob ein {@link AsbKnoten} in einer {@link Collection} von
		 * {@link AsbKnoten} vorhanden ist.
		 * 
		 * @param collection
		 *          {@link Collection}, die zu berprfen ist
		 * @param knoten
		 *          {@link AsbKnoten}, auf dessen vorhanden sein i.d.
		 *          {@link Collection} zu prfen ist.
		 * @return Liefert <code>true</code>, wenn der Knoten in der
		 *         {@link Collection} vorhanden ist, sonst <code>false</code>.
		 */
		private boolean istKnotenAsbInListe(Collection<AsbKnoten> collection,
				AsbKnoten knoten)
		{
			boolean re = false;
			Iterator<AsbKnoten> it = collection.iterator();
			while (it.hasNext())
			{
				AsbKnoten pruefKnoten = it.next();
				if ((pruefKnoten.getKnoten().equals(knoten.getKnoten()))
						&& (pruefKnoten.getTmcRichtung().equals(knoten.getTmcRichtung())))
				{
					re = true;
					break;
				}
			}

			return re;
		}

		public boolean istInBereich(AsbKnoten knotenA, AsbKnoten knotenB,
				long stationierung)
		{
			boolean re = false;

			/*
			 * Knoten, der durch den Angegebenen Bereich bestimmt ist.
			 */
			AsbKnotenAnfangEnde node = null;

			/*
			 * Vorwrts-Annahme: AnfangsKnoten == knotenA, EndKnoten == knotenB
			 */
			if (node == null)
			{
				Iterator<AsbKnotenAnfangEnde> itVor = _asbKnoten.iterator();
				while (itVor.hasNext())
				{
					AsbKnotenAnfangEnde pruefKnoten = itVor.next();
					if ((pruefKnoten.getAnfangsKnoten().equals(knotenA))
							&& (pruefKnoten.getEndKnoten().equals(knotenB)))
					{
						node = pruefKnoten;
						break;
					}
				}
			}

			/*
			 * Rckwrts-Annahme: AnfangsKnoten == knotenB, EndKnoten == knotenA
			 */
			if (node == null)
			{
				Iterator<AsbKnotenAnfangEnde> it = _asbKnoten.iterator();
				while (it.hasNext())
				{
					AsbKnotenAnfangEnde pruefknoten = it.next();
					if ((pruefknoten.getEndKnoten().equals(knotenA))
							&& (pruefknoten.getAnfangsKnoten().equals(knotenB)))
					{
						node = pruefknoten;
						break;
					}
				}
			}

			/*
			 * Bereichsprfung durchfhren.
			 */
			if (node != null)
			{
				re = node.istInBereichen(stationierung);
			}

			return re;
		}

		public String toString()
		{
			String re = "";
			Iterator<AsbKnotenAnfangEnde> it = _asbKnoten.iterator();
			int i = 0;
			while (it.hasNext())
			{
				AsbKnotenAnfangEnde asbKnotenAnfangEnde = it.next();
				re = re + "#" + i + ":\n" + asbKnotenAnfangEnde + "\n";
				i++;
			}
			return re;
		}
	}

	/**
	 * Klasse zur Reprsentation der vorhandenen Stationierungsknoten
	 * 
	 * @author HofmannJ
	 * 
	 */
	private class AsbKnotenAnfangEnde
	{
		/**
		 * AnfangsKnoten
		 */
		private AsbKnoten _anfangsKnoten = null;

		/**
		 * EndKnoten
		 */
		private AsbKnoten _endKnoten = null;

		private LinkedList<AsbStationierung> _asbStationierungsBereiche = new LinkedList<AsbStationierung>();

		/**
		 * Konstruktor, bernimmt Anfangs- und EndKnoten.
		 * 
		 * @param anfangsKnoten
		 *          AnfangsKnoten
		 * @param endKnoten
		 *          EndKnoten
		 */
		public AsbKnotenAnfangEnde(AsbStationierung asbStationierung,
				EnumTmcRichtung tmcRichtung)
		{

			this._anfangsKnoten = new AsbKnoten(asbStationierung.getAnfangsKnoten(),
					tmcRichtung, asbStationierung.getVerkehrsRichtung());
			this._endKnoten = new AsbKnoten(asbStationierung.getEndKnoten(),
					tmcRichtung, asbStationierung.getVerkehrsRichtung());
			addBereich(asbStationierung);
		}

		/**
		 * Fgt einen {@link AsbStationierung}s-Bereich (Anfang, Ende) hinzu,
		 * sofern dieser noch nicht vorhanden ist. Vergleiche ATG AsbStationiereung.
		 * An Hand diesem kann dann bestimmt werden, ob eine Stationierung im
		 * gltigen Bereich liegt.
		 * 
		 * @param Hinzuzufgender
		 *          {@link AsbStationierung}s-Bereich.
		 */
		public void addBereich(AsbStationierung bereich)
		{

			if (!_asbStationierungsBereiche.contains(bereich))
			{
				_asbStationierungsBereiche.add(bereich);
			}
		}

		/**
		 * Prft, ob eine gegebene Stationierungsangabe in Metern im Bereich liegt.<br>
		 * 
		 * @param stationierungsangabe
		 *          Stationierungsangabe in Metern
		 * @return Liefert <code>true</code>, wenn die Stationierungsangabe in
		 *         einem der Bereiche liegt, sonst <code>false</code>.
		 */
		public boolean istInBereichen(long stationierungsangabe)
		{
			boolean re = false;

			Iterator<AsbStationierung> it = _asbStationierungsBereiche.iterator();
			while (it.hasNext())
			{
				AsbStationierung bereich = it.next();
				long anfang = bereich.getAnfang();
				long ende = bereich.getEnde();

				/*
				 * Liegt die Stationierungsangabe genau auf dem Angang oder dem Ende
				 * eines Bereichs?
				 */

				if ((stationierungsangabe == anfang) || (stationierungsangabe == ende))
				{
					re = true;
					break;
				}

				/*
				 * Anfang < Ende: Ist die Stationierungsangabe zwischen den Bereichen?
				 */
				if (anfang < ende)
				{
					if ((stationierungsangabe >= anfang)
							&& (stationierungsangabe <= ende))
					{
						re = true;
						break;
					}
				}

				/*
				 * Anfang > Ende: Ist die Stationierungsangabe zwischen den Bereichen?
				 */
				if (anfang > ende)
				{
					if ((stationierungsangabe <= anfang)
							&& (stationierungsangabe >= ende))
					{
						re = true;
						break;
					}
				}

			}

			return re;
		}

		/**
		 * Liefert den AnfangsKnoten.
		 * 
		 * @return AnfangsKnoten
		 */
		public AsbKnoten getAnfangsKnoten()
		{
			return _anfangsKnoten;
		}

		/**
		 * Liefert den EndKnoten.
		 * 
		 * @return EndKnoten
		 */
		public AsbKnoten getEndKnoten()
		{
			return _endKnoten;
		}

		public String toString()
		{
			String re = "";
			re = "AsbKnotenAnfangEnde: AnfangsKnoten=" + _anfangsKnoten
					+ " EndKnoten=" + _endKnoten + "\n";

			Iterator<AsbStationierung> it = _asbStationierungsBereiche.iterator();
			int i = 0;
			while (it.hasNext())
			{
				AsbStationierung asbStationierung = it.next();
				long anfang = asbStationierung.getAnfang();
				long ende = asbStationierung.getEnde();
				re = re + "Eintrag #" + i + ": Anfang=" + anfang + "m, Ende=" + ende
						+ "m. (Sts:"
						+ asbStationierung.getGehoertZuStrassenTeilSegement().getPid()
						+ ")\n";
				i++;
			}

			return re;
		}

	}

	/**
	 * Klasse zur Erfassung und Darstellung von ASB-Knoten. Erfasst wird die
	 * Kennung des Knotens sowie, in welcher TMC-Richtung dieser Knoten liegt.
	 * 
	 * @author HofmannJ
	 * 
	 */
	public class AsbKnoten implements Comparable<AsbKnoten>
	{
		/**
		 * Kennung des Knotens
		 */
		private String _knoten = null;

		/**
		 * TMC-Richtung des Knotens
		 */
		private EnumTmcRichtung _tmcRichtung = null;

		/**
		 * Speichert die ASB Verkehrsrichtung
		 */
		private EnumAsbVerkehrsRichtung _asbVerkehrsRichtung = null;

		/**
		 * Konstruktor, bernimmt die Kennung sowie die {@link EnumTmcRichtung} des
		 * Knotens.
		 * 
		 * @param knoten
		 *          Kennung des Knotens
		 * @param tmcRichtung
		 *          {@link EnumTmcRichtung} des Knotens
		 * @param asbVerkehrsRichtung
		 *          {@link EnumAsbVerkehrsRichtung}
		 */
		public AsbKnoten(String knoten, EnumTmcRichtung tmcRichtung,
				EnumAsbVerkehrsRichtung asbVerkehrsRichtung)
		{
			this._knoten = knoten;
			this._tmcRichtung = tmcRichtung;
			this._asbVerkehrsRichtung = asbVerkehrsRichtung;
		}

		/**
		 * Liefert die Kennung des Knotens.
		 * 
		 * @return Knotenkennung.
		 */
		public String getKnoten()
		{
			return _knoten;
		}

		/**
		 * Liefert die {@link EnumTmcRichtung} des Knotens.
		 * 
		 * @return {@link EnumTmcRichtung} des Knotens
		 */
		public EnumTmcRichtung getTmcRichtung()
		{
			return _tmcRichtung;
		}

		/**
		 * Liefert die {@link EnumAsbVerkehrsRichtung}
		 * 
		 * @return {@link EnumAsbVerkehrsRichtung}.
		 */
		public EnumAsbVerkehrsRichtung getAsbVerkehrsRichtung()
		{
			return _asbVerkehrsRichtung;
		}

		public boolean equals(Object obj)
		{
			boolean re = false;

			if (obj instanceof AsbKnoten)
			{
				AsbKnoten vergleich = (AsbKnoten) obj;

				if ((this.getKnoten().equals(vergleich.getKnoten()))
						&& (this.getTmcRichtung().equals(vergleich.getTmcRichtung())))
				{
					re = true;
				}

			}

			return re;
		}

		public String toString()
		{
//			String re = _knoten + " (TmcRichtung: " + _tmcRichtung + ")";
			String re = _knoten + " (AsbVerkehrsRichtung: " + _asbVerkehrsRichtung + ")";
			return re;
		}

		@Override
		public int compareTo(AsbKnoten o)
		{
			String aktuell = this.getKnoten();
			String vergleich = o.getKnoten();

			return aktuell.compareTo(vergleich);
		}

	}

	/**
	 * Enumerator fr die Richtung.
	 * 
	 * @author HofmannJ
	 * 
	 */
	public enum EnumTmcRichtung
	{
		/**
		 * Positive TMC-Richtung
		 */
		positiv,

		/**
		 * Negative TMC-Richtung
		 */
		negativ
	}

}
