/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.rw.dav;

import java.util.Iterator;
import java.util.LinkedList;

import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Diese Klasse stellt zur Laufzeit der RCP-Applikation die Daten des aktuell
 * angemeldeten Benutzers zur Verfgung. Damit zur Laufzeit der RCP-Anwendung
 * nur eine Instanz dieser Klasse existiert, ist sie als Singleton
 * implementiert.<br>
 * <br>
 * Desweitern knnen sich Zuhrer vom Typ {@link IBenutzerZuhoerer} registrieren
 * und werden dann darber Informiert, dass sich ein neuer Benutzer angemeldet
 * hat.
 * 
 * @author HofmannJ
 * 
 */
public class BenutzerAngemeldet
{

	/**
	 * Der Debugger
	 */
	private Debug _debug = Debug.getLogger();

	/**
	 * Benutzername
	 */
	private String _name = null;

	/**
	 * Passwort
	 */
	private String _passwort = null;

	/**
	 * Adresse des jeweiligen DaVs
	 */
	private String _ip = null;

	/**
	 * Port
	 */
	private int _tcpPort;

	/**
	 * Liste der {@link IBenutzerZuhoerer}, die ber einen neue angemeldeten
	 * Benuter informiert werden mchten.
	 */
	private LinkedList<IBenutzerZuhoerer> _benutzerZuhoerer = new LinkedList<IBenutzerZuhoerer>();

	/**
	 * Privater Konstruktor dieser Singleton-Klasse
	 * 
	 */
	private BenutzerAngemeldet()
	{
	}

	/**
	 * Speichert die Singleton-Instanz dieser Klasse.
	 * 
	 * @author HofmannJ
	 * 
	 */
	private static class Innen
	{
		private static BenutzerAngemeldet INSTANZ = new BenutzerAngemeldet();
	}

	/**
	 * Liefert die Singleton-Instanz dieser Klasse zurck.
	 * 
	 * @return Singleton-Instanz von {@link BenutzerAngemeldet}.
	 */
	public static BenutzerAngemeldet getInstanz()
	{
		return Innen.INSTANZ;
	}

	/*
	 * Setter und Getter
	 */

	/**
	 * Setzt den angemeltdeten Benutzer
	 * 
	 * @param name
	 *          Name des Benutzers
	 * @param passwort
	 *          Passwort des Benutzers
	 * @param ip
	 *          IP-Adresse des DaV
	 * @param port
	 *          Port-Nummer des DaV
	 */
	public void setAngemeldeterBenutzer(String name, String passwort, String ip,
			int port)
	{
		this._name = name;
		this._passwort = passwort;
		this._ip = ip;
		this._tcpPort = port;

		BenutzerOffline.getInstance().zuruecksetzen();

		benachrichtigeBenutzerZuhoerer();
	}

	/**
	 * Fgt einen {@link IBenutzerZuhoerer} zur Liste der Zuhrer hinzu.
	 * 
	 * @param benutzerZuhoerer
	 *          Hinzuzufgender {@link IBenutzerZuhoerer}.
	 */
	public void hinzufuegenBenutzerZuhoerer(IBenutzerZuhoerer benutzerZuhoerer)
	{
		if (!_benutzerZuhoerer.contains(benutzerZuhoerer))
		{
			_benutzerZuhoerer.add(benutzerZuhoerer);
		} else
		{
			_debug
					.error(getClass().getName()
							+ "::hinzufuegenBenutzerZuhoerer \n\t=>Der BenutzerZuhrer ist bereits in der Zuhrerliste vorhanden und wird nicht erneut hinzugefgt.");
		}
	}

	/**
	 * Entfernt einen {@link IBenutzerZuhoerer} von der Liste der Zuhrer.
	 * 
	 * @param benutzerZuhoerer
	 *          Zu entfernender {@link IBenutzerZuhoerer}.
	 */
	public void entferneBenutzerZuhoerer(IBenutzerZuhoerer benutzerZuhoerer)
	{
		if (_benutzerZuhoerer.contains(benutzerZuhoerer))
		{
			_benutzerZuhoerer.remove(benutzerZuhoerer);
		} else
		{
			_debug
					.error(getClass().getName()
							+ "::entferneBenutzerZuhoerer=>Der BenutzerZuhrer ist in der Zuhrerliste nicht vorhanden und wird somit nicht entfernt.");
		}
	}

	/**
	 * Benachrichtigt alle registrierten {@link IBenutzerZuhoerer} darber, dass
	 * ein neuer Benutzer sich angemeldet hat.
	 * 
	 */
	private void benachrichtigeBenutzerZuhoerer()
	{
		BenutzerEreignis benutzerEreignis = new BenutzerEreignis(_name);
		Iterator<IBenutzerZuhoerer> it = _benutzerZuhoerer.iterator();
		while (it.hasNext())
		{
			IBenutzerZuhoerer benutzerZuhoerer = it.next();
			benutzerZuhoerer.benutzerNeuAngemeldet(benutzerEreignis);
		}
	}

	/**
	 * Getter IP-Adresse
	 * 
	 * @return IP-Adresse als String
	 */
	public String getIp()
	{
		return _ip;
	}

	/**
	 * Setter IP-Adresse
	 * 
	 * @param m_ip
	 *          IP-Adresse als String
	 * @deprecated <code>setAngemeldeterBenutzer</code> verwenden.
	 */
	public void setIp(String m_ip)
	{
		this._ip = m_ip;
	}

	/**
	 * Getter Benutzername
	 * 
	 * @return Benutzername Liefert den Benutzernamen.
	 */
	public String getName()
	{
		// return _name;

		String name = null;

		if (_name != null)
		{
			name = _name;
		} else
		{
			name = BenutzerOffline.getInstance().getName();
		}

		return name;
	}

	/**
	 * Setter Benutzername
	 * 
	 * @param m_name
	 *          Setzt den Benutzernamen.
	 * 
	 * @deprecated <code>setAngemeldeterBenutzer</code> verwenden.
	 */
	public void setName(String m_name)
	{
		this._name = m_name;
	}

	/**
	 * Getter Passwort
	 * 
	 * @return Liefert das Passwort.
	 */
	public String getPasswort()
	{
		return _passwort;
	}

	/**
	 * Setter Passwort
	 * 
	 * @param m_passwort
	 *          Setzt das Passwort.
	 * 
	 * @deprecated <code>setAngemeldeterBenutzer</code> verwenden.
	 */
	public void setPasswort(String m_passwort)
	{
		this._passwort = m_passwort;
	}

	/**
	 * Getter IP-Adresse
	 * 
	 * @return Liefert die IP-Adresse-
	 * 
	 */
	public int getTcpPort()
	{
		return _tcpPort;
	}

	/**
	 * Setter IP-Adresse
	 * 
	 * @param port
	 *          Setzt IP-Adresse
	 * 
	 * @deprecated <code>setAngemeldeterBenutzer</code> verwenden.
	 */
	public void setTcpPort(int port)
	{
		_tcpPort = port;
	}

	/**
	 * Setzt alle Parameter zurck.
	 * 
	 */
	public void zuruecksetzen()
	{
		_name = null;
		_passwort = null;
		_ip = null;
		_tcpPort = -1;
	}
}
