/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.rw.menu.deskriptoren;

import org.eclipse.jface.action.IAction;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.SWT;

import de.bsvrz.buv.rw.rw.menu.ElementRegistrierung;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Diese Klasse liefert die "Beschreibung" zu durch Plug-Ins registrierten {@link IAction}.
 * 
 * @version $Revision: 1.2 $ / $Date: 2012/08/21 08:25:32 $ / ($Author: HofmannJ $)
 * 
 * @author HofmannJ
 * 
 */
public class DeskriptorAktion implements IDeskriptorElementUndAktion
{
  /**
   * Der Debugger
   */
  private Debug _debug = Debug.getLogger();

  /**
   * Gelieferte {@link IAction}.
   */
  private IAction _aktion = null;

  /**
   * ID des Plug-Ins, dem dieses Element zugeordnet wird.
   */
  private String _pluginID = null;

  /**
   * ID der Action, die von diesem {@link DeskriptorElement} stellvertretend beschrieben wird.
   */
  private String _aktionID = null;

  /**
   * Text der Action
   */
  private String _text = null;

  /**
   * ToolTip der Action
   */
  private String _toolTipText = null;

  /**
   * Beschreibung der Action.
   */
  private String _aktionBeschreibung = null;

  /**
   * Pfad zum Icon der Action
   */
  private String _iconPfad = null;

  /**
   * Pfad zum HoverIcon der Action
   */
  private String _hoverIconPfad = null;

  /**
   * Pfad zum DisabledIcon der Action
   */
  private String _deaktiviertesIconPfad = null;

  /**
   * Flag zum Bezeichnen der globalen Action
   */
  private boolean _globaleAktion = false;

  /**
   * Konstruktor, initialisiert ein Objekt vom Typ {@link DeskriptorAktion}. Kann fr den Aufbau von Men- und
   * Symbolleistenstrukturen verwendet werden.
   * 
   * @param pluginId
   *          ID des Plug-ins, von dem diese Action geliefert wird.
   * @param action
   *          Action.
   * @param globaleAction
   *          Ist <code>true</code>, wenn es sich um eine globaleAction handelt, sonst <code>false</code>.
   */
  public DeskriptorAktion(String pluginId, IAction action, boolean globaleAction)
  {
    this._pluginID = pluginId;
    this._aktion = action;
    this._aktionID = action.getId();
    this._text = action.getText();
    this._toolTipText = action.getToolTipText();
    this._aktionBeschreibung = action.getDescription();
    this._globaleAktion = globaleAction;

    ImageDescriptor imageDescriptor = action.getImageDescriptor();
    if (imageDescriptor != null)
    {
      this._iconPfad = imageDescriptor.toString();
    }
    else
    {
      this._iconPfad = "";
    }

    ImageDescriptor hoverImageDescriptor = action.getHoverImageDescriptor();
    if (hoverImageDescriptor != null)
    {
      this._hoverIconPfad = hoverImageDescriptor.toString();
    }

    ImageDescriptor disabledImageDescriptor = action.getDisabledImageDescriptor();
    if (disabledImageDescriptor != null)
    {
      this._deaktiviertesIconPfad = disabledImageDescriptor.toString();
    }

    _debug.fine(getClass().getName() + " => ActionID: " + getAktionID() + " PluginID: " + getPluginID() + " Text: "
        + getText());

  }

  /**
   * Konstruktor, initialisiert ein Objekt vom Typ {@link DeskriptorElement}. Dieses dient als Stellvertreterobjekt fr
   * noch nicht vorhandene Actions, so dass trotzdem schon eine Menstruktur bzw. Symbolleistenstruktur aufgebaut werden
   * kann.
   * 
   * @param pluginId
   *          ID des Plug-Ins.
   * @param actionId
   *          ID der Action.
   * @param labelText
   *          Label der Action.
   * @param toolTipText
   *          ToolTipText der Action.
   * @param actionDescription
   *          ActionDescription
   * @param iconPfad
   *          Pfad zum Icon der Action.
   * @param hoverIconPfad
   *          Pfad zum HoverIcon der Action.
   * @param disabledIconPfad
   *          Pfad zum DisabledIcon der Action.
   */
  public DeskriptorAktion(String pluginId, String actionId, String labelText, String toolTipText,
      String actionDescription, String iconPfad, String hoverIconPfad, String disabledIconPfad)
  {
    this._pluginID = pluginId;
    this._aktionID = actionId;
    this._text = labelText;
    this._toolTipText = toolTipText;
    this._aktionBeschreibung = actionDescription;
    this._iconPfad = iconPfad;
    this._hoverIconPfad = hoverIconPfad;
    this._deaktiviertesIconPfad = disabledIconPfad;
    this._globaleAktion = false;

    // ActionElement aufbauen. Es ist ja schon in der Registrierung vorhanden.
    ElementRegistrierung elementRegistrierung = ElementRegistrierung.getInstanz();

    IAction tempAction = elementRegistrierung.getDeskriptorAktion(this._aktionID).getAktion();
    if (tempAction != null)
    {
      this._aktion = tempAction;
    }
  }

  /**
   * Konstruktor, initialisiert ein Objekt vom Typ {@link DeskriptorElement}. Dieses dient als Stellvertreterobjekt fr
   * noch nicht vorhandene Actions, so dass trotzdem schon eine Menstruktur bzw. Symbolleistenstruktur aufgebaut werden
   * kann.
   * 
   * @param pluginId
   *          ID des Plug-Ins.
   * @param actionId
   *          ID der Action.
   * @param globaleAction
   *          Flag fr GlobaleAction
   * @param labelText
   *          Label der Action.
   * @param toolTipText
   *          ToolTipText der Action.
   * @param actionDescription
   *          ActionDescription
   * @param iconPfad
   *          Pfad zum Icon der Action.
   * @param hoverIconPfad
   *          Pfad zum HoverIcon der Action.
   * @param disabledIconPfad
   *          Pfad zum DisabledIcon der Action.
   */
  public DeskriptorAktion(String pluginId, String actionId, boolean globaleAction, String labelText,
      String toolTipText, String actionDescription, String iconPfad, String hoverIconPfad, String disabledIconPfad)
  {
    this._pluginID = pluginId;
    this._aktionID = actionId;
    this._globaleAktion = globaleAction;
    this._text = labelText;
    this._toolTipText = toolTipText;
    this._aktionBeschreibung = actionDescription;
    this._iconPfad = iconPfad;
    this._hoverIconPfad = hoverIconPfad;
    this._deaktiviertesIconPfad = disabledIconPfad;

    // ActionElement aufbauen. Es ist ja schon in der Registrierung vorhanden.
    ElementRegistrierung elementRegistrierung = ElementRegistrierung.getInstanz();

    IAction tempAction = null;

    if (!_globaleAktion)
    {
      tempAction = elementRegistrierung.getDeskriptorAktion(this._aktionID).getAktion();
    }
    else
    {
      tempAction = elementRegistrierung.getDeskriptorAktionGlobal(this._aktionID).getAktion();
    }

    if (tempAction != null)
    {
      this._aktion = tempAction;
    }
  }

  /*
   * Getter
   */

  /*
   * (Kein Javadoc)
   * 
   * @see de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getAction()
   */
  public IAction getAktion()
  {
    return this._aktion;
  }

  /*
   * Getter
   */

  /**
   * @return PluginID.
   */
  public String getPluginID()
  {
    return _pluginID;
  }

  /**
   * Liefert die ID der reprsentierten Action.
   */
  public String getAktionID()
  {
    return _aktionID;
  }

  /**
   * Liefert den Text der Action.
   * 
   * @return Text der Action.
   */
  public String getText()
  {
    return _text;
  }

  /**
   * Liefert den ToolTipText der Action.
   * 
   * @return ToolTipText de Action.
   */
  public String getToolTipText()
  {
    return _toolTipText;
  }

  /**
   * Liefert die Beschreibung der Action.
   * 
   * @return Beschreibung der Action.
   */
  public String getAktionBeschreibung()
  {
    return _aktionBeschreibung;
  }

  /**
   * Liefert den Pfad zum Icon der Action.
   * 
   * @return Iconpfad.
   */
  public String getIconPfad()
  {
    return _iconPfad;
  }

  /**
   * Liefert den Pfad zum HoverIcon der Action.
   * 
   * @return Hovericonpfad.
   */
  public String getHoverIconPfad()
  {
    return _hoverIconPfad;
  }

  /**
   * Liefert den Pfad zum DisabledIcon der Action.
   * 
   * @return Disablediconpfad.
   */
  public String getDeaktiviertesIconPfad()
  {
    return _deaktiviertesIconPfad;
  }

  /**
   * Liefert das Flag der globalen Action.
   * 
   * @return Liefert <code>true</code>, wenn es sich um eine globale Action handelt, sonst <code>false</code>;
   */
  public boolean istGlobaleAktion()
  {
    return _globaleAktion;
  }

  /**
   * Liefert <code>true</code>, wenn die Aktion <code>SWT.DROP_DOWN</code> ist, sonst <code>false</code>.
   * 
   * @return
   */
  public boolean istAusklappAktion()
  {
    if (_aktion.getStyle() == SWT.DROP_DOWN)
    {
      return true;
    }
    else
    {
      return false;
    }
  }

	@Override
	public String toString()
	{
		return getText();
	}

  // /* (Kein Javadoc)
  // * @see de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getPluginID()
  // */
  // public String getPluginID()
  // {
  // return this._pluginID;
  // }
  //
  // /* (Kein Javadoc)
  // * @see de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getActionID()
  // */
  // public String getActionID()
  // {
  // return _action.getId();
  // }
  //
  // /* (Kein Javadoc)
  // * @see de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getLabelText()
  // */
  // public String getLabelText()
  // {
  // return _action.getText();
  // }
  //
  // /* (Kein Javadoc)
  // * @see de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getToolTipText()
  // */
  // public String getToolTipText()
  // {
  // return _action.getToolTipText();
  // }
  //
  // /* (Kein Javadoc)
  // * @see
  // de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getActionDescription()
  // */
  // public String getActionDescription()
  // {
  // return _action.getDescription();
  // }
  //
  // /* (Kein Javadoc)
  // * @see de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getIcon()
  // */
  // public Image getIcon()
  // {
  // if (_action.getImageDescriptor() != null)
  // {
  // Image image = _action.getImageDescriptor().createImage();
  // return image;
  // } else
  // {
  // return null;
  // }
  //
  // }
  //
  // /* (Kein Javadoc)
  // * @see de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getHoverIcon()
  // */
  // public Image getHoverIcon()
  // {
  // if (_action.getHoverImageDescriptor() != null)
  // {
  // return _action.getHoverImageDescriptor().createImage();
  // } else
  // {
  // return null;
  // }
  //
  // }
  //
  // /* (Kein Javadoc)
  // * @see
  // de.bsvrz.buv.rw.basislib.menu.IDescriptorElement#getDisabledIconPfad()
  // */
  // public String getDisabledIconPfad()
  // {
  // // if (_action.getDisabledImageDescriptor() != null)
  // // {
  // // return _action.getDisabledImageDescriptor().createImage();
  // // } else
  // // {
  // // return null;
  // // }
  // return null;
  // }
  
  

}
