/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.rw.statusleiste;

import org.eclipse.jface.action.ContributionItem;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.IStatusLineManager;
import org.eclipse.swt.graphics.Image;

import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Diese Klasse enthlt einen Verweis auf die globale Statusleiste und ermglicht so den Zugriff. <br>
 * Diese Klasse ist als Singleton ausgelegt, da nur eine globale Statusleiste verfgbar ist. <br>
 * Diese Klasse arbeitet auf den gngigen Eclipse-Klassen {@link IContributionItem} sowie {@link IStatusLineManager}.
 * 
 * @version $Revision: 1.1 $ / $Date: 2009/12/02 09:21:06 $ / ($Author: HofmannJ $)
 * 
 * @author HofmannJ
 * 
 */
public class StatusLeisteZugriff
{
  /**
   * Der Debugger
   */
  private Debug _debug = Debug.getLogger();

  /**
   * Speichert den Statuszeilenmanager.
   */
  private IStatusLineManager _statusZeilenManager = null;

  /**
   * Privater Konstruktor dieser Singleton-Klasse.
   * 
   */
  private StatusLeisteZugriff()
  {
  }

  /**
   * Speichert die Singleton-Instanz der Klasse {@link StatusLeisteZugriff}.
   * 
   * @author HofmannJ
   * 
   */
  private static class Innen
  {
    private static StatusLeisteZugriff INSTANZ = new StatusLeisteZugriff();
  }

  /**
   * Liefert die Singleton-Instanz der Klasse {@link StatusLeisteZugriff}.
   * 
   * @return Singleton-Instanz der Klasse {@link StatusLeisteZugriff}
   */
  public static StatusLeisteZugriff getInstanz()
  {
    return Innen.INSTANZ;
  }

  /**
   * Setzt den globalen StatusLineManager.
   * 
   * @param statusLineManager
   *          Zu setzender StatusLineManager.
   */
  public void setStatusZeilenManager(IStatusLineManager statusLineManager)
  {
    if (statusLineManager != null)
    {
      _debug.fine(getClass().getName() + "::setStatusZeilenManager => StatusLineManager wurde gesetzt.");
    }
    this._statusZeilenManager = statusLineManager;
  }

  /**
   * Liefert den globalen StatuLineManager.
   * 
   * @return {@link IStatusLineManager#}
   */
  public IStatusLineManager getStatusZeilenManager()
  {
    return this._statusZeilenManager;
  }

  /**
   * Fgt ein Element zur Statuszeile hinzu.
   * 
   * @param statusleisteElement
   *          Element, dass zur Statuszeile hinzugefgt werden soll.
   */
  public void hinzufuegenStatusZeilenElement(IContributionItem statusleisteElement)
  {
    if (_statusZeilenManager != null)
    {
      _debug.fine("Lnge vorher: " + _statusZeilenManager.getItems().length);

      _statusZeilenManager.add(statusleisteElement);
      _statusZeilenManager.update(true);
    } 
  }

  /**
   * Entfernt ein Element aus der Statusleiste und liefert dieses als {@link IContributionItem} zurck. Wird das
   * entsprechende Element nicht in der Statuszeile gefunden, wird <code>null</code> zurckgeliefert.
   * 
   * @param item
   *          Element, dass aus der Statuszeile entfernt werden soll.
   * @return Entferntes Element als {@link IContributionItem} oder <code>null</code>.
   */
  public IContributionItem entferneStatusZeilenElement(ContributionItem item)
  {
    IContributionItem itemRemoved = _statusZeilenManager.remove(item);
    _statusZeilenManager.update(true);
    // itemRemoved.dispose();
    return itemRemoved;
  }

  /**
   * Entfernte alle Statuszeilenelemente aus der Statuszeile.
   * 
   */
  public void entferneAlleStatusZeilenElemente()
  {
    _statusZeilenManager.removeAll();
    _statusZeilenManager.update(true);
    loescheStatusZeilenMeldung();
    loescheStatusZeilenFehlerMeldung();
  }

  /**
   * Setzt eine Meldung i.d. Statuszeile.<br>
   * Diese Meldung ist <b>nur</b> sichtbar, wenn keine Fehlermeldung gesetzt ist.
   * 
   * @param meldungsText
   *          Text der Meldung.
   */
  public void setStatusZeilenMeldung(String meldungsText)
  {
    _statusZeilenManager.setMessage(meldungsText);
  }

  /**
   * Setzt eine Meldung mit Bild i.d. Statuszeile.<br>
   * Diese Meldung ist <b>nur</b> sichtbar, wenn keine Fehlermeldung gesetzt ist.
   * 
   * @param meldungsText
   *          Text der Meldung.
   * @param meldungsBild
   *          Bild der Meldung.
   */
  public void setStatusZeilenMeldung(String meldungsText, Image meldungsBild)
  {
    _statusZeilenManager.setMessage(meldungsBild, meldungsText);
  }

  /**
   * Setzt eine Fehlermeldung i.d. Stautzeile.<br>
   * <b>berlargert</b> eine Meldung, wenn gesetzt.
   * 
   * @param fehlerMeldungsText
   *          Text der Fehlermeldung.
   */
  public void setStatusZeilenFehlerMeldung(String fehlerMeldungsText)
  {
    _statusZeilenManager.setErrorMessage(fehlerMeldungsText);
  }

  /**
   * Setzt eine Fehlermeldung mit Bild i.d. Statuszeile.<br>
   * <b>berlargert</b> eine Meldung, wenn gesetzt.
   * 
   * @param fehlerMeldungsText
   *          Text der Fehlermeldung.
   * @param fehlerMeldungsBild
   *          Bild der Fehlermeldung.
   */
  public void setStatusZeilenFehlerMeldung(String fehlerMeldungsText, Image fehlerMeldungsBild)
  {
    _statusZeilenManager.setErrorMessage(fehlerMeldungsBild, fehlerMeldungsText);
  }

  /**
   * Lscht den Meldungsbereich i.d. Statuszeile.
   * 
   */
  public void loescheStatusZeilenMeldung()
  {
    _statusZeilenManager.setMessage(null);
  }

  /**
   * Lscht dern Fehlermeldungsbereich i.d. Statuszeile.
   * 
   */
  public void loescheStatusZeilenFehlerMeldung()
  {
    _statusZeilenManager.setErrorMessage(null);
  }

}
