/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.rw.statusleiste.xml.datenstruktur;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Stack;

import org.jdom.Document;

/**
 * Diese Klasse liefert einen UNDO-Stack fr die Struktur der Statusleiste, definiert in {@link StatusLeisteVerwaltung}.
 * 
 * @author HofmannJ
 * 
 */
public class UndoStackStatusLeiste
{
  /**
   * UNDO-Stack
   */
  private Stack<Document> _undoStack = new Stack<Document>();

  /**
   * Eine REDO-Operation
   */
  private Document _redo = null;

  /**
   * Liste aller {@link IZuhoererUndoStackStatusLeiste} Listener, die ber den aktuellen Zustand des UNDO-Stacks
   * informiert werden mchten.
   */
  private LinkedList<IZuhoererUndoStackStatusLeiste> _zuhoerer = new LinkedList<IZuhoererUndoStackStatusLeiste>();

  /**
   * Konstruktor der Klasse {@link UndoStackStatusLeiste}.
   */
  public UndoStackStatusLeiste()
  {
  }

  /**
   * Legt eine Struktur auf den Stack.
   * 
   * @param doc
   *          Hinzuzufgende Struktur.
   */
  public void push(Document doc)
  {
    _undoStack.push(doc);

    benachrichtigeZuhoerer();
  }

  /**
   * Liefert die oberste Struktur auf dem {@link Stack} zurck.
   * 
   * @param aktZustand
   *          Aktueller Zustand, der fr eine Rckgngig-Operation gesichert werden soll.
   * @return Liefert die oberste Struktur auf dem {@link Stack} zurck oder <code>null</code>, wenn keine existiert.
   */
  public Document pop(Document aktZustand)
  {
    Document re = null;

    if (_undoStack.size() > 0)
    {
      _redo = aktZustand;      

      re = _undoStack.pop();

      benachrichtigeZuhoerer();
    }

    return re;
  }

  /**
   * Liefert den Redo-Zustand, sofern vorhanden.
   * 
   * @return Letzter Redo-Zustand als Instanz der Klasse {@link Document} oder <code>null</code>, wenn nicht
   *         vorhanden.
   */
  public Document getRedo()
  {
    Document tempRedo = _redo;
    _redo = null;
    benachrichtigeZuhoerer();
    return tempRedo;
  }

  /**
   * Prft, ob der Redo-Zustand gefllt ist.
   * 
   * @return Liefert <code>true</code>, wenn der Redo-Zustand gefllt ist, sonst <code>false</code>.
   */
  public boolean isRedoFilled()
  {
    if (_redo != null)
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  /**
   * Prft, ob der Undo-Stack gefllt ist.
   * 
   * @return Liefert <code>true</code>, wenn sich Elemente auf dem Undo-Stack befinden, sonst <code>false</code>.
   */
  public boolean isStackFilled()
  {
    if (_undoStack.size() > 0)
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  /**
   * Fgt einen Listener vom Typ {@link IZuhoererUndoStackStatusLeiste} zur Liste der Listener hinzu, sofern dieser noch
   * nicht in der Liste vorhanden ist.
   * 
   * @param zuhoerer
   *          Hinzuzufgender {@link IZuhoererUndoStackStatusLeiste}
   */
  public void addZuhoererUndoStack(IZuhoererUndoStackStatusLeiste zuhoerer)
  {
    if (!_zuhoerer.contains(zuhoerer))
    {
      _zuhoerer.add(zuhoerer);
      benachrichtigeZuhoerer();
    }
  }

  /**
   * Entfernt einen Listener vom Typ {@link IZuhoererUndoStackStatusLeiste} von der Liste der Listener, sofern dieser in
   * der Liste vorhanden ist.
   * 
   * @param zuhoerer
   *          Zu entfernender {@link IZuhoererUndoStackStatusLeiste}
   */
  public void removeZuhoererUndoStack(IZuhoererUndoStackStatusLeiste zuhoerer)
  {
    if (_zuhoerer.contains(zuhoerer))
    {
      _zuhoerer.remove(zuhoerer);
    }
  }

  /**
   * Benachrichtigt alle registrierten Listener vom Typ {@link IZuhoererUndoStackStatusLeiste} ber den aktuellen
   * Zustand des Undo-Stacks. Sie werden darber informiert, ob z.Zt.<br>
   * <ul>
   * <li>ein UNDO (Rckgngig machen) mglich ist,</li>
   * <li>ein REDO (Wiederholen) mglich ist.</li>
   * <ul>
   */
  private void benachrichtigeZuhoerer()
  {
    if (_zuhoerer.size() > 0)
    {
      boolean redoFilled = isRedoFilled();
      boolean undoStackFilled = isStackFilled();

      Iterator<IZuhoererUndoStackStatusLeiste> it = _zuhoerer.iterator();
      while (it.hasNext())
      {
        IZuhoererUndoStackStatusLeiste zuhoererUndoStack = it.next();
        zuhoererUndoStack.rueckgaengigmachenMoeglich(undoStackFilled);
        zuhoererUndoStack.wiederholenMoeglich(redoFilled);
      }
    }
  }

  public void resetRedo()
  {
    _redo = null;
  }

}
