/*
 * Copyright 2007 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.dav.daf.
 * 
 * de.bsvrz.dav.daf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dav.daf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with de.bsvrz.dav.daf; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.dav.daf.communication.lowLevel;

import java.util.*;

/**
 * Klasse, die zum gepufferten Austausch von Telegrammen zwischen verschiedenen Threads verwendet werden kann. Die Gesamtgre der gepufferten Telegramme ist
 * beschrnkt. Es werden verschiedene Telegrammprioritten untersttzt.
 * <p/>
 * Telegramme knnen mit der Methode {@link #put} gespeichert werden und mit der Methode {@link #take} wieder ausgelesen werden. Die Methoden blockieren, wenn
 * beim Speichern nicht gengend Platz vorhanden ist, bzw., wenn beim Auslesen kein Telegramm mehr zur Verfgung steht. Der Methode {@link #close} dient zum
 * Schlieen der Queue. blockiert keine der beiden Methoden mehr.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5047 $
 */
public class TelegramQueue<Telegram extends QueueableTelegram> {

	/** Maximale Gesamtgre fr zwischengespeicherte Telegramme. */
	final private int _capacity;

	/** Gesamtgre der aktuell zwischengespeicherten Telegramme. */
	private int _size;

	/**
	 * Array, das je mgliche Prioritt eine verkettete Liste mit den zwischengespeicherten Telegrammen enthlt. Es dient auerdem der Synchronisation von Threads
	 * beim lesenden und schreibenden Zugriff.
	 */
	final private LinkedList<Telegram>[] _priorityLists;

	private boolean _closed = false;

	/**
	 * Erzeugt eine neue Queue mit den angegebenen Eigenschaften.
	 *
	 * @param capacity        Maximale Gesamtgre der gepufferten Telegramme.
	 * @param maximumPriority Maximale von Telegrammen verwendete Prioritt.
	 */
	public TelegramQueue(int capacity, int maximumPriority) {
		if(capacity <= 0) throw new IllegalArgumentException("capacity muss positiv sein: " + capacity);
		if(maximumPriority < 0) throw new IllegalArgumentException("maximumPriority darf nicht negativ sein: " + maximumPriority);
		if(maximumPriority > 127) throw new IllegalArgumentException("maximumPriority darf nicht grer als 127 sein: " + maximumPriority);
		_capacity = capacity;
		_size = 0;
		_priorityLists = (LinkedList<Telegram>[])new LinkedList[maximumPriority + 1]; // Compiler-Warnung nicht vermeidbar
		for(int i = 0; i < _priorityLists.length; i++) {
			_priorityLists[i] = new LinkedList<Telegram>();
		}
	}

	/**
	 * Gibt das lteste in der Queue gespeicherte Telegramm mit der hchsten Prioritt zurck. Wenn die Queue noch nicht geschlossen wurde, wartet diese Methode,
	 * bis ein Telegramm in der Queue zur Verfgung steht.
	 *
	 * @return Nchstes gespeicherte Telegramm mit der hchsten Prioritt. Wenn die Queue geschlossen wurde und kein gespeichertes Telegramm mehr verfgbar ist
	 *         wird <code>null</code> zurckgegeben.
	 *
	 * @throws InterruptedException Wenn der Thread whrend des Wartens unterbrochen wurde.
	 */
	public Telegram take() throws InterruptedException {
		synchronized(this) {
			while(_size == 0) {
				// Wenn die Queue leer ist und geschlossen wurde, wird null zurckgegeben
				if(_closed) return null;
				// Wenn die Queue leer ist und nicht geschlossen wurde, wird gewartet
				wait();
			}
			for(int i = _priorityLists.length - 1; i >= 0; i--) {
				LinkedList<Telegram> priorityList = _priorityLists[i];
				if(!priorityList.isEmpty()) {
					final Telegram telegram = priorityList.removeFirst();
					_size -= telegram.getSize();
					notifyAll();
					return telegram;
				}
			}
		}
		throw new IllegalStateException("Interner Fehler: Es wurde kein Telegramm gefunden, obwohl die Gesamtgre " + _size + " ist");
	}

	/**
	 * Speichert das angegebene Telegramm in der Queue. Bei Bedarf wartet diese Methode bis gengend Platz in der Queue fr das zu speichernde Telegramm zur
	 * Verfgung steht.
	 *
	 * @param telegram Das zu speichernde Telegramm
	 *
	 * @throws InterruptedException Wenn der Thread whrend des Wartens unterbrochen wurde.
	 */
	public void put(Telegram telegram) throws InterruptedException {
		if(_closed) return;
		final int length = telegram.getSize();
		if(length <= 0) throw new IllegalArgumentException("Telegrammlnge muss grer 0 sein, ist aber " + length + ": " + telegram);
		final byte priority = telegram.getPriority();
		synchronized(this) {
			while(!_closed && _size + length > _capacity) {
				wait();
			}
			if(_closed) return;
			_priorityLists[priority].add(telegram);
			_size += length;
			notifyAll();
		}
	}

	/**
	 * Bestimmt die maximale Gesamtgre fr zwischengespeicherte Telegramme.
	 *
	 * @return Maximale Gesamtgre fr zwischengespeicherte Telegramme.
	 */
	public int getCapacity() {
		return _capacity;
	}

	/**
	 * Bestimmt die Gesamtgre der aktuell zwischengespeicherten Telegramme.
	 *
	 * @return Gesamtgre der aktuell zwischengespeicherten Telegramme.
	 */
	public int getSize() {
		synchronized(this) {
			return _size;
		}
	}


	/**
	 * Diese Methode schliet die Verbindung. Danach ignoriert die Methode {@link #put} smtliche weitere zu speichernde Telegramme und die Methode {@link #take}
	 * liefert noch alle bisher gespeicherten Telegramme und danach <code>null</code> zurck. Eventuell blockierte Threads werden geweckt.
	 */
	public void close() {
		synchronized(this) {
			_closed = true;
			notifyAll();
		}
	}

	/**
	 * Diese Methode schliet die Verbindung und lscht alle noch gespeicherten Telegramme. Danach ignoriert die Methode {@link #put} smtliche weitere zu
	 * speichernde Telegramme und die Methode {@link #take} liefert anschlieend immer <code>null</code> zurck. Eventuell blockierte Threads werden geweckt.
	 */
	public void abort() {
		synchronized(this) {
			_closed = true;
			notifyAll();
			try {
				while(take() != null) ;
			}
			catch(InterruptedException ignored) {
			}
		}
	}
}
