/*
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2006 by Kappich Systemberatung Aachen
 * Copyright 2005 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.dav.daf.
 * 
 * de.bsvrz.dav.daf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dav.daf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with de.bsvrz.dav.daf; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.dav.daf.main.impl.config;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectInfo;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.sys.funclib.debug.Debug;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;

/**
 * Implementierung der Schnittstelle zum Zugriff auf die Eigenschaften eines System-Objekts. Datenstze, die vom Datenverteiler transportiert werden, sind immer
 * genau einem System-Objekt zugeordnet. Zur internen Identifizierung eines System-Objekts wird die <code>id</code> des Objekts benutzt. Das ist ein 64-Bit-Wert
 * der systemweit eindeutig ist. Zur expliziten Referenzierung von Objekten (z.B. in Aufrufparametern von Applikationen, Versorgungsskripten etc.) kann die
 * permanente ID (PID) eines Objekts verwendet werden. Diese ist optional und muss nur bei den Objekten gesetzt werden, bei denen sie gebraucht wird. Die PID
 * ist eine leicht zu merkende Zeichenkette, die systemweit eindeutig ist und sich (wenn sie einmal vergeben wurde) nie mehr ndern kann. Auerdem kann jedem
 * System-Objekt ein Name zugeordnet werden, der zu Darstellungszwecken benutzt werden kann. Der Name eines Objekts kann i.a. jederzeit gendert werden und
 * sollte aus diesem Grund nicht zur Referenzierung von Objekten eingesetzt werden.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5946 $ / $Date: 2008-03-18 20:57:43 +0100 (Tue, 18 Mar 2008) $ / ($Author: rs $)
 */
public abstract class DafSystemObject implements SystemObject {

	/** Typfeld fr die Serialisierung von <code>null</code>-Referenzen */
	public static final byte NULL_OBJECT = 0;

	/** Typfeld fr die Serialisierung von Konfigurationsobjekten */
	public static final byte CONFIGURATION_OBJECT = 1;

	/** Typfeld fr die Serialisierung von dynamischen Objekten */
	public static final byte DYNAMIC_OBJECT = 2;

	/** Typfeld fr die Serialisierung von Typen von Systemobjekten */
	public static final byte SYSTEM_OBJECT_TYPE = 3;

	/** Typfeld fr die Serialisierung von Typen von dynamischen Objekten */
	public static final byte DYNAMIC_OBJECT_TYPE = SYSTEM_OBJECT_TYPE;

	/** Typfeld fr die Serialisierung von Typen von Konfigurationsobjekten */
	public static final byte CONFIGURATION_OBJECT_TYPE = 4;

	/** Typfeld fr die Serialisierung von Mengentypen */
	public static final byte OBJECT_SET_TYPE = 5;

	/** Typfeld fr die Serialisierung von Attributgruppen */
	public static final byte ATTRIBUTE_GROUP = 6;

	/** Typfeld fr die Serialisierung von Aspekten */
	public static final byte ASPECT = 7;

	/** Typfeld fr die Serialisierung von Mengenverwendungen */
	public static final byte OBJECT_SET_USE = 8;

	/** Typfeld fr die Serialisierung von dynamischen Mengen */
	public static final byte MUTABLE_SET = 9;

	/** Typfeld fr die Serialisierung von konfigurierenden Mengen */
	public static final byte NON_MUTABLE_SET = 10;

	/** Typfeld fr die Serialisierung von Attributen */
	public static final byte ATTRIBUTE = 11;

	/** Typfeld fr die Serialisierung von Ganzzahlattributtypen */
	public static final byte INTEGER_ATTRIBUTE_TYPE = 13;

	/** Typfeld fr die Serialisierung von Kommazahlattributtypen */
	public static final byte FLOATING_POINT_NUMBER_ATTRIBUTE_TYPE = 14;

	/** Typfeld fr die Serialisierung von Referenzattributtypen */
	public static final byte REFERENCE_ATTRIBUTE_TYPE = 15;

	/** Typfeld fr die Serialisierung von Zeichenkettenattributtypen */
	public static final byte STRING_ATTRIBUTE_TYPE = 16;

	/** Typfeld fr die Serialisierung von Zeitstempelattributtypen */
	public static final byte TIME_ATTRIBUTE_TYPE = 17;

	/** Typfeld fr die Serialisierung von Attributlistendefinitionen */
	public static final byte ATTRIBUTE_LIST_TYPE = 18;

	/** Typfeld fr die Serialisierung von Konfigurationsverantwortliche */
	public static final byte CONFIGURATION_AUTHORITY = 19;

	/** Typfeld fr die Serialisierung von Datenverteilerobjekte */
	public static final byte DAV_APPLICATION = 20;

	/** Typfeld fr die Serialisierung von Konfigurationsapplikationen */
	public static final byte CONFIGURATION_APPLICATION = 21;

	/** Typfeld fr die Serialisierung von Applikationen */
	public static final byte CLIENT_APPLICATION = 22;

	/** Typfeld fr die Serialisierung von Aufzhlungswerten von Ganzzahlattributtypen */
	public static final byte INTEGER_VALUE_STATE = 23;

	/** Typfeld fr die Serialisierung von Wertebereichen von Ganzzahlattributtypen */
	public static final byte INTEGER_VALUE_RANGE = 24;

	/** Typfeld fr die Serialisierung von Konfigurationsbereichen */
	public static final byte CONFIGURATION_AREA = 25;

	/** Typfeld fr die Serialisierung von Attributgruppenverwendungen */
	public static final byte ATTRIBUTE_GROUP_USAGE = 26;

	/** Statusfeld fr ungltig gewordene (dynamische) Objekte */
	public static final byte OBJECT_DELETED = 0;

	/** Statusfeld fr gltige Objekte */
	public static final byte OBJECT_EXISTS = 1;

	/** Statusfeld fr nicht gltige (Konfigurations-) Objekte */
	public static final byte OBJECT_INVALID = 2;

	/** Die ID dieses Objekts */
	private long _id;

	/** Die PID dieses Objekts */
	private String _pid;

	/** Der Name dieses Objekts */
	private String _name;

	/** Die ID des Objekttyps dieses Objekts */
	private long _typeId;

	/** Der Objekttyp dieses Objekts */
	private DafSystemObjectType _systemObjectType;

	/**
	 * Statusfeld dieses Objektes.
	 *
	 * @see #OBJECT_DELETED
	 * @see #OBJECT_EXISTS
	 */
	private byte _state;

	/** Fehlerbeschreibung wenn vorhanden, sonst null. */
	private String _error;

	/** Typfeld dieses Objekts. Dieses Feld dient zur Unterscheidung der verschiedenartigen Systemobjekte bei der Serialisierung */
	protected byte _internType;

	/** Objekt zum Zugriff auf die Konfiguration */
	protected DafDataModel _dataModel;

	/** Id des Konfigurationsbereichs zu dem dieses Objekt gehrt. */
	private long _configurationAreaId;

	/** DebugLogger fr Debug-Ausgaben */
	private static final Debug _debug = Debug.getLogger();

	/**
	 * Erzeugt ein neues Objekt dessen Eigenschaften im Anschluss mit der read-Methode eingelesen werden sollten.
	 *
	 * @param dataModel DataModel Implementierung, der das neue Objekt zugeordnet ist.
	 */
	public DafSystemObject(DafDataModel dataModel) {
		_dataModel = dataModel;
	}

	/**
	 * Konstruktor, zur Erzeugung eines neuen Stellvertreterobjekts fr ein Systemobjekt.
	 *
	 * @param id                  Id des Systemobjekts
	 * @param pid                 Pid des Systemobjekts
	 * @param name                Name des Systemobjekts
	 * @param typeId              Id des Typs des SystemObjects
	 * @param state               Status des SystemObjects
	 * @param error               Fehler des Systemobjekts
	 * @param dataModel           Objekt zum Zugriff auf die Konfiguration
	 * @param configurationAreaId Id des zugeordneten Konfigurationsbereichs
	 */
	public DafSystemObject(
			long id, String pid, String name, long typeId, byte state, String error, DafDataModel dataModel, long configurationAreaId
	) {
		_id = id;
		_pid = pid;
		_name = name;
		_typeId = typeId;
		_state = state;
		_error = error;
		_dataModel = dataModel;
		_configurationAreaId = configurationAreaId;
	}

	public ConfigurationArea getConfigurationArea() {
		return (DafConfigurationArea)_dataModel.getObject(_configurationAreaId);
	}

	/**
	 * Bestimmt die Objekt-Id des Konfigurationsbereichs des Objekts
	 *
	 * @return Objekt-Id des Konfigurationsbereichs des Objekts
	 */
	public long getConfigurationAreaId() {
		return _configurationAreaId;
	}

	/**
	 * Vergleicht das Objekt mit einem anderen Objekt. Zwei Objekte sind gleich, wenn sie die gleiche Objekt-Id haben.
	 *
	 * @return <code>true</code>, wenn die Objekte gleich sind, sonst <code>false</code>.
	 */
	public final boolean equals(Object other) {
		if(!(other instanceof DafSystemObject)) {
			return false;
		}
		return ((DafSystemObject)other)._id == _id;
	}

	/**
	 * Bestimmt den Hash-Code dieses Objekts.
	 *
	 * @return Hash-Code dieses Objekts.
	 */
	public final int hashCode() {
		return (int)(_id ^ (_id >>> 32));
	}

	/**
	 * Liefert eine textuelle Beschreibung dieses Objekts fr Debug-Zwecke zurck.
	 *
	 * @return Beschreibender Text dieses Objekts.
	 */
	public String parseToString() {
		String str = "ID: " + _id + "\n";
		str += "PID: " + _pid + "\n";
		str += "Name: " + _name + "\n";
		str += "Typ Id: " + _typeId + "\n";
		return str;
	}

	/**
	 * Liefert eine textuelle Beschreibung dieses Objekts zurck. Es wird der Name des Objekts gefolgt von der Pid in runden Klammern oder der Id in eckigen
	 * Klammern, falls keine Pid vergeben ist.
	 */
	public String toString() {
		StringBuilder text = new StringBuilder();
		if(_name != null && !_name.equals("")) {
			text.append(_name).append(" ");
		}
		if(_pid != null && !_pid.equals("")) {
			text.append("(").append(_pid).append(")");
		}
		else {
			text.append("[").append(_id).append("]");
		}
		return text.toString();
	}

	/**
	 * Serialisiert dieses Objekt.
	 *
	 * @param out Stream auf den das Objekt geschrieben werden soll.
	 *
	 * @throws IOException, wenn beim Schreiben auf den Ausgabe-Stream Fehler aufgetreten sind.
	 */
	public void write(DataOutputStream out) throws IOException {
		out.writeLong(_id);
		out.writeLong(_typeId);
		out.writeByte(_state);
		boolean b = (_pid != null);
		out.writeBoolean(b);
		if(b) {
			out.writeUTF(_pid);
		}
		b = (_name != null);
		out.writeBoolean(b);
		if(b) {
			out.writeUTF(_name);
		}
		b = (_error != null);
		out.writeBoolean(b);
		if(b) {
			out.writeUTF(_error);
		}
		out.writeLong(_configurationAreaId);
	}

	/**
	 * Deserialisiert dieses Objekt.
	 *
	 * @param in Stream von dem das Objekt gelesen werden soll.
	 *
	 * @throws IOException, wenn beim Lesen vom Eingabe-Stream Fehler aufgetreten sind.
	 */
	public void read(DataInputStream in) throws IOException {
		_id = in.readLong();
		_typeId = in.readLong();
		_state = in.readByte();
		boolean b = in.readBoolean();
		if(b) {
			_pid = in.readUTF();
		}
		b = in.readBoolean();
		if(b) {
			_name = in.readUTF();
		}
		b = in.readBoolean();
		if(b) {
			_error = in.readUTF();
		}
		_configurationAreaId = in.readLong();
	}

	/**
	 * Bestimmt den Status dieses Objektes
	 *
	 * @return Liefert den Wert <code>OBJECT_DELETED</code> fr ungltige oder <code>OBJECT_EXISTS</code> fr gltige Objekte zurck.
	 *
	 * @see #OBJECT_DELETED
	 * @see #OBJECT_EXISTS
	 */
	public final byte getState() {
		return _state;
	}


	/**
	 * Setzt den Status dieses Objektes.
	 *
	 * @param state <code>OBJECT_DELETED</code> fr ungltige oder <code>OBJECT_EXISTS</code> fr gltige Objekte.
	 *
	 * @see #OBJECT_DELETED
	 * @see #OBJECT_EXISTS
	 */
	void setState(byte state) {
		_state = state;
	}

	/**
	 * Implementiert die Vergleichsfunktion des Comparable-Interface.
	 *
	 * @param o Systemobjekt mit dem dieses Objekt verglichen werden sollte.
	 *
	 * @return Negative Zahl, wenn dieses Objekt kleiner ist als das Vergleichsobjekt; positive Zahl, wenn dieses Objekt grer ist als das Vergleichsobjekt;
	 *         <code>0</code>, wenn dieses Objekt mit dem Vergleichsobjekt bereinstimmt.
	 */
	public final int compareTo(Object o) {
		DafSystemObject other = (DafSystemObject)o;
		if(_id < other._id) {
			return -1;
		}
		if(_id > other._id) {
			return 1;
		}
		return 0;
	}

	public final String getNameOrPidOrId() {
		if(_name != null && !_name.equals("")) return _name;
		if(_pid != null && !_pid.equals("")) return _pid;
		return Long.toString(_id);
	}

	public String getPidOrNameOrId() {
		if(_pid != null && !_pid.equals("")) return _pid;
		if(_name != null && !_name.equals("")) return _name;
		return Long.toString(_id);
	}

	public String getPidOrId() {
		if(_pid != null && !_pid.equals("")) return _pid;
		return Long.toString(_id);
	}

	public final long getId() {
		return _id;
	}

	public final String getPid() {
		return (_pid == null ? "" : _pid);
	}

	public final String getName() {
		return _name;
	}

	public final void setName(String _name) throws ConfigurationChangeException {
		final String name = _name != null ? _name : "";

		if(_name.length() > 255) {
			throw new ConfigurationChangeException(
					"Der Name des Objekts ist zu lang, es sind nur 255 Zeichen erlaubt. Name, der gesetzt werden sollte: " + _name + " Lnge des Strings: "
					+ _name.length()
			);
		}

		if(_dataModel.setName(this, name)) {
			this._name = name;
		}
		else {
			throw new ConfigurationChangeException("Der Name konnte nicht gesetzt werden");
		}
	}

	/**
	 * Diese Methode wird aufgerufen, wenn die Konfiguration den Namen des Objekts auf den aktuellen Stand bringen mchte. Der Aufruf der Methode fhrt nicht zu
	 * einer Anfrage an die Konfiguration wie bei {@link #setName(String)}.
	 * <p/>
	 * Eventuell angemeldete Listener werden nicht informiert, soll dies geschehen muss die Methode {@link DafDynamicObjectType#updateName(long,String)} benutzt
	 * werden.
	 *
	 * @param newName Aktueller Name des Objekts
	 */
	protected void storeName(String newName) {
		this._name = newName;
	}

	public final SystemObjectType getType() {
		if(_systemObjectType == null) {
			_systemObjectType = (DafSystemObjectType)_dataModel.getObject(_typeId);
		}
		return _systemObjectType;
	}

	public final void invalidate() throws ConfigurationChangeException {
		if(!_dataModel.invalidate(this)) {
			throw new ConfigurationChangeException("Das Objekt konnte nicht ungltig gemacht werden.");
		}
	}

	public final boolean isOfType(String typePid) {
		if(typePid == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		if(_systemObjectType == null) {
			getType();
		}
		if(typePid.equals(_systemObjectType.getPid())) {
			return true;
		}
		List list = _systemObjectType.getSuperTypes();
		while((list != null) && (list.size() > 0)) {
			ArrayList tmpList = new ArrayList();
			for(int i = 0; i < list.size(); ++i) {
				DafSystemObjectType _systemObjectType = (DafSystemObjectType)list.get(i);
				if(typePid.equals(_systemObjectType.getPid())) {
					return true;
				}
				List tmp = _systemObjectType.getSuperTypes();
				if(tmp != null) {
					tmpList.addAll(tmp);
				}
			}
			list = tmpList;
		}
		return false;
	}

	public final boolean isOfType(SystemObjectType objectType) {
		String _pid = objectType.getPid();
		if(_pid != null && !_pid.equals("")) {
			return isOfType(_pid);
		}
		else {
			long typeId = objectType.getId();
			if(_systemObjectType == null) {
				getType();
			}
			if(typeId == _systemObjectType.getId()) {
				return true;
			}
			ArrayList list = (ArrayList)_systemObjectType.getSuperTypes();
			while((list != null) && (list.size() > 0)) {
				ArrayList tmpList = new ArrayList();
				for(int i = 0; i < list.size(); ++i) {
					DafSystemObjectType _systemObjectType = (DafSystemObjectType)list.get(i);
					if(typeId == _systemObjectType.getId()) {
						return true;
					}
					tmpList.addAll(_systemObjectType.getSuperTypes());
				}
				list = tmpList;
			}
		}
		return false;
	}

	/**
	 * Gibt den internen Typ zurck
	 *
	 * @return der interne Typ
	 */
	public final byte getInternType() {
		return _internType;
	}

	/**
	 * Gibt ein neues nicht initialisiertes Objekt der richtigen Klasse abhngig vom angegebenen Typfeld zurck
	 *
	 * @param internType Typfeld des gewnschten Systemobjekts
	 * @param dataModel  Objekt zum Zugriff auf die Konfiguration
	 *
	 * @return Neues nicht initialisiertes Systemobjekt
	 */
	public final static DafSystemObject getObject(byte internType, DafDataModel dataModel) {
		switch(internType) {
			case(CONFIGURATION_OBJECT): {
				return new DafConfigurationObject(dataModel);
			}
			case(DYNAMIC_OBJECT): {
				return new DafDynamicObject(dataModel);
			}
			case(DYNAMIC_OBJECT_TYPE): {
				return new DafDynamicObjectType(dataModel);
			}
			case(CONFIGURATION_OBJECT_TYPE): {
				return new DafConfigurationObjectType(dataModel);
			}
			case(OBJECT_SET_TYPE): {
				return new DafObjectSetType(dataModel);
			}
			case(ATTRIBUTE_GROUP): {
				return new DafAttributeGroup(dataModel);
			}
			case(ASPECT): {
				return new DafAspect(dataModel);
			}
			case(OBJECT_SET_USE): {
				return new DafObjectSetUse(dataModel);
			}
			case(MUTABLE_SET): {
				return new DafMutableSet(dataModel);
			}
			case(NON_MUTABLE_SET): {
				return new DafNonMutableSet(dataModel);
			}
			case(ATTRIBUTE): {
				return new DafAttribute(dataModel);
			}
			case(INTEGER_ATTRIBUTE_TYPE): {
				return new DafIntegerAttributeType(dataModel);
			}
			case(INTEGER_VALUE_RANGE): {
				return new DafIntegerValueRange(dataModel);
			}
			case(INTEGER_VALUE_STATE): {
				return new DafIntegerValueState(dataModel);
			}
			case(FLOATING_POINT_NUMBER_ATTRIBUTE_TYPE): {
				return new DafDoubleAttributeType(dataModel);
			}
			case(REFERENCE_ATTRIBUTE_TYPE): {
				return new DafReferenceAttributeType(dataModel);
			}
			case(STRING_ATTRIBUTE_TYPE): {
				return new DafStringAttributeType(dataModel);
			}
			case(TIME_ATTRIBUTE_TYPE): {
				return new DafTimeAttributeType(dataModel);
			}
			case(ATTRIBUTE_LIST_TYPE): {
				return new DafAttributeListDefinition(dataModel);
			}
			case(CONFIGURATION_AUTHORITY): {
				return new DafConfigurationAuthority(dataModel);
			}
			case(CONFIGURATION_AREA): {
				return new DafConfigurationArea(dataModel);
			}
			case(DAV_APPLICATION): {
				return new DafDavApplication(dataModel);
			}
			case(CLIENT_APPLICATION): {
				return new DafClientApplication(dataModel);
			}
			case(ATTRIBUTE_GROUP_USAGE): {
				return new DafAttributeGroupUsage(dataModel);
			}
			default: {
				return null;
			}
		}
	}

	public DataModel getDataModel() {
		return _dataModel;
	}

	public final Data getConfigurationData(AttributeGroup atg) {
		return _dataModel.getConfigurationData(this, atg);
	}

	public Data getConfigurationData(AttributeGroup atg, Aspect asp) {
		return _dataModel.getConfigurationData(this, atg.getAttributeGroupUsage(asp));
	}

	public Data getConfigurationData(AttributeGroupUsage atgUsage) {
		return _dataModel.getConfigurationData(this, atgUsage);
	}

	public void setConfigurationData(AttributeGroup atg, Data data)
			throws ConfigurationChangeException {
		setConfigurationData(atg.getAttributeGroupUsage(_dataModel.getAspect("asp.eigenschaften")), data);
	}

	public void setConfigurationData(
			AttributeGroup atg, Aspect aspect, Data data
	) throws ConfigurationChangeException {
		setConfigurationData(atg.getAttributeGroupUsage(aspect), data);
	}

	public void setConfigurationData(AttributeGroupUsage atgUsage, Data data) throws ConfigurationChangeException {
		_dataModel.setConfigurationData(this, atgUsage, data);
	}

	/**
	 * Diese Methode wird seitens der Datenverteiler-Applikationsfunktionen nicht untersttzt.
	 *
	 * @throws UnsupportedOperationException Immer, da die Methode nicht untersttzt wird.
	 */
	public Collection<AttributeGroupUsage> getUsedAttributeGroupUsages() {
		
		throw new UnsupportedOperationException("Diese Methode 'getUsedAttributeGroupUsage()' wird hier nicht untersttzt.");
	}

	public abstract boolean isValid();

	public SystemObjectInfo getInfo() {
		Data infoData;
		try {
			infoData = getConfigurationData(_dataModel.getAttributeGroup("atg.info"));
		}
		catch(Exception e) {
			infoData = null;
		}
		if(infoData == null) {
			return SystemObjectInfo.UNDEFINED;
		}
		return new SystemObjectInfo(infoData.getTextValue("kurzinfo").getValueText(), infoData.getTextValue("beschreibung").getValueText());
	}
}
