/*
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2004 by Kappich+Kni Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.dav.dav.
 * 
 * de.bsvrz.dav.dav is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dav.dav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.dav.dav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.dav.dav.main;

import de.bsvrz.dav.daf.main.impl.config.AttributeGroupUsageIdentifications;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.BaseSubscriptionInfo;

import java.util.*;

/**
 * Diese Klasse regelt die Authentifizierung der Benutzer(DAV / DAF). In einer Tabelle werden die Benutzer mit ihren Rechten verwaltet.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5068 $
 */
public class AuthentificationManager {

	/** Der Verbindungsmanager */
	private ConnectionsManager _connectionsManager;

	/** User-Tabelle in der die Rechte der Benutzer verwaltet werden */
	private Hashtable<Long, UserInfo> _userTable;

	AuthentificationManager(ConnectionsManager connectionsManager) {
		_connectionsManager = connectionsManager;
		_userTable = new Hashtable<Long, UserInfo>();
	}

	/**
	 * Fgt eine Benutzerinformation zu der Benutzertabelle hinzu, wenn der Datenverteiler die Benutzerrechte prfen soll. Existiert der Benutzer bereits, wird
	 * lediglich die interne Referenz inkrementiert.
	 *
	 * @param userId BenutzerID
	 */
	final synchronized void addUser(long userId) {
		if(_connectionsManager.getServerDavParameters().isUserRightsCheckingEnabled()) {
			Long key = new Long(userId);
			UserInfo userInfo = _userTable.get(key);
			if(userInfo == null) {
				userInfo = new UserInfo(userId, _connectionsManager);
				_userTable.put(key, userInfo);
			}
			else {
				userInfo.incrementReference();
			}
		}
	}

	/**
	 * Lscht einen Benutzer aus der Benutzertabelle, wenn der Datenverteiler die Benutzerrechte prfen soll. Wenn die interne Referenz eines Benutzers 0 ist, dann
	 * wird die Benutzerinformation aus der Tabelle enfernt.
	 *
	 * @param userId BenutzerID
	 */
	final synchronized void removeUser(long userId) {
		if(_connectionsManager.getServerDavParameters().isUserRightsCheckingEnabled()) {
			Long key = new Long(userId);
			UserInfo user = _userTable.get(key);
			if(user != null) {
				user.decrementReference();
				if(user.canBeSafelyDeleted()) {
					user.disable();
					_userTable.remove(key);
				}
			}
		}
	}

	/**
	 * berprft, ob die Applikation oder der Datenverteiler mit der Id berechtigt ist diese Aktion durchzufhren.
	 *
	 * @param userId die Id der Interressanten (Applikation oder Datenverteiler)
	 * @param info   Die Basisanmeldeinformationen
	 * @param state  <ul><li>state 0: Als Sender</li> <li>1: Als Empfnger</li> <li>2: Als Quelle</li> <li>3: Als Senke</li></ul>
	 *
	 * @return <ul> <li>1: Die Aktion ist berechtigt</li> <li>2: Die Aktion ist nicht berechtigt</li> </ul>
	 */
	final byte isActionAllowed(long userId, BaseSubscriptionInfo info, byte state) {
		if(!_connectionsManager.getServerDavParameters().isUserRightsCheckingEnabled()) {
			return 1;
		}
		// Konfiguration oder Parametrierung
		if(userId == 0) {
			return 1;
		}
		// Interne Applikation der Datenverteiler beim lokalen Betrieb
		else if(userId == _connectionsManager.getTransmitterId()) {
			return 1;
		}
		else {
			// Applikation darf Konfigurationsanfragen als Sender anmelden
			if(AttributeGroupUsageIdentifications.isConfigurationRequest(info.getUsageIdentification())) {
				if(state == 0) {
					return 1;
				}
				else {
					return 2;
				}
			}

			// Applikation darf Konfigurationsantworten als Senke anmelden
			if(AttributeGroupUsageIdentifications.isConfigurationRequest(info.getUsageIdentification())) {
				if(state == 3) {
					return 1;
				}
				else {
					return 2;
				}
			}

			// Andere Anmeldungen
			UserInfo userInfo = (UserInfo)_userTable.get(new Long(userId));
			if(userInfo == null) {
				return 2;
			}
			return userInfo.authentify(info, state) ? (byte)1 : (byte)2;
		}
	}

	/**
	 * Diese Methode wird fr automatisierte JUnit-Tests bentigt und gibt zu einer Benutzer Id die UserInfo zurck.
	 *
	 * @param userId Id, die das Objekt referenziert, das einen Benutzer darstellt.
	 *
	 * @return UserInfo des Benutzer oder <code>null</code>, falls zu der Id kein Objekt gefunden werden konnte.
	 */
	public UserInfo getUserInfo(final long userId) {
		return _userTable.get(new Long(userId));
	}
}
