/*
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2004 by Kappich+Kni Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.dav.dav.
 * 
 * de.bsvrz.dav.dav is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dav.dav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.dav.dav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.dav.dav.main;

import de.bsvrz.dav.daf.communication.dataRepresentation.datavalue.DataValue;
import de.bsvrz.dav.daf.communication.dataRepresentation.datavalue.SendDataObject;
import de.bsvrz.dav.daf.communication.dataRepresentation.datavalue.StreamFetcher;
import de.bsvrz.dav.daf.communication.lowLevel.TelegramUtility;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.ApplicationDataTelegram;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.AttributeGroupAspectCombination;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.BaseSubscriptionInfo;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.ReceiveSubscriptionInfo;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.TransmitterDataSubscriptionReceipt;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.TransmitterDataTelegram;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.TransmitterSubscriptionsConstants;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.impl.CommunicationConstant;
import de.bsvrz.dav.dav.communication.appProtocol.T_A_HighLevelCommunication;
import de.bsvrz.dav.dav.communication.davProtocol.T_T_HighLevelCommunication;
import de.bsvrz.sys.funclib.debug.Debug;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;

/**
 * Diese Komponente ist fr die Verwaltung der Anmeldungen zustndig. Fr jede Anmeldung wird ein Reprsentant erzeugt, der in einer Tabelle gehalten wird.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 6343 $
 */
public class SubscriptionsManager {

	private static final Debug _debug = Debug.getLogger();

	/** Die Tabelle, wo die registrierten Applikationen oder Datenverteiler gehalten werden */
	private final Hashtable _subscriptionsTable;

	/** Die Tabelle der Anmeldungsantrge, wo drin steht was empfangen wurde und was raus geschickt wurde. */
	private final Hashtable _inOutSubscriptionsTable;

	/** Die Tabelle in der der Index der Datenstze verwaltet werden */
	private final Hashtable _dataIndexTable;

	/** Die Tabelle in der der Index der TeilDatenstze verwaltet werden */
	private final Hashtable _usedDataIndexTable;

	/** Die verbindungsverwaltung */
	private final ConnectionsManager _connectionsManager;

	/** Der Zwichenspeicher der zerstckelten Telegramme */
	private final SplittedTransmitterTelegramsTable _splittedTelegramsTable;

	/**
	 * Erzeugt ein neues Objekt mit den gegebenen Parametern.
	 *
	 * @param connectionsManager Verbindungsverwaltung
	 */
	SubscriptionsManager(ConnectionsManager connectionsManager) {
		_connectionsManager = connectionsManager;
		_subscriptionsTable = new Hashtable();
		_inOutSubscriptionsTable = new Hashtable();
		_dataIndexTable = new Hashtable();
		_usedDataIndexTable = new Hashtable();
		_splittedTelegramsTable = new SplittedTransmitterTelegramsTable();
	}


	/** Dieser Konstruktor wird fr Testzwecke bentigt. */
	public SubscriptionsManager() {
		_connectionsManager = null;
		_subscriptionsTable = null;
		_inOutSubscriptionsTable = null;
		_dataIndexTable = null;
		_usedDataIndexTable = null;
		_splittedTelegramsTable = null;
	}

	/**
	 * Anmelden einer Anmeldekomponente
	 *
	 * @param subscriptionComponent Die Anmeldekomponente
	 */
	public void subscribe(SubscriptionComponent subscriptionComponent) {
		if(subscriptionComponent == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		if(_subscriptionsTable == null) {
			throw new IllegalArgumentException("Anmeldemanager ist nicht fertig initiallisiert");
		}
		ServerHighLevelCommunication connection = subscriptionComponent.getConnection();
		if(connection != null) {
			Long key = new Long(connection.getId());
			_subscriptionsTable.put(key, subscriptionComponent);
		}
	}

	/**
	 * Abmelden einer Anmeldekomponente
	 *
	 * @param subscriptionComponent Die Anmeldekomponente
	 */
	public final void unsubscribe(SubscriptionComponent subscriptionComponent) {
		if(subscriptionComponent == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		if(_subscriptionsTable == null) {
			throw new IllegalArgumentException("Anmeldemanager ist nicht fertig initiallisiert");
		}
		ServerHighLevelCommunication connection = subscriptionComponent.getConnection();
		if(connection != null) {
			Long key = new Long(connection.getId());
			_subscriptionsTable.remove(key);
		}
	}

	/**
	 * Gibt die Quelle der Applikationsverbindung zurck.
	 *
	 * @param info Anmeldeinformationen
	 *
	 * @return die Quelle der Applikationsverbindung
	 */
	final T_A_SubscriptionComponent getSendingComponent(BaseSubscriptionInfo info) {
		if(_subscriptionsTable == null) {
			throw new IllegalArgumentException("Anmeldemanager ist nicht fertig initiallisiert");
		}
		synchronized(_subscriptionsTable) {
			ArrayList list = new ArrayList(_subscriptionsTable.values());
			for(int i = 0; i < list.size(); ++i) {
				SubscriptionComponent subscriptionComponent = (SubscriptionComponent)list.get(i);
				if(subscriptionComponent != null) {
					if(subscriptionComponent.getType() == SubscriptionComponent.T_A) {
						T_A_SubscriptionComponent _subscriptionComponent = (T_A_SubscriptionComponent)subscriptionComponent;
						if(_subscriptionComponent.canSend(info)) {
							return _subscriptionComponent;
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * Gibt die Senke fr die bergene Anmeldeinformationen zurck.
	 *
	 * @param info Anmeldeinformationen
	 *
	 * @return Senke oder <code>null</code>, falls nicht vorhanden
	 */
	final T_A_SubscriptionComponent getReceivingComponent(BaseSubscriptionInfo info) {
		if(_subscriptionsTable == null) {
			throw new IllegalArgumentException("Anmeldemanager ist nicht fertig initiallisiert");
		}
		synchronized(_subscriptionsTable) {
			ArrayList list = new ArrayList(_subscriptionsTable.values());
			for(int i = 0; i < list.size(); ++i) {
				SubscriptionComponent subscriptionComponent = (SubscriptionComponent)list.get(i);
				if(subscriptionComponent != null) {
					if(subscriptionComponent.getType() == SubscriptionComponent.T_A) {
						T_A_SubscriptionComponent _subscriptionComponent = (T_A_SubscriptionComponent)subscriptionComponent;
						if(_subscriptionComponent.canReceive(info)) {
							return _subscriptionComponent;
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * berprft ob eine hnliche Anmeldung schon positiv abgeschlossen und vorhanden ist.
	 *
	 * @param key               Basisanmeldeinformationen
	 * @param subscriptionState Status der Anmeldung
	 *
	 * @return Eintrag mit den Anmeldungsinformationen
	 */
	final InAndOutSubscription isSuccessfullySubscribed(BaseSubscriptionInfo key, byte subscriptionState) {
		if(key == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						if((subscription.subscriptionState == subscriptionState) && (subscription.receip == TransmitterSubscriptionsConstants.POSITIV_RECEIP)
						   && (subscription.outSubscriptions != null)) {
							return subscription;
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * berprft ob eine hnliche Anmeldung schon positive abgeschlossen und vorhanden ist.
	 *
	 * @param key               Basisanmeldeinformationen
	 * @param subscriptionState Status der Anmeldung
	 * @param transmitterId     Id des DAV
	 * @param ids               die erreichbaren DAV, ausgehend vom lokalen DAV.
	 *
	 * @return Eintrag mit den Anmeldungsinformationen
	 */
	final InAndOutSubscription isSuccessfullySubscribed(
			BaseSubscriptionInfo key, byte subscriptionState, long transmitterId, long ids[]
	) {
		if(key == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						if((subscription.subscriptionState == subscriptionState) && (subscription.receip == TransmitterSubscriptionsConstants.POSITIV_RECEIP)
						   && (subscription.outSubscriptions != null)) {
							for(int j = 0; j < ids.length; ++j) {
								if(transmitterId == ids[j]) {
									continue;
								}
								if(subscription.mainTransmitterId == ids[j]) {
									return subscription;
								}
							}
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * berprft ob eine hnliche Anmeldung gerade bearbeitet wird.
	 *
	 * @param key               Basisanmeldeinformationen
	 * @param subscriptionState Status der Anmeldung
	 *
	 * @return Eintrag mit den Anmeldungsinformationen
	 */
	final boolean isUnsuccessfullySubscribed(BaseSubscriptionInfo key, byte subscriptionState) {
		if(key == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						if((subscription.subscriptionState == subscriptionState) && (subscription.receip == TransmitterSubscriptionsConstants.NEGATIV_RECEIP)
						   && (subscription.outSubscriptions != null)) {
							return true;
						}
					}
				}
			}
		}
		return false;
	}

	/**
	 * berprft ob eine hnliche Anmeldung gerade bearbeitet wird.
	 *
	 * @param key               Basisanmeldeinformationen
	 * @param subscriptionState Status der Anmeldung
	 * @param transmitterId     Id des DAV
	 * @param ids               die erreichbaren DAV, ausgehend vom lokalen DAV
	 *
	 * @return true: hnliche Anmeldung gefunden false: keine Ahnliche Anmeldung gefunden
	 */
	final boolean isUnsuccessfullySubscribed(
			BaseSubscriptionInfo key, byte subscriptionState, long transmitterId, long ids[]
	) {
		if((key == null) || (ids == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						if((subscription.subscriptionState == subscriptionState) && (subscription.receip == TransmitterSubscriptionsConstants.NEGATIV_RECEIP)
						   && (subscription.outSubscriptions != null) && (subscription.actualTransmitterToConsider != null)) {
							boolean subSet = true;
							for(int j = 0; j < ids.length; ++j) {
								if(transmitterId == ids[j]) {
									continue;
								}
								boolean found = false;
								for(int k = 0; k < subscription.actualTransmitterToConsider.length; ++k) {
									if(ids[j] == subscription.actualTransmitterToConsider[k]) {
										found = true;
										break;
									}
								}
								if(!found) {
									subSet = false;
									break;
								}
							}
							if(subSet) {
								return true;
							}
						}
					}
				}
			}
		}
		return false;
	}

	/**
	 * berprft ob eine hnliche Anmeldung gerade bearbeitet wird
	 *
	 * @param transmitterId Id des DAV
	 * @param subscription  ein Vermerk in der Verwaltungstabelle fr Anmeldungsinformationen
	 *
	 * @return true: hnliche Anmeldung gefunden false: keine Ahnliche Anmeldung gefunden
	 */
	final boolean needToSendSubscription(long transmitterId, InAndOutSubscription subscription) {
		if(subscription == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		long ids[] = subscription.actualTransmitterToConsider;
		if(ids == null) {
			return false;
		}
		BaseSubscriptionInfo key = subscription.baseSubscriptionInfo;
		byte subscriptionState = subscription.subscriptionState;
		if(key == null) {
			throw new IllegalStateException("Argument ist inkonsistent");
		}

		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				int size = list.size();
				for(int i = 0; i < size; ++i) {
					InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
					if(_subscription != null) {
						if(_subscription == subscription) {
							continue;
						}
						if((_subscription.subscriptionState == subscriptionState) && (_subscription.outSubscriptions != null) && (_subscription
								.actualTransmitterToConsider != null)) {
							if(_subscription.actualTransmitterToConsider.length == ids.length) {
								boolean subSet = true;
								for(int j = 0; j < ids.length; ++j) {
									boolean found = false;
									for(int k = 0; k < _subscription.actualTransmitterToConsider.length; ++k) {
										if(ids[j] == _subscription.actualTransmitterToConsider[k]) {
											found = true;
											break;
										}
									}
									if(!found) {
										subSet = false;
										break;
									}
								}
								if(subSet) {
									return false;
								}
							}
						}
					}
				}
			}
		}
		return true;
	}

	/**
	 * Addiert einen Anmeldeantrag in der Tabelle.
	 *
	 * @param subscription  ein Vermerk in der Verwaltungstabelle fr Anmeldungsinformationen
	 * @param transmitterId Id des DAV
	 *
	 * @return
	 */
	final InAndOutSubscription addSubscriptionRequest(InAndOutSubscription subscription, long transmitterId) {
		BaseSubscriptionInfo key = subscription.baseSubscriptionInfo;
		if(key != null) {
			ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
			if(list == null) {
				list = new ArrayList();
				_inOutSubscriptionsTable.put(key, list);
			}
			synchronized(list) {
				if(subscription.isTransmitterConnection) {
					for(int i = list.size() - 1; i > -1; --i) {
						InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
						if((_subscription != null) && (_subscription.sourceTransmitterId == subscription.sourceTransmitterId) && (_subscription
								.subscriptionState == subscription.subscriptionState)) {
							if((_subscription.transmitterToConsider != null) && (subscription.transmitterToConsider != null)) {
								if(_subscription.transmitterToConsider.length == subscription.transmitterToConsider.length) {
									boolean subSet = true;
									for(int k = 0; k < _subscription.transmitterToConsider.length; ++k) {
										boolean found = false;
										for(int h = 0; h < subscription.transmitterToConsider.length; ++h) {
											if(subscription.transmitterToConsider[h] == _subscription.transmitterToConsider[k]) {
												found = true;
												break;
											}
										}
										if(!found) {
											subSet = false;
											break;
										}
									}
									if(subSet) {
										_subscription.count = _subscription.count + 1;
										return _subscription;
									}
								}
							}
						}
					}
				}
				else {
					for(int i = list.size() - 1; i > -1; --i) {
						InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
						if((_subscription != null) && (_subscription.sourceTransmitterId == subscription.sourceTransmitterId) && (_subscription
								.subscriptionState == subscription.subscriptionState)) {
							list.set(i, subscription);
							return _subscription;
						}
					}
				}
				list.add(subscription);
				return subscription;
			}
		}
		return null;
	}

	/**
	 * @param key               Basisanmeldeinformationen
	 * @param subscriptionState Status der Anmeldung
	 * @param sourceTransmitter Ausgangs - DAV
	 * @param ids               die erreichbaren DAV, ausgehend vom lokalen DAV.
	 *
	 * @return Eintrag mit der Anmeldungsinformationen
	 */
	final InAndOutSubscription[] getSameSubscriptionPostedTo(
			BaseSubscriptionInfo key, byte subscriptionState, long sourceTransmitter, long ids[]
	) {
		if((key == null) || (ids == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList result = new ArrayList();
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				int size = list.size();
				for(int i = 0; i < size; ++i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						if((subscription.outSubscriptions != null) && (subscription.subscriptionState == subscriptionState)) {
							for(int j = 0; j < subscription.outSubscriptions.length; ++j) {
								if((subscription.outSubscriptions[j] != null) && (sourceTransmitter == subscription.outSubscriptions[j].targetTransmitterId)) {
									if(subscription.outSubscriptions[j].transmitterToConsider != null) {
										if(subscription.outSubscriptions[j].transmitterToConsider.length == ids.length) {
											boolean subSet = true;
											for(int k = 0; k < ids.length; ++k) {
												boolean found = false;
												for(int h = 0; h < subscription.outSubscriptions[j].transmitterToConsider.length; ++h) {
													if(subscription.outSubscriptions[j].transmitterToConsider[h] == ids[k]) {
														found = true;
														break;
													}
												}
												if(!found) {
													subSet = false;
													break;
												}
											}
											if(subSet) {
												result.add(subscription);
												break;
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		int size = result.size();
		if(size == 0) {
			return null;
		}
		else {
			InAndOutSubscription array[] = new InAndOutSubscription[size];
			return (InAndOutSubscription[])result.toArray(array);
		}
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um einen Anmeldungsvermerk zu quittieren. Anhand der Basisanmeldeinformation in receipt wird
	 * die Menge der Anmeldungsvermerke bestimmt, die fr diese Basisanmeldeinformation bei dieser Komponente registriert sind. Danach wird in dieser Liste der
	 * Vermerk gesucht, der den gleichen Anmeldestatus und die gleichen zu bercksichtigenden Datenverteiler hat. Die Folgeanmeldung aus diesem Anmeldungsvermerk
	 * bei dem Datenverteiler mit der in targetId angegebenen Datenverteiler-ID wird mit dem Quittierzustand aus receipt versehen, und der Anmeldungsvermerk wird
	 * von der Methode zurckgegeben. Wenn kein Vermerk existiert, wird Null zurckgegeben.
	 *
	 * @param targetId Ziel - DAV
	 * @param receipt  Quittung der Anmeldung
	 *
	 * @return Eintrag mit der Anmeldungsinformationen
	 */
	final InAndOutSubscription[] getSubscriptionRequestToUpdate(long targetId, TransmitterDataSubscriptionReceipt receipt) {
		ArrayList result = new ArrayList();
		BaseSubscriptionInfo key = receipt.getBaseSubscriptionInfo();
		if(key != null) {
			ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
			if(list != null) {
				synchronized(list) {
					long transmitterIds[] = receipt.getTransmitters();
					if(transmitterIds != null) {
						byte subscriptionState = receipt.getSubscriptionState();
						int size = list.size();
						for(int i = 0; i < size; ++i) {
							InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
							if(subscription != null) {
								if((subscription.outSubscriptions != null) && (subscription.subscriptionState == subscriptionState)) {
									InAndOutSubscription.OutSubscription outSubscription = null;
									for(int j = 0; j < subscription.outSubscriptions.length; ++j) {
										if((subscription.outSubscriptions[j] != null) && (targetId == subscription.outSubscriptions[j].targetTransmitterId)) {
											outSubscription = subscription.outSubscriptions[j];
											break;
										}
									}
									if((outSubscription != null) && (outSubscription.transmitterToConsider != null)) {
										if(outSubscription.transmitterToConsider.length == transmitterIds.length) {
											boolean subSet = true;
											for(int k = 0; k < transmitterIds.length; ++k) {
												boolean found = false;
												for(int h = 0; h < outSubscription.transmitterToConsider.length; ++h) {
													if(outSubscription.transmitterToConsider[h] == transmitterIds[k]) {
														found = true;
														break;
													}
												}
												if(!found) {
													subSet = false;
													break;
												}
											}
											if(subSet) {
												if(outSubscription.receip != receipt.getReceipt()) {
													outSubscription.receip = receipt.getReceipt();
													outSubscription.mainTransmitterId = receipt.getMainTransmitterId();
													result.add(subscription);
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		int size = result.size();
		if(size == 0) {
			return null;
		}
		else {
			InAndOutSubscription array[] = new InAndOutSubscription[size];
			return (InAndOutSubscription[])result.toArray(array);
		}
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um einen Anmeldungsvermerk eines Umleitungsantrags zu quittieren. Anhand der Informationen in
	 * receip wird berprft, ob der vorhandene Vermerk die gleiche Basisanmeldeinformation, den gleichen Anmeldestatus und die gleichen zu bercksichtigenden
	 * Datenverteiler hat. Ist dies der Fall, so wird die Folgeanmeldung aus diesem Anmeldungsvermerk bei dem Datenverteiler mit der in targetId angegebenen
	 * Datenverteiler-ID mit dem Quittierzustand aus receip versehen, und true wird von der Methode zurckgegeben. Wenn kein Vermerk existiert, wird false
	 * zurckgegeben.
	 *
	 * @param subscription      Eintrag mit der Anmeldungsinformationen
	 * @param targetId          Ziel - DAV
	 * @param key               Basisanmeldeinformationen
	 * @param transmitterIds    die erreichbaren DAV, ausgehend vom lokalen DAV
	 * @param mainTransmitterId Haupt - DAV
	 * @param subscriptionState Status der Anmeldung
	 * @param receip            Quittunngszustand
	 *
	 * @return true: Es existiert ein Vermerk in der Verbindungsverwaltung, false: Es existiert kein entsprechender Vermerk.
	 */
	final boolean updateExchangeSubscriptionRequest(
			InAndOutSubscription subscription,
			long targetId,
			BaseSubscriptionInfo key,
			long transmitterIds[],
			long mainTransmitterId,
			byte subscriptionState,
			byte receip
	) {
		if((subscription == null) || (key == null) || (transmitterIds == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}
		if(key.equals(subscription.baseSubscriptionInfo)) {
			if(subscription != null) {
				synchronized(subscription) {
					if((subscription.outSubscriptions != null) && (subscription.subscriptionState == subscriptionState)) {
						InAndOutSubscription.OutSubscription outSubscription = null;
						for(int j = 0; j < subscription.outSubscriptions.length; ++j) {
							if((subscription.outSubscriptions[j] != null) && (targetId == subscription.outSubscriptions[j].targetTransmitterId)) {
								outSubscription = subscription.outSubscriptions[j];
								break;
							}
						}
						if((outSubscription != null) && (outSubscription.transmitterToConsider != null)) {
							if(outSubscription.transmitterToConsider.length == transmitterIds.length) {
								boolean subSet = true;
								for(int k = 0; k < transmitterIds.length; ++k) {
									boolean found = false;
									for(int h = 0; h < outSubscription.transmitterToConsider.length; ++h) {
										if(outSubscription.transmitterToConsider[h] == transmitterIds[k]) {
											found = true;
											break;
										}
									}
									if(!found) {
										subSet = false;
										break;
									}
								}
								if(subSet) {
									outSubscription.receip = receip;
									outSubscription.mainTransmitterId = mainTransmitterId;
									return true;
								}
							}
						}
					}
				}
			}
		}
		return false;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um einen Anmeldungsvermerk einer Applikation zu lschen. Zunchst wird die Liste der
	 * Anmeldungsvermerke der angegebenen Basisanmeldeinformation bestimmt. Dann werden alle Vermerke aus der Liste entfernt, die von der Applikation applicationId
	 * stammen und den gleichen Anmeldestatus wie subscriptionState haben.
	 *
	 * @param applicationId     Applikation
	 * @param key               Basisanmeldeinformation
	 * @param subscriptionState Anmeldestatus
	 *
	 * @return aktualisierter Eintrag der Anmeldungsinformationen
	 */
	final InAndOutSubscription removeSubscriptionRequest(
			long applicationId, BaseSubscriptionInfo key, byte subscriptionState
	) {
		if(key == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						if((subscription.sourceTransmitterId == applicationId) && (subscription.subscriptionState == subscriptionState)) {
							return (InAndOutSubscription)list.remove(i);
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um einen Anmeldungsvermerk eines Datenverteilers zu lschen. Zunchst wird die Liste der
	 * Anmeldungsvermerke mit der angegebenen Basisanmeldeinformation bestimmt. Dann werden alle Vermerke aus der Liste entfernt, die vom Datenverteiler
	 * transmitterId stammen, den gleichen Anmeldestatus wie subscriptionState haben und sich auch auf einen der in transmitterIds angegebenen Datenverteiler
	 * beziehen.
	 *
	 * @param sourceTransmitterId Ausgangs - DAV
	 * @param key                 Basisanmeldeinformation
	 * @param subscriptionState   Anmeldestatus
	 * @param transmitterIds      die erreichbaren DAV, ausgehend vom lokalen DAV
	 * @param transmitterId       Id des DAV
	 *
	 * @return aktualisierter Eintrag der Anmeldungsinformationen
	 */
	final InAndOutSubscription removeSubscriptionRequest(
			long sourceTransmitterId, BaseSubscriptionInfo key, byte subscriptionState, long transmitterIds[], long transmitterId
	) {
		if((key == null) || (transmitterIds == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				int size = list.size();
				for(int i = 0; i < size; ++i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						if(subscription.transmitterToConsider != null) {
							if((subscription.sourceTransmitterId == sourceTransmitterId) && (subscription.subscriptionState == subscriptionState)) {
								if(transmitterIds.length == subscription.transmitterToConsider.length) {
									boolean subSet = true;
									for(int k = 0; k < transmitterIds.length; ++k) {
										boolean found = false;
										for(int h = 0; h < subscription.transmitterToConsider.length; ++h) {
											if(subscription.transmitterToConsider[h] == transmitterIds[k]) {
												found = true;
												break;
											}
										}
										if(!found) {
											subSet = false;
											break;
										}
									}
									if(subSet) {
										subscription.count = subscription.count - 1;
										if(subscription.count <= 0) {
											return (InAndOutSubscription)list.remove(i);
										}
									}
								}
							}
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um einen Anmeldungsvermerk aus der Verwaltung zu lschen. Zunchst wird die Liste der
	 * Anmeldungsvermerke mit der angegebenen Basisanmeldeinformation bestimmt. Dann werden alle Vermerke subscription aus der Liste entfernt. Dieser Wert wird von
	 * der Methode auch zurckgegeben. Existiert der Vermerk nicht, so wird Null zurckgegeben.
	 *
	 * @param subscription Eintrag mit der Anmeldungsinformation
	 *
	 * @return aktualisierter Eintrag der Anmeldungsinformationen
	 */
	final InAndOutSubscription removeSubscriptionRequest(InAndOutSubscription subscription) {
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(subscription.baseSubscriptionInfo);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
					if(_subscription == subscription) {
						return (InAndOutSubscription)list.remove(i);
					}
				}
			}
		}
		return null;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, wenn eine Verbindung nicht mehr vorhanden ist. Alle Anmeldungsvermerke mit der sourceId als
	 * Anmeldeinstanz werden aus der Verwaltung entfernt. Dieser Wert wird von der Methode auch zurckgegeben.
	 *
	 * @param sourceId Ausgangs - DAV
	 *
	 * @return aktualisierter Eintrag der Anmeldungsinformationen
	 */
	final InAndOutSubscription[] removeSubscriptionRequests(long sourceId) {
		ArrayList listList = new ArrayList(_inOutSubscriptionsTable.values());
		ArrayList requests = new ArrayList();
		for(int i = listList.size() - 1; i > -1; --i) {
			ArrayList list = (ArrayList)listList.get(i);
			if(list != null) {
				synchronized(list) {
					InAndOutSubscription subscription = null;
					for(int j = list.size() - 1; j > -1; --j) {
						subscription = (InAndOutSubscription)list.get(j);
						if((subscription != null) && (subscription.sourceTransmitterId == sourceId)) {
							list.remove(j);
							requests.add(subscription);
						}
					}
					if((list.size() == 0) && (subscription != null)) {
						_inOutSubscriptionsTable.remove(subscription.baseSubscriptionInfo);
					}
				}
			}
		}
		int size = requests.size();
		if(size == 0) {
			return null;
		}
		else {
			InAndOutSubscription retValue[] = new InAndOutSubscription[size];
			for(int i = 0; i < size; ++i) {
				retValue[i] = (InAndOutSubscription)requests.get(i);
			}
			return retValue;
		}
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um zu berprfen, ob fr die bergebene Basisanmeldeinformation ein Anmeldungsvermerk vorliegt.
	 * Wenn mindestens ein Anmeldungsvermerk existiert, dann wird true zurckgegeben, sonst false.
	 *
	 * @param key Basisanmeldeinformationen
	 *
	 * @return true: es existiert mind. ein Anmeldungsvermerk, false: es existiert keine Anmeldungsvermerk.
	 */
	final boolean isSubscriptionRequestAvaillable(BaseSubscriptionInfo key) {
		if(key == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if((list != null) && (list.size() > 0)) {
			return true;
		}
		return false;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um zu berprfen, ob ein positiv quittierter Anmeldungsvermerk fr die gleiche
	 * Basisanmeldeinformation und den gleichen Anmeldstatus vorliegt, wie in subscription bergeben. Ist dies der Fall, so wird true zurckgegeben, sonst false.
	 *
	 * @param subscription Eintrag mit der Anmeldungsinformation
	 *
	 * @return true: es existiert ein Anmeldungsvermerk, false: es existiert keine Anmeldungsvermerk.
	 */
	final boolean isSubscriptionRequestAvaillable(InAndOutSubscription subscription) {
		if((subscription == null) || (subscription.baseSubscriptionInfo == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(subscription.baseSubscriptionInfo);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
					if(_subscription != null) {
						if(subscription == _subscription) {
							continue;
						}
						if((_subscription.sourceTransmitterId == subscription.sourceTransmitterId)
						   && (_subscription.subscriptionState == subscription.subscriptionState) && (_subscription.receip == TransmitterSubscriptionsConstants
								.POSITIV_RECEIP)) {
							return true;
						}
					}
				}
			}
		}
		return false;
	}


	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um zu berprfen, ob ein positiv quittierter Anmeldungsvermerk fr die gleiche
	 * Basisanmeldeinformation und den gleichen Anmeldstatus vorliegt, wie in subscription bergeben. Auerdem darf sie keine Folgeanmeldungen besitzen. Ist dies
	 * der Fall, so wird der Vermerk zurckgegeben, sonst wird Null zurckgegeben.
	 *
	 * @param subscription Anmeldung
	 *
	 * @return falls vorhanden, wird die Anmmeldung zurck gegeben, sonst NULL
	 */
	final InAndOutSubscription getAlternativeSubscriptionRequest(InAndOutSubscription subscription) {
		if(subscription == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		BaseSubscriptionInfo key = subscription.baseSubscriptionInfo;
		if(key == null) {
			throw new IllegalArgumentException("Argument ist Inkonsistent");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
					if(_subscription != null) {
						if(_subscription == subscription) {
							continue;
						}
						else if(_subscription.outSubscriptions != null) {
							continue;
						}
						else if((_subscription.subscriptionState == subscription.subscriptionState) && (_subscription.receip == subscription.receip) && (
								_subscription.mainTransmitterId == subscription.mainTransmitterId)) {
							return _subscription;
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um zu berprfen, welche Anmeldungsvermerke von subscription abhngen. Dazu wird zunchst die
	 * Liste der Anmeldungsvermerke bestimmt, die die gleiche Basisanmeldeinformation wie subscription haben. Fr die positiv quittierten aus dieser Liste wird
	 * bestimmt, welche zustzlich den gleichen Anmeldestatus und eine leere Folgeanmeldungsliste besitzen. Alle Vermerke, die diese Bedingungen erfllen, werden
	 * in einem Feld zurckgegeben. Wenn kein solcher Vermerk existiert, wird Null zurckgegeben.
	 *
	 * @param subscription Anmeldung
	 *
	 * @return Feld der vom bergabeparameter abhngigen Anmeldungen.
	 */
	final InAndOutSubscription[] getInvolvedPositiveSubscriptionRequests(InAndOutSubscription subscription) {
		if(subscription == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		BaseSubscriptionInfo key = subscription.baseSubscriptionInfo;
		if(key == null) {
			throw new IllegalArgumentException("Argument ist Inkonsistent");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			ArrayList array = new ArrayList();
			synchronized(list) {
				int size = list.size();
				for(int i = 0; i < size; ++i) {
					InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
					if(_subscription != null) {
						if(_subscription == subscription) {
							array.add(_subscription);
						}
						else if(_subscription.outSubscriptions != null) {
							continue;
						}
						else if((_subscription.subscriptionState == subscription.subscriptionState) && (_subscription.receip
						                                                                                == TransmitterSubscriptionsConstants.POSITIV_RECEIP)) {
							array.add(_subscription);
						}
					}
				}
			}
			int size = array.size();
			if(size > 0) {
				InAndOutSubscription retValue[] = new InAndOutSubscription[size];
				for(int i = 0; i < size; ++i) {
					retValue[i] = (InAndOutSubscription)array.get(i);
				}
				return retValue;
			}
		}
		return null;
	}


	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um zu berprfen, welche Anmeldungsvermerke von subscription abhngen. Dazu wird zunchst die
	 * Liste der Anmeldungsvermerke bestimmt, die die gleiche Basisanmeldeinformation wie subscription haben. Fr die positiv quittierten aus dieser Liste wird
	 * bestimmt, welche zustzlich den gleichen Anmeldestatus und eine leere Folgeanmeldungsliste besitzen. Alle Vermerke, die diese Bedingungen erfllen, werden
	 * in einem Feld zurckgegeben. Wenn kein solcher Vermerk existiert, wird Null zurckgegeben.<br>Ein Anmeldeantrag A ist abhngig von B wenn:<br>
	 * -Basisanmeldeinformation von A ist gleich von B<br> -Anmeldungsart von A ist gleich von B (Sender oder empfnger)<br>
	 *
	 * @param subscription Anmeldung
	 *
	 * @return vom bergabeparameter abhngige Anmeldungsvermerke
	 */
	final InAndOutSubscription[] getInvolvedSubscriptionRequests(InAndOutSubscription subscription) {
		if(subscription == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		BaseSubscriptionInfo key = subscription.baseSubscriptionInfo;
		if(key == null) {
			throw new IllegalArgumentException("Argument ist Inkonsistent");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			ArrayList array = new ArrayList();
			synchronized(list) {
				int size = list.size();
				for(int i = 0; i < size; ++i) {
					InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
					if(_subscription != null) {
						if(_subscription == subscription) {
							array.add(_subscription);
						}
						else if(_subscription.outSubscriptions != null) {
							continue;
						}
						else if(_subscription.subscriptionState == subscription.subscriptionState) {
							array.add(_subscription);
						}
					}
				}
			}
			int size = array.size();
			if(size > 0) {
				InAndOutSubscription retValue[] = new InAndOutSubscription[size];
				for(int i = 0; i < size; ++i) {
					retValue[i] = (InAndOutSubscription)array.get(i);
				}
				return retValue;
			}
		}
		return null;
	}

	/**
	 * Gibt die Anmeldeantrage zurck, die vom bergebenen Anmeldeantrag abhngig sein knnten, zurck. <br>Ein Anmeldeantrag A ist abhngig von B wenn:<br>
	 * -Basisanmeldeinformation von A ist gleich von B<br>-Anmeldungsart von A ist gleich von B (Sender oder empfnger) <br>-Zu berucksichtigenden Datenverteiler
	 * von A sind gleich von B
	 */
	final InAndOutSubscription[] getInvolvedSubscriptionRequests(InAndOutSubscription subscription, boolean checkIds) {
		if(subscription == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		BaseSubscriptionInfo key = subscription.baseSubscriptionInfo;
		if(key == null) {
			throw new IllegalArgumentException("Argument ist Inkonsistent");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			ArrayList array = new ArrayList();
			synchronized(list) {
				int size = list.size();
				for(int i = 0; i < size; ++i) {
					InAndOutSubscription _subscription = (InAndOutSubscription)list.get(i);
					if(_subscription != null) {
						if(_subscription == subscription) {
							array.add(_subscription);
						}
						else if(_subscription.outSubscriptions != null) {
							continue;
						}
						else if(_subscription.subscriptionState == subscription.subscriptionState) {
							if((_subscription.actualTransmitterToConsider != null) && (subscription.actualTransmitterToConsider != null)) {
								if(_subscription.actualTransmitterToConsider.length == subscription.actualTransmitterToConsider.length) {
									boolean subSet = true;
									for(int k = 0; k < _subscription.actualTransmitterToConsider.length; ++k) {
										boolean found = false;
										for(int h = 0; h < subscription.actualTransmitterToConsider.length; ++h) {
											if(subscription.actualTransmitterToConsider[h] == _subscription.actualTransmitterToConsider[k]) {
												found = true;
												break;
											}
										}
										if(!found) {
											subSet = false;
											break;
										}
									}
									if(subSet) {
										array.add(_subscription);
									}
								}
							}
						}
					}
				}
			}
			int size = array.size();
			if(size > 0) {
				InAndOutSubscription retValue[] = new InAndOutSubscription[size];
				for(int i = 0; i < size; ++i) {
					retValue[i] = (InAndOutSubscription)array.get(i);
				}
				return retValue;
			}
		}
		return null;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, um zu berprfen, welche Anmeldungsvermerke von key abhngen. Dazu wird die Liste der
	 * Anmeldungsvermerke bestimmt, die die gleiche Basisanmeldeinformation wie key haben und in einem Feld zurckgegeben. Wenn kein solcher Vermerk existiert,
	 * wird Null zurckgegeben.
	 *
	 * @param key Basisanmeldeinformationen
	 *
	 * @return Anmeldeantrage, die vom gegebenen Anmeldeinformationen abhngig sein knnten
	 */
	final InAndOutSubscription[] getInvolvedSubscriptionRequests(BaseSubscriptionInfo key) {
		if(key == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			ArrayList array = new ArrayList();
			synchronized(list) {
				int size = list.size();
				for(int i = 0; i < size; ++i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						array.add(subscription);
					}
				}
			}
			int size = array.size();
			if(size > 0) {
				InAndOutSubscription retValue[] = new InAndOutSubscription[size];
				for(int i = 0; i < size; ++i) {
					retValue[i] = (InAndOutSubscription)array.get(i);
				}
				return retValue;
			}
		}
		return null;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, wenn neue Zuliefererinformationen eingetroffen sind.  Es werden alle nicht positiv quittierten
	 * Anmeldungsvermerke darauf berprft, ob die neuen Objekt-IDs oder Attributgruppen-Aspekt-Kombinationen sie betreffen. Dazu werden die Basisanmeldeinformation der
	 * Vermerke mit den ids bzw. combinations verglichen. Ein Feld mit den betroffenen Vermerken wird gebildet und zurckgegeben. Wenn kein solcher Vermerk
	 * existiert, wird Null zurckgegeben.
	 *
	 * @param ids          Die Objekt-Ids
	 * @param combinations Die Attributgruppen-Aspekt-Kombinationen
	 *
	 * @return Feld mit den betroffenen Kombinationen
	 */
	final InAndOutSubscription[] getInvolvedSubscriptionRequests(
			long ids[], AttributeGroupAspectCombination combinations[]
	) {
		if((ids == null) && (combinations == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}

		ArrayList listList = new ArrayList(_inOutSubscriptionsTable.values());
		ArrayList requests = new ArrayList();
		for(int i = listList.size() - 1; i > -1; --i) {
			ArrayList list = (ArrayList)listList.get(i);
			if(list != null) {
				synchronized(list) {
					int size = list.size();
					for(int j = 0; j < size; ++j) {
						InAndOutSubscription subscription = (InAndOutSubscription)list.get(j);
						if(subscription != null) {
							if(subscription.receip == TransmitterSubscriptionsConstants.POSITIV_RECEIP) {
								continue;
							}
							BaseSubscriptionInfo info = subscription.baseSubscriptionInfo;
							if(info != null) {
								if(ids != null) {
									long objectId = info.getObjectID();
									boolean found = false;
									for(int k = 0; k < ids.length; ++k) {
										if(ids[k] == objectId) {
											found = true;
											break;
										}
									}
									if(found) {
										requests.add(subscription);
										continue;
									}
								}
								if(combinations != null) {
									long atgUsageIdentification = info.getUsageIdentification();
									for(int k = 0; k < combinations.length; ++k) {
										if(combinations[k].getAtgUsageIdentification() == atgUsageIdentification) {
											requests.add(subscription);
											break;
										}
									}
								}
							}
						}
					}
				}
			}
		}
		int size = requests.size();
		if(size == 0) {
			return null;
		}
		else {
			InAndOutSubscription retValue[] = new InAndOutSubscription[size];
			for(int i = 0; i < size; ++i) {
				retValue[i] = (InAndOutSubscription)requests.get(i);
			}
			return retValue;
		}
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, wenn neue Zuliefererinformationen eingetroffen sind.  Es werden alle nicht positiv quittierten
	 * Anmeldungsvermerke darauf berprft, ob die neuen Objekt-IDs oder Attributgruppen-Aspekt-Kombinationen sie betreffen. Dazu werden die Basisanmeldeinformation der
	 * Vermerke mit den ids bzw. combinations verglichen. Ein Feld mit den betroffenen Vermerken wird gebildet und zurckgegeben. Wenn kein solcher Vermerk
	 * existiert, wird Null zurckgegeben.
	 *
	 * @param ids          Die Objekt-Ids
	 * @param combinations Die Attributgruppen-Aspekt-Kombinationen
	 */
	final InAndOutSubscription[] getInvolvedPositiveSubscriptionRequests(
			long ids[], AttributeGroupAspectCombination combinations[]
	) {
		if((ids == null) && (combinations == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}

		ArrayList listList = new ArrayList(_inOutSubscriptionsTable.values());
		ArrayList requests = new ArrayList();
		for(int i = listList.size() - 1; i > -1; --i) {
			ArrayList list = (ArrayList)listList.get(i);
			if(list != null) {
				synchronized(list) {
					int size = list.size();
					for(int j = 0; j < size; ++j) {
						InAndOutSubscription subscription = (InAndOutSubscription)list.get(j);
						if((subscription != null) && (subscription.outSubscriptions != null) && (subscription.receip == TransmitterSubscriptionsConstants
								.POSITIV_RECEIP)) {
							BaseSubscriptionInfo info = subscription.baseSubscriptionInfo;
							if(info != null) {
								if(ids != null) {
									long objectId = info.getObjectID();
									boolean found = false;
									for(int k = 0; k < ids.length; ++k) {
										if(ids[k] == objectId) {
											found = true;
											break;
										}
									}
									if(found) {
										requests.add(subscription);
										continue;
									}
								}
								if(combinations != null) {
									final long atgUsageIdentification = info.getUsageIdentification();
									for(int k = 0; k < combinations.length; ++k) {
										if(combinations[k].getAtgUsageIdentification() == atgUsageIdentification) {
											requests.add(subscription);
											break;
										}
									}
								}
							}
						}
					}
				}
			}
		}
		int size = requests.size();
		if(size == 0) {
			return null;
		}
		else {
			InAndOutSubscription retValue[] = new InAndOutSubscription[size];
			for(int i = 0; i < size; ++i) {
				retValue[i] = (InAndOutSubscription)requests.get(i);
			}
			return retValue;
		}
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, wenn neue Weginformationen eingetroffen sind. Es werden alle Anmeldungsvermerke gesucht, die
	 * den Datenverteiler transmitterId in ihrer Liste der zu bercksichtigenden Datenverteiler haben. Weiterhin mssen die Anmeldungsvermerke eine Folgeanmeldung
	 * beim Datenverteiler overTransmitterId haben. Diese gefundenen Anmeldungsvermerke werden umgeleitet, da der Datenverteiler transmitterId ber einen neuen Weg
	 * erreichbar ist. Diese Anmeldungsvermerke werden als Feld zurckgegeben. Wenn kein solcher Eintrag existiert, wird Null zurckgegeben.
	 *
	 * @param transmitter       DAV
	 * @param overTransmitterId
	 *
	 * @return alle Anmeldungen, die positiv quittiert sind oder auf eine quittung warten
	 */
	final InAndOutSubscription[] getSubscriptionRequestsToRedirect(long transmitter, long overTransmitterId) {
		ArrayList listList = new ArrayList(_inOutSubscriptionsTable.values());
		ArrayList requests = new ArrayList();
		for(int i = listList.size() - 1; i > -1; --i) {
			ArrayList list = (ArrayList)listList.get(i);
			if(list != null) {
				synchronized(list) {
					int size = list.size();
					for(int j = 0; j < size; ++j) {
						InAndOutSubscription subscription = (InAndOutSubscription)list.get(j);
						if((subscription != null) && (subscription.outSubscriptions != null)) {
							if(subscription.actualTransmitterToConsider != null) {
								boolean found = false;
								for(int k = 0; k < subscription.actualTransmitterToConsider.length; ++k) {
									if(subscription.actualTransmitterToConsider[k] == transmitter) {
										found = true;
										break;
									}
								}
								if(found) {
									for(int k = 0; k < subscription.outSubscriptions.length; ++k) {
										if(subscription.outSubscriptions[k] != null) {
											if(subscription.outSubscriptions[k].targetTransmitterId == overTransmitterId) {
												requests.add(subscription);
												break;
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		int size = requests.size();
		if(size == 0) {
			return null;
		}
		else {
			InAndOutSubscription retValue[] = new InAndOutSubscription[size];
			for(int i = 0; i < size; ++i) {
				retValue[i] = (InAndOutSubscription)requests.get(i);
			}
			return retValue;
		}
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung aufgerufen, wenn eine Verbindung zu einem Datenverteiler aufgehoben wurde. Es werden alle
	 * Anmeldungsvermerke gesucht, die eine Folgeanmeldung bei diesem Datenverteiler targetId besitzen. Ein Feld der betroffenen Anmeldungsvermerke wird erstellt
	 * und zurckgegeben. Wenn kein solcher Vermerk existiert, wird Null zurckgegeben.
	 *
	 * @param targetId DAV mit Folgeanmeldungen
	 *
	 * @return Feld der betroffenen Anmeldungsvermerke wird zurckgegeben. Wenn kein solcher Vermerk existiert, wird Null zurckgegeben.
	 */
	final InAndOutSubscription[] getAffectedOutgoingSubscriptions(long targetId) {
		ArrayList listList = new ArrayList(_inOutSubscriptionsTable.values());
		ArrayList requests = new ArrayList();
		for(int i = listList.size() - 1; i > -1; --i) {
			ArrayList list = (ArrayList)listList.get(i);
			if(list != null) {
				synchronized(list) {
					InAndOutSubscription subscription = null;
					int size = list.size();
					for(int j = 0; j < size; ++j) {
						subscription = (InAndOutSubscription)list.get(j);
						if(subscription != null) {
							if(subscription.outSubscriptions != null) {
								for(int k = subscription.outSubscriptions.length - 1; k > -1; --k) {
									if((subscription.outSubscriptions[k] != null) && (subscription.outSubscriptions[k].targetTransmitterId == targetId)) {
										requests.add(subscription);
										break;
									}
								}
							}
						}
					}
				}
			}
		}
		int size = requests.size();
		if(size == 0) {
			return null;
		}
		else {
			InAndOutSubscription retValue[] = new InAndOutSubscription[size];
			for(int i = 0; i < size; ++i) {
				retValue[i] = (InAndOutSubscription)requests.get(i);
			}
			return retValue;
		}
	}

	final InAndOutSubscription[] getSubscriptionsFromSource(long sourceId) {
		ArrayList listList = new ArrayList(_inOutSubscriptionsTable.values());
		ArrayList requests = new ArrayList();
		for(int i = listList.size() - 1; i > -1; --i) {
			ArrayList list = (ArrayList)listList.get(i);
			if(list != null) {
				synchronized(list) {
					int size = list.size();
					for(int j = 0; j < size; ++j) {
						InAndOutSubscription subscription = (InAndOutSubscription)list.get(j);
						if(subscription != null) {
							if(subscription.sourceTransmitterId == sourceId) {
								requests.add(subscription);
							}
						}
					}
				}
			}
		}
		int size = requests.size();
		if(size == 0) {
			return null;
		}
		else {
			InAndOutSubscription retValue[] = new InAndOutSubscription[size];
			for(int i = 0; i < size; ++i) {
				retValue[i] = (InAndOutSubscription)requests.get(i);
			}
			return retValue;
		}
	}

	final InAndOutSubscription getSubscription(long sourceId, BaseSubscriptionInfo info) {
		if(info == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(info);
		if(list != null) {
			synchronized(list) {
				int size = list.size();
				for(int i = 0; i < size; ++i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if((subscription != null) && (subscription.sourceTransmitterId == sourceId)) {
						return subscription;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Diese Methode wird von der Protokollsteuerung DaV-DAF aufgerufen, wenn von einer Applikation ein neuer Datensatz empfangen wurde. Zuerst wird berprft ob
	 * dieser Datenverteiler der Zentraldatenverteiler der Datenstze, d. h., ob die Quelle oder Senke der Datenstze bei diesen Datenverteiler angemeldet ist. Ist
	 * dies nicht der Fall, so wird berprft, ob ein positiv quittierter Anmeldungsvermerk fr einen Sender vorhanden ist. In diesem Fall wird der Datensatz mit
	 * der ID des Senders versehen und dem zum Datenverteiler weitergeleitet, bei dem eine positiv quittierte Folgeanmeldung vorliegt. Erst der
	 * Zentraldatenverteiler sorgt fr einen gltigen, eindeutigen Datensatzindex. Bis zu diesem Zeitpunkt muss der Sender eindeutig benannt werden. Dies ist
	 * gerade dann notwendig, wenn mehrere Sender ihre Datenstze ber einen Datenverteiler im System verteilen und die Telegramme zerstckelt sind. Wenn dieser
	 * Datenverteiler der Zentraldatenverteiler fr diese Datenstze ist, wird berprft ob der Datensatz zerstckelt ist. In diesem Fall wird der Datensatzindex
	 * dieses Datensatzes bestimmt und an die Applikationen und Datenverteiler weitergeleitet, die positiv quittierte Anmeldungen initiiert haben. Wenn der
	 * Datensatz kein nachgelieferter Datensatz ist, wird er in der Datenbestand aufgenommen. Fr den Fall das der Datensatz zerstckelt ist gibt es zwei
	 * Mglichkeiten: <UL><LI>Das empfangene Telegramm ist das erste. Dann wird der Datensatzindex bestimmt und die Relation zwischen altem und neuem
	 * Datensatzindex in einer Tabelle festgehalten. Diese Zuordnung ist notwendig, damit die anderen Teile des zerstckelten Datensatzes den gleichen Index
	 * erhalten knnen. Das Telegramm wird in die sogenannte Zerstckelungstabelle aufgenommen und an die Applikationen und Datenverteiler weitergeleitet, die
	 * positiv quittierte Anmeldungen initiiert haben. <LI>Das empfangene Telegramm ist nicht das erste. Dann wird es mit dem Datensatzindex versehen, der dem
	 * ersten Telegramm der Zerstckelung zugewiesen wurde. Das Telegramm wird in die Zerstckelungstabelle aufgenommen, und es wird berprft, ob alle Teile der
	 * Zerstckelung angekommen sind. Sind alle vorhanden, und sind sie nicht als nachgeliefert markiert, so werden sie in den Datenbestand des DaVs aufgenommen.
	 * Nachgelieferte Telegramme hingegen werden nicht im DaV vorgehalten. Das empfangene Telegramm wird dann an die Applikationen und Datenverteiler
	 * weitergeleitet, die positiv quittierte Anmeldungen initiiert haben. </UL>Der Datensatzindex wird so gebildet, dass die oberen 32 Bits die Zeit und die
	 * unteren 30 Bits den um eins erhhten Index der Basisanmeldeinformation enthalten. Die Zeit wird entweder aus dem Datensatzindex des im Datenbestand
	 * vorhandenen Datensatzes gewonnen oder ist die Zeit des Startens dieses Datenverteilers. Fr jede Basisanmeldeinformation wird ein Index in einer Tabelle
	 * eingerichtet, der bei jedem neuen Datensatz um eins erhht wird und in der Tabelle wieder abgespeichert wird. Wenn ein Telegramm an eine Applikation
	 * weitergeleitet wird, muss berprft werden, ob sich die Applikation evtl. nur fr Delta-Daten oder fr einen Teil der Attribute (oder beides) angemeldet
	 * hat. Wenn keine Einschrnkung vorliegt, wird das Telegramm an die Applikation weitergeleitet. Bei Einschrnkungen hingegen wird zunchst gewartet, bis der
	 * Datensatz komplett verfgbar ist (Zerstckelung). Je nachdem, welche Einschrnkungen vorliegen, wird der Datensatz wie folgt an die Applikation
	 * weitergeleitet: <UL><LI>Keine nachgelieferten Daten: Sind die Daten nachgeliefert, werden sie nicht gesendet. <LI>Delta-Daten: Es wird berprft, ob die
	 * nderungsindikatoren gesetzt sind. Nur dann wird der Datensatz zur Applikation weitergeleitet. <LI>Teilattribute: die angemeldeten Teilattribute werden aus
	 * dem Datensatz herausge- filtert und als neuer Datensatz zur Applikation weitergeleitet. <LI>Delta-Daten und Teilattribute: Es wird die Schnittmenge aus den
	 * Indikatoren der angemeldeten Attribute und den Indikatoren der genderten Attribute gebildet. Aus den Attributen dieser Schnittmenge wird ein neuer
	 * Datensatz gebildet und an die Applikation weitergeleitet. </UL>Das Senden eines Datensatzes wird natrlich nur dann erfolgen, wenn der Empfnger be-
	 * rechtigt ist, die Daten zu empfangen. Andernfalls wird ein spezieller Datensatz generiert, welcher den Empfnger darber informiert, dass er keine
	 * Berechtigung hat.
	 *
	 * @param inputApplicationConnection              Verbindung
	 * @param applicationDataTelegram Applikations-Daten-Telegramm
	 */
	public void sendData(T_A_HighLevelCommunication inputApplicationConnection, ApplicationDataTelegram applicationDataTelegram) {
		if(applicationDataTelegram == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		BaseSubscriptionInfo info = applicationDataTelegram.getBaseSubscriptionInfo();
		if(info == null) {
			throw new IllegalArgumentException("Argument ist inkonsistent");
		}

		InAndOutSubscription subscription = getSubscription(inputApplicationConnection.getId(), info);
		if((subscription == null) || (subscription.receip != TransmitterSubscriptionsConstants.POSITIV_RECEIP)) {
			return;
		}

		boolean save = false;
		// (See if main transmitter)
		T_A_SubscriptionComponent subscriptionComponent = getReceivingComponent(info);
		if(subscriptionComponent == null) {
			// Es wurde keine Anmeldung als Senke gefunden
			save = true;
			// Gibt es eine Anmeldung als Quelle?
			subscriptionComponent = getSendingComponent(info);
		}
		if(subscriptionComponent == null) {
			// Weder Quelle noch Senke am lokalen DAV verbunden, dann an andere DAV weiterleiten
			InAndOutSubscription inAndOutSubscription = isSuccessfullySubscribed(
					info, TransmitterSubscriptionsConstants.SENDER_SUBSCRIPTION
			);
			if((inAndOutSubscription == null) || (inAndOutSubscription.outSubscriptions == null)) {
				return; // no successfull subscription for this data
			}
			// Sender no source send to main transmitter
			TransmitterDataTelegram transmitterDataTelegram = new TransmitterDataTelegram(applicationDataTelegram, (byte)0);
			transmitterDataTelegram.setDataIndex(inputApplicationConnection.getId());
			for(int i = inAndOutSubscription.outSubscriptions.length - 1; i > -1; --i) {
				if(inAndOutSubscription.outSubscriptions[i] == null) {
					continue;
				}
				if(inAndOutSubscription.outSubscriptions[i].receip == TransmitterSubscriptionsConstants.POSITIV_RECEIP) {
					long transmitterId = inAndOutSubscription.outSubscriptions[i].targetTransmitterId;
					T_T_HighLevelCommunication _connection = _connectionsManager.getTransmitterConnection(transmitterId);
					if(_connection != null) {
						_connection.sendTelegram(transmitterDataTelegram);
					}
					break;
				}
			}
		}
		else {
			// Folgendes synchronized stellt sicher, dass Vergabe von Datensatzindex und Versand eines Telegramms zusammen durchgefhrt werden, da eine
			// verzahnte Abarbeitung durch verschiedene Threads zu Vertauschungen im Datensatzindex fhren knnten
			synchronized(subscriptionComponent) {
				final int totalTelegramsCount = applicationDataTelegram.getTotalTelegramsCount();
				if(totalTelegramsCount == 1 || (save == true && subscriptionComponent == inputApplicationConnection.getSubscriptionComponent())) {
					// Nicht gestckelte Telegramme oder Telegramme von der Quelle sofort verarbeiten
					processApplicationDataTelegramAsCentralDistributor(applicationDataTelegram, save, subscriptionComponent, inputApplicationConnection, info);
				}
				else {
					// Gestckelte Telegramme, die nicht von einer Quelle kommen
					// Es wird unterschieden nach dem ersten, einem mittleren und dem letzen Telegramm eines zerstckelten Datensatzes
					final int telegramNumber = applicationDataTelegram.getTelegramNumber();
					if(telegramNumber == 0) {
						// Das erste Telegramm eines zerstckelten Datensatzes
						List<ApplicationDataTelegram> stalledTelegramList = inputApplicationConnection.createStalledTelegramList(info, totalTelegramsCount);
						stalledTelegramList.add(applicationDataTelegram);
					}
					else {
						// Nicht das erste Telegramm eines zerstckelten Datensatzes
						if(telegramNumber + 1 != totalTelegramsCount) {
							// Ein mittleres Telegramm eines zerstckelten Datensatzes
							List<ApplicationDataTelegram> stalledTelegramList = inputApplicationConnection.getStalledTelegramList(info);
							if(stalledTelegramList == null) {
								_debug.error(
										"Ein mittleres Telegramm eines zerstckelten Datensatzes wurde nach Empfang eines Folgetelegramms nicht gefunden", info
								);
							}
							else {
								stalledTelegramList.add(applicationDataTelegram);
							}
						}
						else {
							// Das letzte Telegramm eines zerstckelten Datensatzes
							List<ApplicationDataTelegram> stalledTelegramList = inputApplicationConnection.deleteStalledTelegramList(info);
							if(stalledTelegramList == null) {
								_debug.error(
										"Das letzte Telegramm eines zerstckelten Datensatzes wurde nicht gefunden", info
								);
							}
							else {
								stalledTelegramList.add(applicationDataTelegram);
								for(ApplicationDataTelegram stalledApplicationDataTelegram : stalledTelegramList) {
									processApplicationDataTelegramAsCentralDistributor(
											stalledApplicationDataTelegram, save, subscriptionComponent, inputApplicationConnection, info
									);
								}
							}
						}
					}
				}
//				processApplicationDataTelegramAsCentralDistributor(applicationDataTelegram, save, subscriptionComponent, inputApplicationConnection, info);
			}
		}
	}

	private void processApplicationDataTelegramAsCentralDistributor(
			final ApplicationDataTelegram applicationDataTelegram,
			boolean save,
			final T_A_SubscriptionComponent subscriptionComponent,
			final T_A_HighLevelCommunication connection,
			final BaseSubscriptionInfo info
	) {
		// Wenn save==false, dann gibt es eine lokale Senke, sonst eine lokale Quelle
		TransmitterDataTelegram transmitterDataTelegram = new TransmitterDataTelegram(applicationDataTelegram, (byte)1);
		long index = transmitterDataTelegram.getDataNumber();
		Long originIndex = new Long(index);
		if(transmitterDataTelegram.getTelegramNumber() == 0) {
			if(save == true && subscriptionComponent == connection.getSubscriptionComponent()) {
				index = getDataIndex(info, true, index);
			}
			else {
				index = getDataIndex(info, false, index);
			}
			if(transmitterDataTelegram.getTotalTelegramsCount() != 1) {
				if(_usedDataIndexTable.put(originIndex, new Long(index)) != null) {
					_debug.error("Empfangener Datensatzindex nicht eindeutig");
				}
				;
			}
		}
		else {
			Long value = (Long)_usedDataIndexTable.get(originIndex);
			if(value == null) {
				throw new IllegalStateException("Inkonsistente Indexierung der Teiltelegramme");
			}
			index = value.longValue();
		}

		transmitterDataTelegram.setDataIndex(index);
		applicationDataTelegram.setDataIndex(index);

		ApplicationDataTelegram telegrams[] = null;
		final TransmitterDataTelegram[] t_telegrams;
		if(transmitterDataTelegram.getTotalTelegramsCount() == 1) {
			t_telegrams = new TransmitterDataTelegram[]{transmitterDataTelegram};
			telegrams = new ApplicationDataTelegram[]{applicationDataTelegram};
		}
		else {
			t_telegrams = _splittedTelegramsTable.put(transmitterDataTelegram);
			if(t_telegrams != null) {
				_usedDataIndexTable.remove(originIndex);
				telegrams = new ApplicationDataTelegram[t_telegrams.length];
				for(int i = 0; i < t_telegrams.length; ++i) {
					telegrams[i] = t_telegrams[i].getApplicationDataTelegram();
				}
			}
			else {
				save = false;
			}
		}
		saveAndSendToInterestedReceivingComponents(save, t_telegrams, transmitterDataTelegram, info, telegrams, applicationDataTelegram);
	}

	/**
	 * Speichert Telegramme eines Datensatzes im Cache und leitet die Telegramme an Applikationen oder andere Datenverteiler weiter, wenn sie sich dafr interessieren.
	 * @param save <code>true</code> falls der Datensatz gespeichert werden soll.
	 * @param transmitterTelegrams Alle Datenverteilertelegramme eines in mehrere Teiltelegramme zerlegten Datensatzes wenn alle Telegramme vorliegen, sonst <code>null</code>.
	 * @param info Datenidentifikation des Telegramms
	 * @param applicationTelegrams Alle Applikationstelegramme eines in mehrere Teiltelegramme zerlegten Datensatzes wenn alle Telegramme vorliegen, sonst <code>null</code>.
	 * @param applicationTelegram Telegramm fr Applikationen
	 * @param transmitterTelegram Telegramm fr andere Datenverteiler
	 */
	private void saveAndSendToInterestedReceivingComponents(
			final boolean save, final TransmitterDataTelegram[] transmitterTelegrams, final TransmitterDataTelegram transmitterTelegram, final BaseSubscriptionInfo info,
			final ApplicationDataTelegram[] applicationTelegrams,
			final ApplicationDataTelegram applicationTelegram) {
		boolean send = true;
		TransmitterDataTelegram[] previousTransmitterTelegrams = null;
		if(save) {
			final boolean result;
			synchronized(_connectionsManager.cacheManager) {
				long newDataIndex = transmitterTelegrams[0].getDataNumber();
				if((newDataIndex & 0x0000000000000003L) == 0) {
					CacheManager.CacheEntry cacheEntry = _connectionsManager.cacheManager.getCacheEntry(info);
					if(cacheEntry == null) {
						_connectionsManager.cacheManager.putNewCacheEntry(info, newDataIndex, transmitterTelegrams);
						//_cache.put(info, new CacheEntry(newDataIndex, transmitterTelegrams));
					}
					else {
						if(cacheEntry.getDataIndex() < newDataIndex) {
							cacheEntry.setDataIndex(newDataIndex);
							previousTransmitterTelegrams = cacheEntry.getTelegrams();
							if(!transmitterTelegram.getDelayedDataFlag()) cacheEntry.setTelegrams(transmitterTelegrams);
						}
						else if(cacheEntry.getDataIndex() == newDataIndex) {
							_debug.fine("Datensatz wird weitergeleitet, obwohl der Datensatzindex sich nicht verndert hat. Neuer Index: " + newDataIndex + ". "
							               + "Alter Index: " + cacheEntry.getDataIndex() + ". Info: " + info);
						}
						else {
							send = false;
							_debug.warning("Datensatz wird nicht weitergeleitet, weil der Datensatzindex nicht monoton ist. Neuer Index: " + newDataIndex + ". "
							               + "Alter Index: " + cacheEntry.getDataIndex() + ". Info: " + info);
						}
					}
				}
			}
		}
		if(send) {
			SubscriptionComponent subscriptionComponents[] = getInterrestedReceivingComponent(info);
			if(subscriptionComponents != null) {
				for(int j = 0; j < subscriptionComponents.length; ++j) {
					if(subscriptionComponents[j] != null) {
						if(subscriptionComponents[j].getType() == SubscriptionComponent.T_A) {
							try {
								sendDataToT_A_Component(
										(T_A_SubscriptionComponent)subscriptionComponents[j], applicationTelegrams, applicationTelegram,
										transmitterTelegrams, previousTransmitterTelegrams
								);
							}
							catch(ConfigurationException e) {
								e.printStackTrace();
							}
						}
						else {
							sendDataToT_T_Component((T_T_SubscriptionComponent)subscriptionComponents[j], transmitterTelegram);
						}
					}
				}
			}
		}
	}


	/**
	 * Diese Methode wird von der Protokollsteuerung DaV-DaV aufgerufen, wenn von einem Datenverteiler ein neuer Datensatz empfangen wurde. <UL><LI> Sind die Daten
	 * auf dem Weg von einem Sender zu einem Zentraldatenverteiler, so wird bei jedem Datenverteiler zunchst berprft, ob dieser bereits der gesuchte
	 * Zentraldatenverteiler dieser Datenstze ist, d. h., ob die Quelle oder Senke der Datenstze an diesem Datenverteiler angemeldet ist. Ist der Datensatz nicht
	 * zerstckelt, wird der Datensatzindex dieses Datensatzes bestimmt, und er wird an die Applikationen und Datenverteiler weitergeleitet, fr die bei diesem
	 * Datenverteiler positiv quittierte Anmeldungen vorliegen. Wenn der Datensatz nicht nachgeliefert ist, wird er in den Datenbestand aufgenommen.  <br>Ist der
	 * Datensatz zerstckelt, dann gibt es zwei Mglichkeiten: <UL><LI>Das empfangene Telegramm ist das erste. <LI>Das empfangene Telegramm ist nicht das
	 * erste.</UL>Wenn dieser Datenverteiler nicht der Zentraldatenverteiler fr diese Datenstze ist dann wird berprft, ob ein positiv quittierter
	 * Anmeldungsvermerk fr einen Sender vorliegt. Ist dies der Fall, dann wird der Datensatz zum dem Datenverteiler weitergeleitet, an dem eine positiv
	 * quittierte Folgeanmeldung vorliegt. <LI> Sind die Daten auf dem Weg von einem Zentraldatenverteiler zu einem Empfnger, so wird bei jedem Datenverteiler
	 * berprft, ob der Datensatz zerstckelt ist. Ist der Datensatz nicht zerstckelt, wird der Datensatzindex dieses Datensatzes bestimmt und an die
	 * Applikationen und Datenverteiler weitergeleitet, fr die bei diesem Datenverteiler positiv quittierte Anmeldungen vorliegen. Wenn der Datensatz nicht
	 * nachgeliefert ist, wird er in den Datenbestand aufgenommen. Anschlieend wird die updatePendingSubscription-Methode der Verbindungsverwaltung aufgerufen, um
	 * die Umleitungsantrge mit den Informationen der neuen Datenstze zu versehen.  <br>Ist der Datensatz zerstckelt, dann gibt es zwei Mglichkeiten:
	 * <UL><LI>Das empfangene Telegramm ist das erste.<LI>Das empfangener Telegramm ist nicht das erste. Dann wird es mit dem Datensatzindex versehen, der dem
	 * ersten Telegramm der Zerstckelung zugewiesen wurde. Das Telegramm wird in die Zerstckelungstabelle aufgenommen, und es wird berprft, ob alle Teile der
	 * Zerstckelung angekommen sind. Sind alle vorhanden, und sind sie nicht als nachgeliefert markiert, so werden sie alle in den Datenbestand aufgenommen.
	 * Anschlieend wird die updatePendingSubscription- Methode der Verbindungsverwaltung aufgerufen, um die Umleitungsantrge mit den Informationen der neuen
	 * Datenstze zu versehen. Das empfangene Telegramm wird dann an die Applikationen und Datenverteiler weitergeleitet, die positiv quittierte Anmeldungen
	 * initiiert haben.</UL></UL> Der Datensatzindex wird so gebildet, dass die oberen 32 Bits die Zeit und die unteren 30 Bits den um eins erhhten Index der
	 * Basisanmeldeinformation enthalten. Die Zeit wird entweder aus dem Datensatzindex des im Datenbestand vorhandenen Datensatzes gewonnen oder ist die Zeit des
	 * Startens dieses Datenverteilers. Fr jede Basisanmeldeinformation wird ein Index in einer Tabelle eingerichtet, der bei jedem neuen Datensatz um eins erhht
	 * wird und in der Tabelle wieder abgespeichert wird. Wenn ein Telegramm an eine Applikation weitergeleitet wird, muss berprft werden, ob sich die
	 * Applikation evtl. nur fr Delta-Daten oder fr einen Teil der Attribute (oder beides) angemeldet hat. Wenn keine Einschrnkung vorliegt, wird das Telegramm
	 * an die Applikation weitergeleitet. Bei Einschrnkungen hingegen wird zunchst gewartet, bis der Datensatz komplett verfgbar ist (Zerstckelung). Je
	 * nachdem, welche Einschrnkungen vorliegen, wird der Datensatz wie folgt an die Applikation weitergeleitet: <UL><LI>Keine nachgelieferten Daten: Sind die
	 * Daten nachgeliefert, werden sie nicht gesendet.<LI>Delta-Daten: Es wird berprft, ob die nderungsindikatoren gesetzt sind. Nur dann wird der Datensatz zur
	 * Applikation weitergeleitet. <LI>Teilattribute: die angemeldeten Teilattribute werden aus dem Datensatz herausgefiltert und als neuer Datensatz zur
	 * Applikation weitergeleitet. <LI>Delta-Daten und Teilattribute: Es wird die Schnittmenge aus den Indikatoren der angemeldeten Attribute und den Indikatoren
	 * der genderten Attribute gebildet. Aus den Attributen dieser Schnittmenge wird ein neuer Datensatz gebildet und an die Applikation weitergeleitet. </UL>Das
	 * Senden eines Datensatzes wird natrlich nur dann erfolgen, wenn der Empfnger berechtigt ist, die Daten zu empfangen. Andernfalls wird ein spezieller
	 * Datensatz generiert, welcher den Empfnger darber informiert, dass er keine Berechtigung hat.
	 *
	 * @param inputTransmitterConnection              Verbindung
	 * @param transmitterDataTelegram Transmitter-Daten-Telegramm
	 */
	public final void sendData(T_T_HighLevelCommunication inputTransmitterConnection, TransmitterDataTelegram transmitterDataTelegram) {
		if(transmitterDataTelegram == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		BaseSubscriptionInfo info = transmitterDataTelegram.getBaseSubscriptionInfo();
		if(info == null) {
			throw new IllegalArgumentException("Argument ist inkonsistent");
		}

		T_A_SubscriptionComponent subscriptionComponent = getReceivingComponent(info);
		boolean save = false;
		//(See if main transmitter)
		if(subscriptionComponent == null) {
			save = true;
			subscriptionComponent = getSendingComponent(info);
		}
		if(transmitterDataTelegram.getDirection() == 0) { // Send to main transmitter
			if(subscriptionComponent == null) {
				InAndOutSubscription inAndOutSubscription = isSuccessfullySubscribed(
						info, TransmitterSubscriptionsConstants.SENDER_SUBSCRIPTION
				);
				if((inAndOutSubscription == null) || (inAndOutSubscription.outSubscriptions == null)) {
					return; // no successfull subscription for this data
				}
				for(int i = inAndOutSubscription.outSubscriptions.length - 1; i > -1; --i) {
					if(inAndOutSubscription.outSubscriptions[i] == null) {
						continue;
					}
					if(inAndOutSubscription.outSubscriptions[i].receip == TransmitterSubscriptionsConstants.POSITIV_RECEIP) {
						long transmitterId = inAndOutSubscription.outSubscriptions[i].targetTransmitterId;
						T_T_HighLevelCommunication _connection = _connectionsManager.getTransmitterConnection(transmitterId);
						if(_connection != null) {
							_connection.sendTelegram(transmitterDataTelegram);
						}
						break;
					}
				}
			}
			else {
				if(isInvalidSubscription(info, TransmitterSubscriptionsConstants.RECEIVER_SUBSCRIPTION)) {
					return;
				}
				// Folgendes synchronized stellt sicher, dass Vergabe von Datensatzindex und Versand eines Telegramms zusammen durchgefhrt werden, da eine
				// verzahnte Abarbeitung durch verschiedene Threads zu Vertauschungen im Datensatzindex fhren knnten
				synchronized(subscriptionComponent) {
					ApplicationDataTelegram telegrams[] = null;
					final TransmitterDataTelegram t_telegrams[];
					boolean noNeedToWait = false;
					long index = transmitterDataTelegram.getDataNumber();
					Long originIndex = new Long(index);
					if(transmitterDataTelegram.getTelegramNumber() == 0) {
						index = getDataIndex(info, false, index);
						if(transmitterDataTelegram.getTotalTelegramsCount() == 1) {
							noNeedToWait = true;
						}
						else {
							_usedDataIndexTable.put(originIndex, new Long(index));
						}
					}
					else {
						Long value = (Long)_usedDataIndexTable.get(originIndex);
						if(value == null) {
							throw new IllegalStateException("Inkonsistente Indexierung der Teiltelegramme");
						}
						index = value.longValue();
					}
					transmitterDataTelegram.setDataIndex(index);
					transmitterDataTelegram.setDirection((byte)1);
					ApplicationDataTelegram applicationDataTelegram = transmitterDataTelegram.getApplicationDataTelegram();
					if(noNeedToWait) {
						t_telegrams = new TransmitterDataTelegram[]{transmitterDataTelegram};
						telegrams = new ApplicationDataTelegram[]{applicationDataTelegram};
					}
					else {
						t_telegrams = _splittedTelegramsTable.put(transmitterDataTelegram);
						if(t_telegrams != null) {
							_usedDataIndexTable.remove(originIndex);
							telegrams = new ApplicationDataTelegram[t_telegrams.length];
							for(int i = 0; i < t_telegrams.length; ++i) {
								telegrams[i] = t_telegrams[i].getApplicationDataTelegram();
							}
						}
						else {
							save = false;
						}
					}

					saveAndSendToInterestedReceivingComponents(save, t_telegrams, transmitterDataTelegram, info, telegrams, applicationDataTelegram);
				}
			}
		}
		else {
			if(isInvalidSubscription(info, TransmitterSubscriptionsConstants.RECEIVER_SUBSCRIPTION)) {
				return;
			}
			ApplicationDataTelegram telegrams[] = null;
			ApplicationDataTelegram applicationDataTelegram = transmitterDataTelegram.getApplicationDataTelegram();
			TransmitterDataTelegram[] t_telegrams = null;
			if((applicationDataTelegram.getTotalTelegramsCount() == 1) && (applicationDataTelegram.getTelegramNumber() == 0)) {
				if(_connectionsManager.updatePendingSubscription(info, inputTransmitterConnection.getId(), transmitterDataTelegram.getDataNumber())) {
					telegrams = new ApplicationDataTelegram[]{applicationDataTelegram};
					if(save) {
						t_telegrams = new TransmitterDataTelegram[]{transmitterDataTelegram};
					}
				}
			}
			else {
				t_telegrams = _splittedTelegramsTable.put(transmitterDataTelegram);
				if(t_telegrams != null) {
					if(_connectionsManager.updatePendingSubscription(
							info, inputTransmitterConnection.getId(), transmitterDataTelegram.getDataNumber()
					)) {
						telegrams = new ApplicationDataTelegram[t_telegrams.length];
						for(int i = 0; i < t_telegrams.length; ++i) {
							telegrams[i] = t_telegrams[i].getApplicationDataTelegram();
						}
					}
				}
				else {
					save = false;
				}
			}
			saveAndSendToInterestedReceivingComponents(save, t_telegrams, transmitterDataTelegram, info, telegrams, applicationDataTelegram);
		}
	}

	/**
	 * Leitet ein Telegramm an einen anderen Datenverteiler weiter.
	 *
	 * @param subscriptionComponent
	 * @param transmitterDataTelegram
	 */
	private void sendDataToT_T_Component(
			T_T_SubscriptionComponent subscriptionComponent, TransmitterDataTelegram transmitterDataTelegram
	) {
		T_T_HighLevelCommunication _connection = (T_T_HighLevelCommunication)subscriptionComponent.getConnection();
		if(_connection != null) {
			BaseSubscriptionInfo info = transmitterDataTelegram.getBaseSubscriptionInfo();
			_connection.sendTelegram(transmitterDataTelegram);
		}
	}

	/**
	 * Leitet ein Telegramm an eine Applikation weiter.
	 *
	 * @param subscriptionComponent
	 * @param allAppTelegrams
	 * @param lastAppTelegram
	 * @param transmitterTelegrams
	 * @param previousTransmitterTelegrams
	 */
	private void sendDataToT_A_Component(
			T_A_SubscriptionComponent subscriptionComponent,
			ApplicationDataTelegram allAppTelegrams[],
			ApplicationDataTelegram lastAppTelegram,
			final TransmitterDataTelegram[] transmitterTelegrams,
			final TransmitterDataTelegram[] previousTransmitterTelegrams) {
		T_A_HighLevelCommunication _connection = (T_A_HighLevelCommunication)subscriptionComponent.getConnection();
		if(_connection != null) {
			BaseSubscriptionInfo info = lastAppTelegram.getBaseSubscriptionInfo();
			ReceiveSubscriptionInfo receiveSubscriptionInfo = subscriptionComponent.getReceiveSubscription(info);
			if(receiveSubscriptionInfo != null) {
				if((receiveSubscriptionInfo.getDelayedDataFlag() == false) && (lastAppTelegram.getDelayedDataFlag() == true)) {
					return;
				}
				long lastDataIndex = receiveSubscriptionInfo.getLastDataIndex();
				receiveSubscriptionInfo.setLastDataIndex(lastAppTelegram.getDataNumber());
				lastAppTelegram.setAttributesIndicator(null);
				if(receiveSubscriptionInfo.getDeltaDataFlag()) {
					if((allAppTelegrams != null) && (allAppTelegrams.length > 0) && (allAppTelegrams[0] != null)) {
						boolean sendAll = true;
						if(isDeltaAllowed(info)) {
							if((lastDataIndex > 0) && ((lastDataIndex & 0x0000000000000003L) == 0)
							   && transmitterTelegrams != null && previousTransmitterTelegrams != null
							   && transmitterTelegrams.length == previousTransmitterTelegrams.length && !transmitterTelegrams[0].getDelayedDataFlag()
							   && transmitterTelegrams[0].getErrorFlag() == 0 && previousTransmitterTelegrams[0].getErrorFlag() == 0) {
								sendAll = false;
								for(int i = transmitterTelegrams.length - 1; i >= 0; i--) {
									TransmitterDataTelegram transmitterTelegram = transmitterTelegrams[i];
									TransmitterDataTelegram previousTransmitterTelegram = previousTransmitterTelegrams[i];
									if(!Arrays.equals(transmitterTelegram.getData(), previousTransmitterTelegram.getData())) {
										sendAll = true;
										break;
									}

								}
							}
						}
						if(sendAll) {
							// Alle Telegramme des Datensatzes weiterleiten
							_connection.sendData(allAppTelegrams);
						}
					}
				}
				else {
					_connection.sendData(lastAppTelegram);
				}
			}
		}
	}

	/**
	 * Leitet ein Telegramm an eine Applikation weiter.
	 *
	 * @param subscriptionComponent
	 * @param info
	 */
	final void sendDataToT_A_Component(T_A_SubscriptionComponent subscriptionComponent, BaseSubscriptionInfo info) {
		if((subscriptionComponent == null) && (info == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ApplicationDataTelegram telegrams[] = _connectionsManager.cacheManager.getCurrentDataForApplication(info);
		if((telegrams == null) || (telegrams.length == 0) || (telegrams[0] == null)) {
			return;
		}
		T_A_HighLevelCommunication _connection = (T_A_HighLevelCommunication)subscriptionComponent.getConnection();
		if(_connection != null) {
			ReceiveSubscriptionInfo receiveSubscriptionInfo = subscriptionComponent.getReceiveSubscription(info);
			if(receiveSubscriptionInfo != null) {
				if((receiveSubscriptionInfo.getDelayedDataFlag() == false) && (telegrams[0].getDelayedDataFlag() == true)) {
					return;
				}

				receiveSubscriptionInfo.setLastDataIndex(telegrams[0].getDataNumber());
				if(receiveSubscriptionInfo.getDeltaDataFlag()) {
					if(false && isDeltaAllowed(info)) {
						ApplicationDataTelegram applicationDataTelegrams[] = getDeltaSubDataObject(telegrams, null, null);
						if(applicationDataTelegrams != null) {
							_connection.sendData(applicationDataTelegrams);
						}
					}
					else {
						telegrams[0].setAttributesIndicator(null);
						_connection.sendData(telegrams);
					}
				}
				else {
					telegrams[0].setAttributesIndicator(null);
					_connection.sendData(telegrams);
				}
			}
		}
	}


	/**
	 * Erzeugt ein Teildatensatz mit den spezifizierten und genderten Attributen.
	 *
	 * @param telegrams
	 * @param indicators
	 * @param changes
	 *
	 * @return
	 */
	@Deprecated
	private ApplicationDataTelegram[] getDeltaSubDataObject(
			ApplicationDataTelegram telegrams[], byte[] indicators, byte[] changes
	) {
		if((telegrams == null) || (telegrams.length == 0) || (telegrams[0] == null)) {
			throw new IllegalArgumentException("bergabeparameter ist null");
		}

		// Put all the bytes together
		ByteArrayOutputStream byteBuffer = new ByteArrayOutputStream();
		try {
			for(int i = 0; i < telegrams.length; ++i) {
				if(telegrams[i] != null) {
					byte byteArray[] = telegrams[i].getData();
					if(byteArray != null) {
						byteBuffer.write(byteArray);
					}
				}
			}
		}
		catch(IOException ex) {
			ex.printStackTrace();
			return null;
		}

		//Get the Datavalues out of the stream
		DataValue values[] = null;
		if(byteBuffer.size() > 0) {
			try {
				DataInputStream in = new DataInputStream(new ByteArrayInputStream(byteBuffer.toByteArray()));
				final DataModel dataModel = _connectionsManager.dataModel;
				final BaseSubscriptionInfo baseSubscriptionInfo = telegrams[0].getBaseSubscriptionInfo();
				final AttributeGroupUsage attributeGroupUsage = dataModel.getAttributeGroupUsage(baseSubscriptionInfo.getUsageIdentification());
				final AttributeGroup atg = attributeGroupUsage.getAttributeGroup();
				values = StreamFetcher.getInstance().getDataValuesFromStream(
						_connectionsManager.dataModel, atg, in
				);
			}
			catch(IOException ex) {
				ex.printStackTrace();
				return null;
			}
		}

		// Empty telegram (no data inside)
		if(values == null) {
			byte errorFlag = telegrams[0].getErrorFlag();
			long dataIndex = telegrams[0].getDataNumber();
			if((telegrams.length == 1) && ((errorFlag != 0x00) || ((dataIndex & 0x0000000000000003) != 0))) { // only one error telegram
				if(telegrams[0] != null) {
					telegrams[0].setAttributesIndicator(indicators);
				}
				return telegrams;
			}
			else {
				return null;
			}
		}

		byte lastMask = (byte)0xFF;
		int length = (values.length / 8);
		int diff = values.length - (length * 8);
		if(diff > 0) {
			++length;
			lastMask = (byte)(Math.pow(2, diff) - 1);
		}
		// Gather the specified attributes
		byte _indicators[] = new byte[length];
		ArrayList subList = new ArrayList();
		for(int i = 0; i < length; ++i) {
			byte b1 = indicators == null ? (byte)0xFF : indicators[i];
			byte b2 = changes == null ? (byte)0xFF : changes[i];
			_indicators[i] = (byte)(b1 & b2);

			int index = i * 8;
			if(((b1 & 0x01) == 0x01) && ((b2 & 0x01) == 0x01) && (index < values.length)) {
				subList.add(values[index]);
			}
			++index;
			if(((b1 & 0x02) == 0x02) && ((b2 & 0x02) == 0x02) && (index < values.length)) {
				subList.add(values[index]);
			}
			++index;
			if(((b1 & 0x04) == 0x04) && ((b2 & 0x04) == 0x04) && (index < values.length)) {
				subList.add(values[index]);
			}
			++index;
			if(((b1 & 0x08) == 0x08) && ((b2 & 0x08) == 0x08) && (index < values.length)) {
				subList.add(values[index]);
			}
			++index;
			if(((b1 & 0x10) == 0x10) && ((b2 & 0x10) == 0x10) && (index < values.length)) {
				subList.add(values[index]);
			}
			++index;
			if(((b1 & 0x20) == 0x20) && ((b2 & 0x20) == 0x20) && (index < values.length)) {
				subList.add(values[index]);
			}
			++index;
			if(((b1 & 0x40) == 0x40) && ((b2 & 0x40) == 0x40) && (index < values.length)) {
				subList.add(values[index]);
			}
			++index;
			if(((b1 & 0x80) == 0x80) && ((b2 & 0x80) == 0x80) && (index < values.length)) {
				subList.add(values[index]);
			}
		}
		_indicators[length - 1] &= lastMask;
		// put them to one Object
		int size = subList.size();
		if(size == 0) {
			return null;
		}
		byteBuffer = new ByteArrayOutputStream();
		try {
			DataOutputStream out = new DataOutputStream(byteBuffer);
			if(size > 0) {
				for(int i = 0; i < size; ++i) {
					DataValue attribute = (DataValue)subList.get(i);
					if(attribute != null) {
						attribute.write(out);
					}
				}
			}
			out.flush();
			out.close();
		}
		catch(IOException ex) {
			ex.printStackTrace();
			return null;
		}
		byte data[] = byteBuffer.toByteArray();
		SendDataObject sendDataObject = new SendDataObject(
				telegrams[0].getBaseSubscriptionInfo(),
				telegrams[0].getDelayedDataFlag(),
				telegrams[0].getDataNumber(),
				telegrams[0].getDataTime(),
				telegrams[0].getErrorFlag(),
				_indicators,
				data
		);
		// splitt the new object to small telegrams
		return TelegramUtility.splitToApplicationTelegrams(sendDataObject);
	}

	/**
	 *  Gibt die fortlaufende Nummer der Daten zurck
	 *
	 * @param info
	 * @param source
	 * @param originIndex
	 *
	 * @return
	 */
	private synchronized long getDataIndex(BaseSubscriptionInfo info, boolean source, long originIndex) {
		if(info == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		long time;
		if(source) {
			time = originIndex & 0xFFFFFFFF00000000L;
		}
		else {
			time = _connectionsManager.cacheManager.getCurrentDataIndex(info);
			if(time == 0) {
				time = CommunicationConstant.START_TIME;
			}
			else {
				time = time & 0xFFFFFFFF00000000L;
			}
		}
		Object obj = _dataIndexTable.get(info);
		int sendIndex;
		if(obj == null) {
			sendIndex = (int)((originIndex >> 2) & 0x000000003FFFFFFFL);
			if(sendIndex > 0x3FFFFFFF) {
				sendIndex = 1;
			}
		}
		else {
			sendIndex = ((Integer)obj).intValue();
			if(++sendIndex > 0x3FFFFFFF) {
				sendIndex = 1;
			}
		}
		_dataIndexTable.put(info, new Integer(sendIndex));
		long dataNumber = ((time | (((long)sendIndex << 2) & 0x00000000FFFFFFFCL)) & 0xFFFFFFFFFFFFFFFCL);
		//System.out.println("originIndex = " + originIndex + " source = " + source + " dataNumber = " + dataNumber);
		return dataNumber;
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung und von sich selbst aufgerufen, um festzustellen, welche Verbindungen am Empfang der Datenstze fr die
	 * spezifizierte Basisanmeldeinformation interessiert sind. Alle Anmeldekomponenten der Verbindungen werden darauf untersucht, ob eine Senken- oder
	 * Empfangsanmeldung fr das spezifizierte Datum existiert. Ist keine Senke und kein Empfnger angemeldet, so wird Null zurckgegeben.
	 *
	 * @param info Basisanmeldeinformationen
	 *
	 * @return Feld der interessierten Empfnger(Empfnger oder Senke)
	 */
	final SubscriptionComponent[] getInterrestedReceivingComponent(BaseSubscriptionInfo info) {
		if(_subscriptionsTable == null) {
			throw new IllegalArgumentException("Anmeldemanager ist nicht fertig initiallisiert");
		}
		ArrayList result = new ArrayList();
		synchronized(_subscriptionsTable) {
			ArrayList list = new ArrayList(_subscriptionsTable.values());
			for(int i = list.size() - 1; i > -1; --i) {
				SubscriptionComponent subscriptionComponent = (SubscriptionComponent)list.get(i);
				if(subscriptionComponent != null) {
					if(subscriptionComponent.getType() == SubscriptionComponent.T_A) {
						T_A_SubscriptionComponent _subscriptionComponent = (T_A_SubscriptionComponent)subscriptionComponent;
						if(_subscriptionComponent.getReceiveSubscription(info) != null) {
							result.add(_subscriptionComponent);
						}
					}
					else {
						T_T_SubscriptionComponent _subscriptionComponent = (T_T_SubscriptionComponent)subscriptionComponent;
						if(_subscriptionComponent.getReceiveSubscription(info) != null) {
							result.add(_subscriptionComponent);
						}
					}
				}
			}
		}
		int size = result.size();
		if(size == 0) {
			return null;
		}
		else {
			SubscriptionComponent array[] = new SubscriptionComponent[size];
			for(int i = 0; i < size; ++i) {
				array[i] = (SubscriptionComponent)result.get(i);
			}
			return array;
		}
	}

	/**
	 * Diese Methode wird von der Verbindungsverwaltung und von sich selbst aufgerufen, um festzustellen, welche Verbindungen sich fr das Senden von Datenstzen
	 * fr die spezifizierte Basisanmeldeinformation angemeldet haben. Alle Anmeldekomponenten der Verbindungen werden darauf untersucht, ob eine Qell- oder
	 * Sendeanmeldung fr das spezifizierten Datum existiert. Ist keine Quelle und kein Sender angemeldet, so wird Null zurckgegeben.
	 *
	 * @param info Basisanmeldeinformationen
	 *
	 * @return Feld der Sender(Sender oder Quelle) eines bestimmten Datums
	 */
	final SubscriptionComponent[] getInterrestedSendingComponent(BaseSubscriptionInfo info) {
		if(_subscriptionsTable == null) {
			throw new IllegalArgumentException("Anmeldemanager ist nicht fertig initiallisiert");
		}
		ArrayList result = new ArrayList();
		synchronized(_subscriptionsTable) {
			ArrayList list = new ArrayList(_subscriptionsTable.values());
			for(int i = list.size() - 1; i > -1; --i) {
				SubscriptionComponent subscriptionComponent = (SubscriptionComponent)list.get(i);
				if(subscriptionComponent != null) {
					if(subscriptionComponent.getType() == SubscriptionComponent.T_A) {
						T_A_SubscriptionComponent _subscriptionComponent = (T_A_SubscriptionComponent)subscriptionComponent;
						if(_subscriptionComponent.getSendSubscription(info) != null) {
							result.add(_subscriptionComponent);
						}
					}
					else {
						T_T_SubscriptionComponent _subscriptionComponent = (T_T_SubscriptionComponent)subscriptionComponent;
						if(_subscriptionComponent.getSendSubscription(info) != null) {
							result.add(_subscriptionComponent);
						}
					}
				}
			}
		}
		int size = result.size();
		if(size == 0) {
			return null;
		}
		else {
			SubscriptionComponent array[] = new SubscriptionComponent[size];
			for(int i = 0; i < size; ++i) {
				array[i] = (SubscriptionComponent)result.get(i);
			}
			return array;
		}
	}

	/**
	 * Delta ist zulssig wenn nicht mehr als ein Sender oder Quelle eines Datums vorhanden ist.
	 *
	 * @param info
	 *
	 * @return
	 */
	private boolean isDeltaAllowed(BaseSubscriptionInfo info) {
		if(_subscriptionsTable == null) {
			throw new IllegalArgumentException("Anmeldemanager ist nicht fertig initiallisiert");
		}
		int senderNumber = 0;
		synchronized(_subscriptionsTable) {
			ArrayList list = new ArrayList(_subscriptionsTable.values());
			for(int i = list.size() - 1; i > -1; --i) {
				SubscriptionComponent subscriptionComponent = (SubscriptionComponent)list.get(i);
				if(subscriptionComponent != null) {
					if(subscriptionComponent.getType() == SubscriptionComponent.T_A) {
						T_A_SubscriptionComponent _subscriptionComponent = (T_A_SubscriptionComponent)subscriptionComponent;
						if(_subscriptionComponent.getSendSubscription(info) != null) {
							++senderNumber;
						}
					}
					else {
						T_T_SubscriptionComponent _subscriptionComponent = (T_T_SubscriptionComponent)subscriptionComponent;
						if(_subscriptionComponent.getSendSubscription(info) != null) {
							++senderNumber;
						}
					}
					if(senderNumber > 1) {
						return false;
					}
				}
			}
		}
		return true;
	}

	/**
	 * berprft ob mehr als eine Quelle fr ein Datum im System vorhanden ist.
	 *
	 * @param key
	 * @param subscriptionState
	 *
	 * @return
	 */
	private boolean isInvalidSubscription(BaseSubscriptionInfo key, byte subscriptionState) {
		if(key == null) {
			throw new IllegalArgumentException("Argument ist null");
		}
		ArrayList list = (ArrayList)_inOutSubscriptionsTable.get(key);
		if(list != null) {
			synchronized(list) {
				for(int i = list.size() - 1; i > -1; --i) {
					InAndOutSubscription subscription = (InAndOutSubscription)list.get(i);
					if(subscription != null) {
						if((subscription.subscriptionState == subscriptionState)
						   && (subscription.receip == TransmitterSubscriptionsConstants.MORE_THAN_ONE_POSITIV_RECEIP) && (subscription.outSubscriptions
						                                                                                                  != null)) {
							return true;
						}
					}
				}
			}
		}
		return false;
	}

	/** Gibt alle Anmeldungen zu Debug-Zwecken aus*/
	public void printSubscriptions() {
		final Enumeration keys = _subscriptionsTable.keys();
		while(keys.hasMoreElements()) {
			Long id = (Long)keys.nextElement();
			SubscriptionComponent component = (SubscriptionComponent)_subscriptionsTable.get(id);
			System.out.println("id " + id + ", component " + component.toString());
			component.printSubscriptions();
		}
	}
}
