/*
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2005 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.pat.sysbed.
 * 
 * de.bsvrz.pat.sysbed is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pat.sysbed is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pat.sysbed; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.pat.sysbed.dataview;

import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.SystemObject;

/**
 * Ein Objekt dieser Klasse reprsentiert einen Datensatz in der {@link DataViewFrame Online-Tabelle}. Reichen die Informationen eines {@link ResultData} nicht
 * aus, so steht ein weiterer Konstruktor zur Verfgung, dem neben den bentigten Werten fr die Online-Tabelle die {@link de.bsvrz.dav.daf.main.archive.ArchiveDataKind Art der
 * Archivanfrage} bergeben werden kann.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5063 $
 */
public class DataTableObject {

	private final SystemObject _object;

	private final DataDescription _dataDescription;

	private final TimingType _timingType;

	private final long _archiveTime;

	private final long _dataTime;

	private final long _dataIndex;

	private final DataState _dataState;

	private final ArchiveDataKind _dataKind;

	private final Data _data;

	/**
	 * Erstellt einen Datensatz aus den Informationen eines {@link de.bsvrz.dav.daf.main.ResultData Ergebnisdatensatzes} zur Anzeige in der {@link DataViewFrame Online-Tabelle}.
	 *
	 * @param resultData Ergebnisdatensatz
	 */
	public DataTableObject(final ResultData resultData) {
		this(
				resultData.getObject(),
				resultData.getDataDescription(),
				TimingType.DATA_TIME,
				0,
				resultData.getDataTime(),
				resultData.getDataIndex(),
				resultData.getDataState(),
				(resultData.isDelayedData() ? ArchiveDataKind.ONLINE_DELAYED : ArchiveDataKind.ONLINE),
				resultData.getData()
		);
	}

	/**
	 * Erstellt einen Datensatz zur Anzeige in der {@link DataViewFrame Online-Tabelle}.
	 *
	 * @param object          das Systemobjekt
	 * @param dataDescription Datenbeschreibung besteht aus der Attributgruppe, Aspekt und der Simulationsvariante
	 * @param timingType      gibt an, welcher der drei Typen (Archivzeit, Datenzeit und Datenindex) angezeigt werden soll
	 * @param archiveTime     der Archivzeitstempel
	 * @param dataTime        der Datenzeitstempel
	 * @param dataIndex       der Datenindex
	 * @param dataState       Zustand der Daten
	 * @param dataKind        Art des Datensatzes
	 * @param data            die Daten
	 */
	public DataTableObject(
			SystemObject object,
			DataDescription dataDescription,
			TimingType timingType,
			long archiveTime,
			long dataTime,
			long dataIndex,
			DataState dataState,
			ArchiveDataKind dataKind,
			Data data
	) {
		_object = object;
		_dataDescription = dataDescription;
		_timingType = timingType;
		_archiveTime = archiveTime;
		_dataTime = dataTime;
		_dataIndex = dataIndex;
		_dataState = dataState;
		_dataKind = dataKind;
		_data = data;
	}

	/**
	 * Gibt das Systemobjekt zurck.
	 *
	 * @return das Systemobjekt
	 */
	public SystemObject getObject() {
		return _object;
	}

	/**
	 * Gibt die Datenbeschreibung (Attributgruppe, Aspekt und Simulationsvariante) zurck.
	 *
	 * @return die Datenbeschreibung
	 */
	public DataDescription getDataDescription() {
		return _dataDescription;
	}

	/**
	 * Gibt an, welcher Zeitstempel angezeigt werden soll.
	 *
	 * @return die Art des Zeitstempels, der angezeigt werden soll
	 */
	public TimingType getTimingType() {
		return _timingType;
	}

	/**
	 * Gibt die Archivzeit zurck.
	 *
	 * @return die Archivzeit
	 */
	public long getArchiveTime() {
		return _archiveTime;
	}

	/**
	 * Gibt die Datenzeit zurck.
	 *
	 * @return die Datenzeit
	 */
	public long getDataTime() {
		return _dataTime;
	}

	/**
	 * Gibt den Datenindex zurck.
	 *
	 * @return der Datenindex
	 */
	public long getDataIndex() {
		return _dataIndex;
	}

	/**
	 * Gibt den Zustand der Daten zurck.
	 *
	 * @return der Zustand der Daten
	 */
	public DataState getDataState() {
		return _dataState;
	}

	/**
	 * Gibt die Art des Datensatzes zurck (ob OA = online aktuell, ON = online nachgeliefert, NA = nachgefordert aktuell, NN = nachgefordert nachgeliefert - siehe
	 * auch Technische Anforderungen Archivsystem).
	 *
	 * @return die Art des Datensatzes
	 */
	public ArchiveDataKind getDataKind() {
		return _dataKind;
	}

	/**
	 * Gibt die Daten dieses Datensatzes zurck.
	 *
	 * @return die Daten dieses Datensatzes
	 */
	public Data getData() {
		return _data;
	}

	/**
	 * Gibt den Inhalt dieser Klasse als String zurck.
	 *
	 * @return der Inhalt dieser Klasse als String
	 */
	public String toString() {
		return "DataTableObject{" + "_object=" + _object + ", _dataDescription=" + _dataDescription + ", _timingType=" + _timingType + ", _archiveTime="
		       + _archiveTime + ", _dataTime=" + _dataTime + ", _dataIndex=" + _dataIndex + ", _dataState=" + _dataState + ", _dataKind=" + _dataKind
		       + ", _data=" + _data + "}";
	}
}
