/*
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2004 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.pat.sysbed.
 * 
 * de.bsvrz.pat.sysbed is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pat.sysbed is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pat.sysbed; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.pat.sysbed.dataview;

import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.SystemObject;

import javax.swing.*;
import javax.swing.border.EtchedBorder;
import java.awt.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.List;

/**
 * Diese Klasse erstellt zu einem Datensatz der Form {@link de.bsvrz.dav.daf.main.ResultData} eine hierarchische Struktur der Form {@link RowData}. Zustzlich knnen die Komponenten
 * eines Zeilenkopfes und einer Nutzdatenzeile angefordert werden.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5063 $
 */
public class Dataset {

	/** speichert einen Datensatz vom Datenverteiler */
	private final DataTableObject _dataTableObject;

	/** speichert die Struktur des Spaltenheaders, damit die Nutzdaten damit verknpft werden knnen */
	private final HeaderGrid _head;

	/** speichert den Datensatz in einer hierarchischen Struktur */
	private RowData _rowData = null;

	/** speichert die zum Datensatz gehrende Zeilenkopf-Komponente */
	private JComponent _rowHeaderRow = null;

	/** speichert die zum Datensatz gehrende Datenzeilen-Komponente */
	private JComponent _viewportRow = null;

	/** speichert die Hhe der Komponente einer Zeile */
	private int _height = -1;

	/** Gibt die Breite des Headers zurck. */
	private int _headerWidth = 0;

	/* ############# Konstruktor ############## */

	/**
	 * Konstruktor, dem ein Datensatz zugewiesen wird.
	 *
	 * @param head            Element, welches die hierarchische Struktur darstellt
	 * @param dataTableObject neuer Datensatz
	 */
	public Dataset(final HeaderGrid head, final DataTableObject dataTableObject) {
		_head = head;
		_dataTableObject = dataTableObject;
	}

	/* ############ Get- / Set-Methoden ############## */

	/**
	 * Gibt den Datensatz zurck.
	 *
	 * @return Datensatz
	 */
	public DataTableObject getDataTableObject() {
		return _dataTableObject;
	}

	/**
	 * Gibt die hierarchische Struktur des Datensatzes zurck.
	 *
	 * @return hierarchische Struktur des Datensatzes
	 */
	public RowData getRowData() {
		if(_rowData == null) {
			createRowData();
		}
		return _rowData;
	}

	/**
	 * Gibt die Komponente des Zeilenkopfes zurck.
	 *
	 * @param timeFormat das gewnschte Format
	 *
	 * @return Komponente des Zeilenkopfes
	 */
	public JComponent getRowHeaderRow(String timeFormat) {
		if(_rowHeaderRow == null) {
			_rowHeaderRow = createRowHeaderRow(_dataTableObject, timeFormat);
		}
		return _rowHeaderRow;
	}

	/**
	 * Gibt die Komponente der Datenzeile zurck.
	 *
	 * @return Komponente der Datenzeile
	 */
	public JComponent getViewportRow() {
		if(_viewportRow == null) {
			createViewportRow();
		}
		return _viewportRow;
	}

	/**
	 * Gibt die Hhe dieser Zeile in Pixel zurck.
	 *
	 * @return Hhe dieser Zeile
	 */
	public int getHeight() {
		if(_height == -1) {
			setHeight();
		}
		return _height;
	}

	/**
	 * Erstellt die Verbindungen zwischen den Daten und dem Header. Jedes Blattelement meldet sich beim entsprechenden Blatt im Header an, damit etwaige
	 * Grennderungen vom Header an die Datenstruktur bergeben werden kann.
	 */
	public void setLinks() {
		_rowHeaderRow = null;   // wenn die Elemente verbunden werden sollen, mssen die Komponenten erst gelscht
		_viewportRow = null;	// werden. Nach dem nchsten Erstellen, benutzen sie auch die Informationen aus dem
		// Header
		if(_rowData == null) {
			createRowData();
		}
		linkData(_rowData, _head);
		createViewportRow();
	}

	/** Entfernt alle Eintrge, auer den Datensatz und die Hhe einer Zeile. */
	public void unsetLinks() {
		// lscht alles auer ResultData und Height (Objekte einschlielich Verweise)
		if(_rowData != null) { // nur dann macht es Sinn, sonst gibt es nichts zum Lschen
			unlinkData(_rowData, _head);
		}
		removeComponents();
	}

	/* ################# Private Methoden ############# */

	/** Erzeugt aus einem Datensatz eine hierarchische Struktur. */
	private void createRowData() {
		if(_dataTableObject.getData() == null) {
//            _debug.info("Zu diesem Datensatz muss keine hierarchische Struktur erstellt werden.");
			_rowData = new RowData(_head, _dataTableObject);
		}
		else {
			_rowData = new RowData();
			transformData(_dataTableObject.getData(), _rowData);
		}
	}

	/**
	 * Wandelt den Datensatz vom Datenverteiler in eine hierachische Struktur um.
	 *
	 * @param data    Daten vom Datenverteiler
	 * @param rowData neue hierarchische Struktur
	 */
	private void transformData(Data data, RowData rowData) {
		if(data.isPlain()) {
			rowData.setValue(data.valueToString()); // Wert
		}
		else {
			if(data.isList()) {	// kein Array
				Iterator it = data.iterator();
				while(it.hasNext()) {
					Data nextData = (Data)it.next();
					RowData nextRowData = new RowData();
					transformData(nextData, nextRowData);
					rowData.addArrayElement(nextRowData);
				}
			}
			if(data.isArray()) {   // ein Array
				rowData.setIsArray(true);
				Data.Array dataArray = data.asArray();
				for(int i = 0, n = dataArray.getLength(); i < n; i++) {
					Data nextData = dataArray.getItem(i);
					RowSuccessor succ = new RowSuccessor();
//                    System.out.println(nextData.toString());   // ein Array hat viele RowData als Nachfolger
//                    System.out.println(nextData.isList());

					if(nextData.isList()) {
						// eine Liste in einem Array
						Iterator it = nextData.iterator();
						while(it.hasNext()) {
							Data nextNextData = (Data)it.next();
//                            System.out.println(nextNextData.toString());
							RowData nextRowData = new RowData();
							transformData(nextNextData, nextRowData);
							succ.addSuccessor(nextRowData);
						}
						// fr jedes i wird ein Successor im RowData angehngt
					}
					else {	// keine Liste
						RowData nextRowData = new RowData();
						transformData(nextData, nextRowData);
						succ.addSuccessor(nextRowData);
					}
//                    _debug.finest("Anzahl im Successor: " + succ.getSuccessors().size());
					rowData.addArrayElement(succ);
				}
			}
		}
	}

	/**
	 * Rekursive Hilfsmethode. Sie wird von {@link #setLinks()} aufgerufen. Die Verbindungen zum Spaltenheader werden hergestellt und die Breite der Komponenten,
	 * welche die Daten anzeigen wird initial festgelegt. Diese Methode wird nur ausgefhrt, wenn im Datensatz auch Daten vorhanden sind.
	 *
	 * @param rowData    darzustellende Daten
	 * @param headerGrid Spaltenheader
	 */
	private void linkData(RowData rowData, HeaderGrid headerGrid) {
		if(_dataTableObject != null) {					// _resultData.getData() != null
			if(_dataTableObject.getData() == null) {
//				System.out.println("_dataTableObject.getDataState() = " + _dataTableObject.getDataState());
				// Anmelden auf alle Spalten!
				// Idee: bis in die Bltter gehen, dort anmelden und die Summe ergibt dann die Breite des leeren Datensatzes
//				System.out.println("rowData = " + rowData);
				_headerWidth = 0;
				getHeaderWidth(headerGrid, rowData);
//				System.out.println("headerWidth = " + _headerWidth);
				rowData.setInitialWidth(_headerWidth);
			}
			else {
				if(rowData.getArray().isEmpty()) {	 // RowData ist Blatt
					int width = headerGrid.getHeaderElement().getSize().width + headerGrid.getSplitter().getSize().width;
					rowData.setInitialWidth(width);		 // Breite mitteilen
					headerGrid.addRowListener(rowData);	 // beim Listener anmelden
				}
				else {			// RowData ist kein Blatt
					// entweder sind alle vom Typ RowData oder vom Typ RowSuccessor je nachdem, ob es sich um ein Array handelt
					List array = rowData.getArray();
					if(!rowData.isArray()) {   // kein Array -> alle Nachfolger vom Typ RowData
						Iterator gridIt = headerGrid.getHeaderSuccessors().iterator();
						Iterator rowIt = array.iterator();
						while(rowIt.hasNext() && gridIt.hasNext()) {
							RowData nextRowData = (RowData)rowIt.next();
							HeaderGrid nextHeaderGrid = (HeaderGrid)gridIt.next();
							linkData(nextRowData, nextHeaderGrid);
						}
					}
					else {
						for(Iterator iterator = array.iterator(); iterator.hasNext();) {
							RowSuccessor rowSuccessor = (RowSuccessor)iterator.next();
							if(headerGrid.getHeaderSuccessors().isEmpty()) {
								for(final RowData rowData1 : rowSuccessor.getSuccessors()) {
									RowData nextRowData = (RowData)rowData1;
									linkData(nextRowData, headerGrid);
								}
							}
							else {
								Iterator gridIt = headerGrid.getHeaderSuccessors().iterator();
								Iterator succIt = rowSuccessor.getSuccessors().iterator();
								while(succIt.hasNext() && gridIt.hasNext()) {
									RowData nextRowData = (RowData)succIt.next();
									HeaderGrid nextHeaderGrid = (HeaderGrid)gridIt.next();
									linkData(nextRowData, nextHeaderGrid);
								}
							}
						}
					}
				}
			}
		}
	}

	private void getHeaderWidth(HeaderGrid headerGrid, RowData rowData) {
		List<HeaderGrid> succs = headerGrid.getHeaderSuccessors();
//		System.out.println("succs.size() = " + succs.size());
		if(succs.isEmpty()) {	// Blattknoten erreicht
			_headerWidth += (headerGrid.getHeaderElement().getSize().width + headerGrid.getSplitter().getSize().width);
			headerGrid.addRowListener(rowData);
		}
		else {
			for(HeaderGrid grid : succs) {
				getHeaderWidth(grid, rowData);
			}
		}
	}

	/**
	 * Rekursive Hilfsmethode. Sie entfernt die Verbindungen zwischen den Daten und dem Spaltenheader. Diese Methode wird nur ausgefhrt, wenn auch Daten vorhanden
	 * sind.
	 *
	 * @param rowData    Daten, die mit dem Spaltenheader verbunden sind
	 * @param headerGrid Spaltenheader
	 */
	private void unlinkData(RowData rowData, HeaderGrid headerGrid) {
		if(_dataTableObject != null) {
			if(rowData.getArray().isEmpty()) {
				headerGrid.removeRowListener(rowData);
			}
			else {
				// entweder sind alle vom Typ RowData oder vom Typ RowSuccessor
				List array = rowData.getArray();
				Object object = array.get(0);
				if(object instanceof RowData) {
					Iterator gridIt = headerGrid.getHeaderSuccessors().iterator();
					Iterator rowIt = array.iterator();
					while(rowIt.hasNext() && gridIt.hasNext()) {
						RowData nextRowData = (RowData)rowIt.next();
						HeaderGrid nextHeaderGrid = (HeaderGrid)gridIt.next();
						unlinkData(nextRowData, nextHeaderGrid);
					}
				}
				if(object instanceof RowSuccessor) {
					for(Iterator iterator = array.iterator(); iterator.hasNext();) {
						RowSuccessor rowSuccessor = (RowSuccessor)iterator.next();
						if(headerGrid.getHeaderSuccessors().isEmpty()) {
							for(final RowData rowData1 : rowSuccessor.getSuccessors()) {
								RowData nextRowData = (RowData)rowData1;
								unlinkData(nextRowData, headerGrid);
							}
						}
						else {
							Iterator gridIt = headerGrid.getHeaderSuccessors().iterator();
							Iterator succIt = rowSuccessor.getSuccessors().iterator();
							while(succIt.hasNext() && gridIt.hasNext()) {
								RowData nextRowData = (RowData)succIt.next();
								HeaderGrid nextHeaderGrid = (HeaderGrid)gridIt.next();
								unlinkData(nextRowData, nextHeaderGrid);
							}
						}
					}
				}
			}
		}
	}


	/**
	 * Erzeugt anhand der Daten eine neue Zeile im Zeilenheader, bestehend aus einem Zeitstempel und dem dazugehrenden Objekt.
	 *
	 * @param dataTableObject das Objekt der Online-Tabelle
	 * @param format          das Zeitformat, um Datum und Zeit darzustellen
	 *
	 * @return der erstellte Zeilen-Header
	 */
	private JComponent createRowHeaderRow(final DataTableObject dataTableObject, String format) {
		final ArchiveDataKind dataKind = dataTableObject.getDataKind();
		final SystemObject systemObject = dataTableObject.getObject();
		String dataKindText;
		String dataKindTooltipText;
		if(dataKind == ArchiveDataKind.ONLINE) {
			dataKindText = "OA";
			dataKindTooltipText = "online aktueller Datensatz";
		}
		else if(dataKind == ArchiveDataKind.ONLINE_DELAYED) {
			dataKindText = "ON";
			dataKindTooltipText = "online nachgelieferter Datensatz";
		}
		else if(dataKind == ArchiveDataKind.REQUESTED) {
			dataKindText = "NA";
			dataKindTooltipText = "nachgefordert aktueller Datensatz";
		}
		else if(dataKind == ArchiveDataKind.REQUESTED_DELAYED) {
			dataKindText = "NN";
			dataKindTooltipText = "nachgefordert nachgelieferter Datensatz";
		}
		else {
			dataKindText = "??";
			dataKindTooltipText = "Art des Datensatzes nicht bekannt";
		}
		JLabel dataKindLabel = new JLabel(dataKindText);
		dataKindLabel.setToolTipText(dataKindTooltipText);
		dataKindLabel.setBorder(new EtchedBorder());
		dataKindLabel.setHorizontalAlignment(SwingConstants.CENTER);
		dataKindLabel.setVerticalAlignment(SwingConstants.NORTH);

		JPanel dataKindPanel = new JPanel(new BorderLayout());
		dataKindPanel.add(dataKindLabel, BorderLayout.CENTER);

//		if (_dataTableObject.getTimingType() == TimingType.ARCHIVE_TIME || _dataTableObject.getTimingType() == TimingType.DATA_TIME) {
//		} else {
//			long dataIndex = _dataTableObject.getDataIndex();
//			time = (dataIndex >> 32) + "#" + ((dataIndex >> 2) & 0x3fffffff) + "#" + (dataIndex & 0x3);
//		}
		DateFormat timeFormat = new SimpleDateFormat(format);
		String archiveTime = timeFormat.format(new Date(dataTableObject.getArchiveTime()));
		String dataTime = timeFormat.format(new Date(dataTableObject.getDataTime()));
		long dataIndex = _dataTableObject.getDataIndex();
		String dataIndexString = (dataIndex >>> 32) + "#" + ((dataIndex >> 2) & 0x3fffffff) + "#" + (dataIndex & 0x3);

		String labelText = "";
		if(_dataTableObject.getTimingType() == TimingType.ARCHIVE_TIME) {
			labelText = archiveTime;
		}
		else if(_dataTableObject.getTimingType() == TimingType.DATA_INDEX) {
			labelText = dataIndexString;
		}
		else {
			labelText = dataTime;
		}

		String toolTipText = "<html>";
		toolTipText += "Datenzeit: " + dataTime + "<br>";
		if(dataTableObject.getArchiveTime() > 0) {
			toolTipText += "Archivzeit: " + archiveTime + "<br>";
		}
		toolTipText += "Datenindex: " + dataIndexString;
		toolTipText += "</html>";

//		JLabel timeLabel = new JLabel(timeFormat.format(new Date(dataTableObject.getDataTime())));
		JLabel timeLabel = new JLabel(labelText);
		timeLabel.setBorder(new EtchedBorder());
		timeLabel.setHorizontalAlignment(SwingConstants.CENTER);
		timeLabel.setVerticalAlignment(SwingConstants.NORTH);
//		String toolTipText = "<html>";
//		if (dataTableObject.getArchiveTime() != 0) {
//			toolTipText += "Archivzeit: " + timeFormat.format(new Date(dataTableObject.getArchiveTime())) + "<br>";
//		}
//		long dataIndex = _dataTableObject.getDataIndex();
//		toolTipText += "Datenindex: " + ((dataIndex >> 32) + "#" + ((dataIndex >> 2) & 0x3fffffff) + "#" + (dataIndex & 0x3));
//		toolTipText += "</html>";

		timeLabel.setToolTipText(toolTipText);

		String object = systemObject.getNameOrPidOrId();
		JLabel objectLabel = new JLabel(object);
		objectLabel.setBorder(new EtchedBorder());
		objectLabel.setHorizontalAlignment(SwingConstants.CENTER);
		objectLabel.setVerticalAlignment(SwingConstants.NORTH);

		JPanel gridPanel = new JPanel(new GridLayout(1, 2));
		gridPanel.add(timeLabel);
		gridPanel.add(objectLabel);

		JPanel rowHeaderPanel = new JPanel(new BorderLayout());
		rowHeaderPanel.add(dataKindPanel, BorderLayout.WEST);
		rowHeaderPanel.add(gridPanel, BorderLayout.CENTER);

		return rowHeaderPanel;
	}

	/** Erzeugt aus einem Datensatz eine Swing-Komponente, damit die Daten angezeigt werden knnen. */
	private void createViewportRow() {
		if(_rowData == null) {
			createRowData();
		}
		_viewportRow = _rowData.createComponent();
		_height = _viewportRow.getPreferredSize().height;
	}

	/**
	 * Ermittelt die Hhe der diesen Datensatz reprsentierenden Swing-Komponente. Falls die Komponenten extra fr die Ermittlung der Hhe erzeugt werden, dann
	 * werden sie anschlieend auch wieder gelscht.
	 */
	private void setHeight() {
		if(_viewportRow == null) {
			createViewportRow();
			removeComponents();
		}
		else {
			_height = _viewportRow.getPreferredSize().height;
		}
	}

	/** Lscht alle nicht mehr bentigten Komponenten, auer des Datensatzes, des Spaltenheaders und der Hhe. */
	private void removeComponents() {
		_rowData = null;
		_rowHeaderRow = null;
		_viewportRow = null;
	}
}
