/*
 * Copyright 2006 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.puk.config.
 * 
 * de.bsvrz.puk.config is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.puk.config is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.puk.config; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package de.bsvrz.puk.config.main.authentication;

import de.bsvrz.dav.daf.main.config.ConfigurationTaskException;
import de.bsvrz.dav.daf.main.impl.config.request.RequestException;

/**
 * Dieses Interface stellt Methoden zur Verfgung, mit der sich ein Benutzer Authentifizieren und Verwaltungsaufgaben
 * anstossen kann.
 * <p/>
 * Die Methode {@link #isValidUser} prft ob eine bergebene Benutzer/Passwort kombination gltig ist.
 * <p/>
 * Die Methode {@link #processTask} beauftragt die Konfiguration eine der folgenden Auftrge auszufhren:<br> - Neuer
 * Benutzer anlegen<br> - Einmal-Passwort erzeugen<br> - Rechte eines Benutzers ndern<br> - Passwort eines Benuzters
 * ndern<br>
 * <p/>
 * Alle Informationen die fr die oben genannten Aufgaben bentigt werden, werden verschlsselt bertragen.
 * <p/>
 * Die Methode {@link #getText} liefert einen Zufallstext. Der Zufallstext wird bentigt um "Reply-Attacken"
 * (verschicken von Kopien bestimmter Telegramme) zu verhindern. Dieser Text muss in allen Telegrammen, die fr die
 * {@link #processTask} Methode bentigt werden, verschlsselt bertragen werden. Danach darf der verschlsselt
 * bertragenen Text nicht mehr fr andere Aufgaben funktionieren.
 *
 * @author Achim Wullenkord (AW), Kappich Systemberatung
 * @version $Revision:5077 $ / $Date:2007-09-02 14:48:31 +0200 (So, 02 Sep 2007) $ / ($Author:rs $)
 */
public interface Authentication {
	/**
	 * Die Implementierung dieser Methode stellt die Authentifizierung des Benutzers sicher. Dafr wird das original
	 * Passwort mit dem bergebenen <code>authentificationText</code> verschlsselt und mit dem bergebenen verschlsselten
	 * Passwort verglichen. Sind beide Passwrter gleich, und der bergebene Benutzername stimmt mit dem Benutzernamen des
	 * original Passworts berein, so war die Authentifkation erfolgreich.
	 * <p/>
	 * Konnte das original Passwort nicht benutzt werden, muss geprft werden, ob es ein Einmal-Passwort gibt. Das
	 * Einmal-Passwort muss das derzeit aktuell gltige sein und muss mit dem bergebenen verschlsseltem Passwort
	 * bereinstimmen. Gibt es ein entsprechendes Einmal-Passwort, so ist es fr immer zu sperren.
	 * <p/>
	 * Konnte kein Passwort gefunden werden, wird eine IllegalArgumentException geworfen.
	 *
	 * @param username					Benutzername, der zu dem bergebenen verschlsselten Passwort gehrt
	 * @param encryptedPassword		   Passwort, das mit dem bergebenen Text <code>authentificationText</code>
	 *                                    verschlsselt wurde
	 * @param authentificationText		Text, der benutzt wurde um das bergebene Passwort <code>encryptedPassword</code>
	 *                                    zu verschlsseln
	 * @param authentificationProcessName Name des Verschlsslungsverfahren, das benutzt wurde. Mit diesem Verfahren wird
	 *                                    das Originalpasswort verschlsselt
	 * @throws Exception				Fehler beim schreiben der neuen Informationen oder ein technisches Problem beim
	 *                                  verschlsseln der Daten
	 * @throws IllegalArgumentException Dem Benutzernamen konnte das Passwort nicht zugeorndet werden oder der Benutzer war
	 *                                  unbekannt
	 */
	public void isValidUser(String username, byte[] encryptedPassword, String authentificationText, String authentificationProcessName) throws Exception, IllegalArgumentException;

	/**
	 * Bearbeitet eine der folgenden Aufgaben:<br> - Neuer Benutzer anlegen<br> - Einmal-Passwort erzeugen<br> - Rechte
	 * eines Benutzers ndern<br> - Passwort eines Benuzters ndern<br>
	 *
	 * @param usernameCustomer	  Benutzer, der den Auftrag erteilt
	 * @param encryptedMessage	  verschlsselte Aufgabe, die ausgefhrt werden soll
	 * @param encryptionProcessName Verschlsslungsverfahren mit der <code>encryptedMessage</code> erstellt wurde
	 * @throws ConfigurationTaskException Der Auftrag, der durch die Konfiguration ausgefhrt werden sollte, konnte nicht
	 *                                    durchgefhrt werden, weil bestimmte Parameter nicht erfllt waren. Welche
	 *                                    Parameter dies genau sind, hngt vom jeweiligen Auftrag ab, so kann zum Beispiel
	 *                                    ein Passwort fehlerhaft gewesen sein oder der Benutzer besitzt nicht die ntigen
	 *                                    Rechte um einen Auftrag dieser Art anzustoen. Wenn der Auftrag erneut
	 *                                    bermittelt werden wrden, mit den richtigen Parametern, knnte er ausgefhrt
	 *                                    werden.
	 * @throws RequestException		   Der Auftrag konnte aufgrund eines technischen Fehlers nicht ausgefhrt werden
	 *                                    (defektes Speichermedium, Fehler im Dateisystem, usw.). Erst wenn dieser Fehler
	 *                                    behoben ist, knnen weitere Auftrge ausgefhrt werden.
	 */
	public void processTask(String usernameCustomer, byte[] encryptedMessage, String encryptionProcessName) throws ConfigurationTaskException, RequestException;

	/**
	 * Erzeugt einen Zufallstext und gibt diesen als Byte-Array zurck.
	 *
	 * @return Zufallstext
	 */
	public byte[] getText();

	/**
	 * Diese Methode wird aufgerufen, wenn das System heruntergefahren wird. Es ist ein Zustand herzustellen, der es ermglicht das System wieder zu starten.
	 */
	public void close();
}
